/* miscfncs.c
 *
 *
 * $Author: mtm $
 * $Date: 1995/05/09 01:27:15 $
 * $Id: miscfncs.c,v 1.1 1995/05/09 01:27:15 mtm Exp mtm $
 *
 * Mike Macgirvin <Mike_Macgirvin@CAMIS.Stanford.EDU>
 *
 * Copyright 1995 by The Leland Stanford Junior University.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notices appear in all copies and that both the
 * above copyright notices and this permission notice appear in supporting
 * documentation, and that the name of The Leland Stanford Junior University 
 * not be used in advertising or publicity pertaining to distribution of the 
 * software without specific, written prior permission.  This software is made 
 * available "as is", and THE LELAND STANFORD JUNIOR UNIVERSITY DISCLAIMS ALL 
 * WARRANTIES, EXPRESS OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING 
 * WITHOUT LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE LELAND STANFORD JUNIOR 
 * UNIVERSITY BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR 
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, 
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
 * SOFTWARE.
 *
 */

/*
 * $Log: miscfncs.c,v $
 * Revision 1.1  1995/05/09  01:27:15  mtm
 * Initial revision
 *
 * 
 */

#include "ml.h"



#ifdef __STDC__
Mailbox *find_mailbox_from_mailstream(MAILSTREAM *mailstream)
#else
Mailbox *find_mailbox_from_mailstream(mailstream)
     MAILSTREAM *mailstream;
#endif
{
  Mailbox *mailbox;

  for(mailbox = session.mailboxes; mailbox ; mailbox = mailbox->next)
    if(mailbox->mailstream == mailstream)
      return(mailbox);
  return(NULL);
}

#ifdef __STDC__
void free_binary_buffer(Binary_Buffer *binary_buffer)
#else
void free_binary_buffer(binary_buffer)
     Binary_Buffer *binary_buffer;
#endif
{
  if(binary_buffer) {
    if(binary_buffer->data)
      fs_give((void **) &binary_buffer->data);
    fs_give((void **) &binary_buffer);
  }
  return;
}

#ifdef __STDC__
Remote_Auth *new_remote_auth(void)
#else
Remote_Auth *new_remote_auth()
#endif
{
  Remote_Auth *remote_auth = (Remote_Auth *) fs_get(sizeof(Remote_Auth));

  remote_auth->hostname     = NULL;
  remote_auth->authtype     = NULL;
  remote_auth->username     = NULL;
  remote_auth->password     = NULL;
  remote_auth->workspace    = NULL;
  remote_auth->done         = 0;
  remote_auth->timer        = (XtIntervalId) 0;
  remote_auth->next         = NULL;

  return(remote_auth);
}

#ifdef __STDC__
void free_remote_auth(Remote_Auth *remote_auth)
#else
void free_remote_auth(remote_auth)
     Remote_Auth *remote_auth;
#endif
{
  if(remote_auth) {
    if(remote_auth->password) {
      wipeout(remote_auth->password);
      fs_give((void **) &remote_auth->password);
    }
    if(remote_auth->workspace)
      fs_give((void **) &remote_auth->workspace);
    if(remote_auth->username)
      fs_give((void **) &remote_auth->username);
    if(remote_auth->hostname)
      fs_give((void **) &remote_auth->hostname);
    if(remote_auth->authtype)
      fs_give((void **) &remote_auth->authtype);
    
    fs_give((void **) &remote_auth);
  }
  return;
}

#ifdef __STDC__
String_List *new_string_list(void)
#else
String_List *new_string_list()
#endif
{
  String_List *string_list = (String_List *) fs_get(sizeof(String_List));

  string_list->string      = NULL;
  string_list->next        = NULL;

  return(string_list);
}

#ifdef __STDC__
void free_string_list(String_List *string_list, Boolean recurse)
#else
void free_string_list(string_list,recurse)
     String_List *string_list;
     Boolean recurse;
#endif
{
  if(string_list == NULL)
    return;
  if(recurse)
    free_string_list(string_list->next, recurse);

  if(string_list->string != NULL)
    fs_give((void **) &string_list->string);

  fs_give((void **) &string_list);
  return;
}


#ifdef __STDC__
Server_Config *new_server_config(void)
#else
Server_Config *new_server_config()
#endif
{
  Server_Config *server_config 
    = (Server_Config *) fs_get(sizeof(Server_Config));

  server_config->hostname = NULL;
  server_config->service  = NULL;
  server_config->port     = 0;
  server_config->priority = 0;
  server_config->next     = NULL;
  server_config->prev     = NULL;

  return(server_config);
}

#ifdef __STDC__
void free_server_config(Server_Config *server_config, Boolean recurse)
#else
void free_server_config(server_config,recurse)
     Server_Config *server_config;
     Boolean recurse;
#endif
{
  if(server_config == NULL)
    return;
  if(recurse)
    free_server_config(server_config->next, recurse);

  if(server_config->hostname != NULL)
    fs_give((void **) &server_config->hostname);

  fs_give((void **) &server_config);
  return;
}

#ifdef __STDC__
Filter_Map *new_filter_map(void)
#else
Filter_Map *new_filter_map()
#endif
{
  Filter_Map *filter_map = (Filter_Map *) fs_get(sizeof(Filter_Map));

  filter_map->mailboxname   = NULL;
  filter_map->filename      = NULL;
  filter_map->sequence      = 0;
  filter_map->next          = NULL;

  return(filter_map);
}

#ifdef __STDC__
Filter_Map *get_filter_map(Lview *lview)
#else
Filter_Map *get_filter_map(lview)
     Lview *lview;
#endif
{
  Filter_Map *filter_map = NULL;
  
  if((lview == NULL) 
     || (lview->mailbox == NULL) 
     || (lview->mailbox->imapname == NULL))
    return(NULL);

  for(filter_map = session.filter_map; filter_map; 
      filter_map = filter_map->next) {
    if((strcmp(filter_map->mailboxname,
	       lview->mailbox->imapname)) == STRMATCH)
      return(filter_map);
  }
  return(NULL);
}

#ifdef __STDC__
Filter_Map *generate_filter_map(Lview *lview)
#else
Filter_Map *generate_filter_map(lview)
     Lview *lview;
#endif
{
  char buffer[MAXPATHLEN];
  Filter_Map *prev = NULL;
  Filter_Map *filter_map = new_filter_map();
 
  filter_map->mailboxname = cpystr(lview->mailbox->imapname);
 
  if(session.filter_map == NULL) {
    session.filter_map = filter_map;
    sprintf(buffer,"%s/%s.1", local_auth.homedir,FILTERFILE);
    filter_map->sequence = 1;
    filter_map->filename = cpystr(buffer);
  }
  else {
    for(prev = session.filter_map; prev->next; prev = prev->next)
      ;
    filter_map->sequence = prev->sequence + 1;
    sprintf(buffer,"%s/%s.%d",local_auth.homedir,
	    FILTERFILE, filter_map->sequence);
    filter_map->filename = cpystr(buffer);
    prev->next = filter_map;
  }
  (void) save_defaults();
  return(filter_map);
}

#ifdef __STDC__
Server_List *new_server_list(void)
#else
Server_List *new_server_list()
#endif
{
  Server_List *server_list = (Server_List *) fs_get(sizeof(Server_List));

  server_list->name             = NULL;
  server_list->type             = MAILBOX_TYPE_MAIL;
  server_list->active           = FALSE;
  server_list->everything       = FALSE;
  server_list->last_pattern     = NULL;
  server_list->mailbox_list     = NULL;
  server_list->mailbox_all_list = NULL;
  server_list->next             = NULL;
  server_list->prev             = NULL;

  return(server_list);
}

#ifdef __STDC__
Mailbox_List *new_mailbox_list(void)
#else
Mailbox_List *new_mailbox_list()
#endif
{
  Mailbox_List *mailbox_list = (Mailbox_List *) fs_get(sizeof(Mailbox_List));

  mailbox_list->name      = NULL;
  mailbox_list->next      = NULL;
  mailbox_list->prev      = NULL;

  return(mailbox_list);
}

#ifdef __STDC__
void free_message_list(Message_List *message_list, 
		       Boolean recurse, Boolean free_messages)
#else 
void free_message_list(message_list, recurse, free_messages) 
     Message_List *message_list;
     Boolean recurse;
     Boolean free_messages;
#endif
{
  if(message_list == NULL)
    return;
  if(recurse)
    free_message_list(message_list->next, recurse, free_messages);
  if(free_messages)
    free_message(message_list->message);
  fs_give((void **) &message_list);
  return;
}

#ifdef __STDC__
Mailbox *new_mailbox(void)
#else
Mailbox *new_mailbox()
#endif
{
  Mailbox *mailbox = (Mailbox *) fs_get(sizeof(Mailbox));

  mailbox->mailstream     = NULL;
  mailbox->host           = NULL;
  mailbox->port           = 0L;
  mailbox->auth_type      = 0L;
  mailbox->imap_protocol  = 0;
  mailbox->mailboxname    = NULL;
  mailbox->imapname       = NULL;
  mailbox->type           = MAILBOX_TYPE_MAIL;
  mailbox->fetched        = FALSE;
  mailbox->auto_scan      = FALSE;
  mailbox->no_update      = FALSE;
  mailbox->bg_timer       = (XtIntervalId) 0;
  mailbox->first_fetch    = 0L;
  mailbox->has_new_mail   = FALSE;
  mailbox->nmsgs          = 0L;
  mailbox->found          = NULL;
  mailbox->flags_changed  = 0L;
  mailbox->update_needed  = 0;
  mailbox->refresh_needed = 0;
  mailbox->timer          = (XtIntervalId) 0;
  mailbox->next           = NULL;
  mailbox->prev           = NULL;

  return(mailbox);
}

#ifdef __STDC__
void free_mailbox(Mailbox *mailbox)
#else
void free_mailbox(mailbox)
     Mailbox *mailbox;
#endif
{
  if(mailbox != NULL) {
    if(mailbox->host)
      fs_give((void **) &mailbox->host);
    if(mailbox->mailboxname)
      fs_give((void **) &mailbox->mailboxname);
    if(mailbox->imapname)
      fs_give((void **) &mailbox->imapname);

    if(mailbox->found)
      free_message_list(mailbox->found,TRUE,FALSE);

    fs_give((void **) &mailbox);
  }
  return;
}

#ifdef __STDC__
View *new_view(void)
#else
View *new_view()
#endif
{
  View *view = (View *) fs_get(sizeof(View));
  
  view->is_realized       = FALSE;
  view->list_is_realized  = FALSE;
  view->hide_empty        = preferences.hide_empty;
  return view;
}

#ifdef __STDC__
Lview *new_lview(void)
#else
Lview *new_lview()
#endif
{
  Lview *lview = (Lview *) fs_get(sizeof(Lview));

  lview->title            = NULL;
  lview->mailstream       = NULL;
  lview->mailbox          = NULL;
  lview->message_list     = NULL;
  lview->count            = 0L;
  lview->pruned           = 0L;
  lview->prune_happened   = FALSE;
  lview->num_selected     = 0L;
  lview->has_new          = FALSE;
  lview->selected         = FALSE;
  lview->number           = 0L;
  lview->button_state     = BTN_ON;
  lview->level            = 0;
  lview->representation   = REP_ASCENDING;
  lview->scripting        = FALSE;
  lview->script_arg       = NULL;
  lview->script_header    = HEADER_UNKNOWN;

  lview->filter           = NULL;
  lview->parent           = NULL;
  lview->next             = NULL;
  lview->prev             = NULL;
  
  return(lview);
}

#ifdef __STDC__
void free_lview(Lview *lview)
#else
void free_lview(lview)
     Lview *lview;
#endif
{
  if(lview) {
    if(lview->title)
      fs_give((void **) &lview->title);
    if(lview->script_arg)
      fs_give((void **) &lview->script_arg);

    fs_give((void **) &lview);
  }
  return;
}

#ifdef __STDC__
Filter *new_filter(void)
#else
Filter *new_filter()
#endif
{
  Filter *filter = (Filter *) fs_get(sizeof(Filter));

  filter->mailboxname     = NULL;
  filter->name            = NULL;
  filter->text            = NULL;
  filter->action          = NULL;
  filter->search_tree     = NULL;
  filter->search_types    = 0L;
  filter->archived        = FALSE;
  filter->filter_type     = FILTER_TYPE_UNKNOWN;
  filter->representation  = REP_ASCENDING;
  filter->number          = 0L;
  filter->next            = NULL;
  filter->prev            = NULL;

  return(filter);
}

#ifdef __STDC__
void free_filter(Filter *filter)
#else
void free_filter(filter)
     Filter *filter;
#endif
{
  if(filter) {
    if(filter->mailboxname) 
      fs_give((void **) &filter->mailboxname);
    if(filter->name)
      fs_give((void **) &filter->name);
    if(filter->action)
      fs_give((void **) &filter->action);
    if(filter->text)
      fs_give((void **) &filter->text);

    cut_down_the_tree(filter->search_tree);

    fs_give((void **) &filter);
  }
  return;
}

#ifdef __STDC__
Message_List *new_message_list(void)
#else
Message_List *new_message_list()
#endif
{
  Message_List *message_list = (Message_List *) fs_get(sizeof(Message_List));

  message_list->message     = NULL;
  message_list->selected    = FALSE;
  message_list->number      = 0L;
  message_list->prev        = NULL;
  message_list->next        = NULL;

  return(message_list);
}
     
#ifdef __STDC__
Message *new_message(void)
#else
Message *new_message()
#endif
{
  Message *message = (Message *) fs_get(sizeof(Message));
  
  message->msgno            = 0L;
  message->mailbox          = NULL;
  message->mailstream       = NULL;
  message->envelope         = NULL;
  message->body             = NULL;
  message->longcache        = NULL;
  message->viewline         = NULL;
  message->in_view          = FALSE;
  message->fetched          = FALSE;
  message->sender           = NULL;
  message->subject          = NULL;

  return(message);
}

#ifdef __STDC__
void free_message(Message *message)
#else
void free_message(message)
     Message *message;
#endif
{
  if(message == NULL)
    return;

  if(message->viewline)
    XmStringFree(message->viewline);
  if(message->sender)
    fs_give((void **) &message->sender);
  if(message->subject)
    fs_give((void **) &message->subject);
  fs_give((void **) &message);
  return;
}

#ifdef __STDC__
Edit *new_edit(void)
#else
Edit *new_edit()
#endif
{
  Edit *edit = (Edit *) fs_get(sizeof(Edit));
  
  edit->is_realized         = FALSE;
  edit->buffer              = NULL;
  edit->begin               = 0;
  edit->end                 = 0;

  return(edit);
}

#ifdef __STDC__
void free_edit(Edit *edit)
#else
void free_edit(edit)
     Edit *edit;
#endif
{
  if(edit->buffer)
    fs_give((void **) &edit->buffer);
  fs_give((void **) &edit);
}

#ifdef __STDC__
Dir_Struct *new_dir_struct(void)
#else
Dir_Struct *new_dir_struct()
#endif
{
  Dir_Struct *dir_struct = (Dir_Struct *) fs_get(sizeof(Dir_Struct));

  dir_struct->name       = NULL;
  dir_struct->directory  = FALSE;
  dir_struct->next       = NULL;

  return(dir_struct);
}

#ifdef __STDC__
void free_dir_structs(Dir_Struct *dir_struct)
#else
void free_dir_structs(dir_struct)
     Dir_Struct *dir_struct;
#endif
{
  if(dir_struct == NULL)
    return;
  free_dir_structs(dir_struct->next);
  if(dir_struct->name)
    fs_give((void **) &dir_struct->name);
  fs_give((void **) &dir_struct);

  return;
}

#ifdef __STDC__
Address_Book_Info *new_address_book_info(void)
#else
Address_Book_Info *new_address_book_info()
#endif
{
  Address_Book_Info *address_book_info =
    (Address_Book_Info *) fs_get(sizeof(Address_Book_Info));

  address_book_info->name    = NULL;
  address_book_info->address = NULL;
  address_book_info->comment = NULL;
  address_book_info->next    = NULL;
  address_book_info->prev    = NULL;

  return(address_book_info);
}

#ifdef __STDC__
void free_address_book_info(Address_Book_Info *address_book_info)
#else
void free_address_book_info(address_book_info)
     Address_Book_Info *address_book_info;
#endif
{
  if(address_book_info != NULL) {
    if(address_book_info->name != NULL)
      fs_give((void **) &address_book_info->name);
    if(address_book_info->address != NULL)
      fs_give((void **) &address_book_info->address);
    if(address_book_info->comment != NULL)
      fs_give((void **) &address_book_info->comment);
    fs_give((void **) &address_book_info);
  }
  return;
}

#ifdef __STDC__
Read_Info *new_read_info(void)
#else 
Read_Info *new_read_info() 
#endif
{
  Read_Info *read_info = (Read_Info *) fs_get(sizeof(Read_Info));

  read_info->message_type     = MESSAGE_NORMAL;
  read_info->mailbox_type     = MAILBOX_TYPE_MAIL;
  read_info->mailstream       = NULL;
  read_info->msgno            = 0L;
  read_info->fetched          = FALSE;
  read_info->expunged         = FALSE;
  read_info->selected         = FALSE;
  read_info->server           = NULL;
  read_info->current_part_str = NULL;
  read_info->base_part_str    = NULL;
  read_info->short_header     = NULL;
  read_info->header           = NULL;
  read_info->current_text     = NULL;
  read_info->alt_text         = NULL;
  read_info->current_part     = NULL;
  read_info->body             = NULL;
  read_info->envelope         = NULL;
  read_info->attachments      = NULL;
  read_info->button_state     = BTN_ON | BTN_LONG_HEADER;
  read_info->number           = 0L;
  read_info->selection        = 0L;
  read_info->show_long_header = FALSE;
  read_info->show_entire      = FALSE;
  read_info->show_alt         = FALSE;
  read_info->rfc_msg_leaf     = FALSE;
  read_info->part_list        = NULL;
  read_info->next             = NULL;
  read_info->prev             = NULL;

  return(read_info);
}

#ifdef __STDC__
void free_read_info(Read_Info *read_info)
#else
void free_read_info(read_info)
     Read_Info *read_info;
#endif
{
  if(read_info == NULL)
    return;
  read_info->msgno = 0L;
  if(read_info->current_part_str)
    fs_give((void **) &read_info->current_part_str);
  if(read_info->base_part_str)
    fs_give((void **) &read_info->base_part_str);
  if(read_info->short_header)
    fs_give((void **) &read_info->short_header);
  if(read_info->header)
    fs_give((void **) &read_info->header);
  if(read_info->current_text)
    fs_give((void **) &read_info->current_text);
  if(read_info->alt_text)
    fs_give((void **) &read_info->alt_text);
  if(read_info->part_list)
    destroy_part_list(read_info->part_list);

  fs_give((void **) &read_info);
}

#ifdef __STDC__
Read *new_read(void)
#else
Read *new_read()
#endif
{
  Read *read = (Read *) fs_get(sizeof(Read));
  
  read->current       = NULL;
  read->read_info     = NULL;
  read->is_realized   = FALSE;
  read->list_is_realized = FALSE;
  return(read);
}

#ifdef __STDC__
Mime_Handler *new_mime_handler(void)
#else
Mime_Handler *new_mime_handler()
#endif
{
  Mime_Handler *mime_handler = (Mime_Handler *) fs_get(sizeof(Mime_Handler));

  mime_handler->type          = TYPEOTHER;
  mime_handler->subtype       = NULL;
  mime_handler->view_builtin  = FALSE;
  mime_handler->builtin       = NULL;
  mime_handler->view          = NULL;
  mime_handler->compose       = NULL;
  mime_handler->next          = NULL;

  return(mime_handler);
}

#ifdef __STDC__
Note *new_note(void)
#else
Note *new_note()
#endif
{
  Note *note = (Note *) fs_get(sizeof(Note));
  note->is_realized = FALSE;
  note->autosave = (XtIntervalId) 0;
  return(note);
}


#ifdef __STDC__
Button_Menu *new_button_menu(void)
#else
Button_Menu *new_button_menu()
#endif
{
  Button_Menu *button_menu = (Button_Menu *) fs_get(sizeof(Button_Menu));
  button_menu->label = NULL;
  button_menu->func = NULL;
  button_menu->widget = NULL;
  button_menu->statemap = 0L;
  button_menu->next = NULL;
  button_menu->prev = NULL;

  return(button_menu);

}

#ifdef __STDC__
void free_button_menu(Button_Menu *button_menu)
#else
void free_button_menu(button_menu)
     Button_Menu *button_menu;
#endif
{
  if(button_menu->label)
    fs_give((void **) &button_menu->label);
  fs_give((void **) &button_menu);
  return;

}
