/* ml.c 
 *
 *
 * $Author: mtm $
 * $Date: 1995/05/09 01:27:15 $
 * $Id: ml.c,v 1.1 1995/05/09 01:27:15 mtm Exp mtm $
 *
 * Mike Macgirvin <Mike_Macgirvin@CAMIS.Stanford.EDU>
 *
 * Copyright 1995 by The Leland Stanford Junior University.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notices appear in all copies and that both the
 * above copyright notices and this permission notice appear in supporting
 * documentation, and that the name of The Leland Stanford Junior University 
 * not be used in advertising or publicity pertaining to distribution of the 
 * software without specific, written prior permission.  This software is made 
 * available "as is", and THE LELAND STANFORD JUNIOR UNIVERSITY DISCLAIMS ALL 
 * WARRANTIES, EXPRESS OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING 
 * WITHOUT LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE LELAND STANFORD JUNIOR 
 * UNIVERSITY BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR 
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, 
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
 * SOFTWARE.
 *
 */

/*
 * $Log: ml.c,v $
 * Revision 1.1  1995/05/09  01:27:15  mtm
 * Initial revision
 *
 */

#include "ml.h"
#include <locale.h>

/* global structures and variables */

Widget top;
Widget status;
XtAppContext context;
Preferences preferences;
Local_Auth local_auth;
Session session;        /* Most everything in the program is found here */
Display *display;
int errfd[2];
Boolean no_prefs = FALSE;
Boolean silent_pipe = FALSE; 
char **locale_messages = NULL;

XtActionsRec global_actions[] = {
  { "complete-address", (XtActionProc) complete_address },
  { "search-text",      (XtActionProc) search_text      },
  { "hex-input",        (XtActionProc) hex_input        },
};

/* IMAP registered drivers */

extern DRIVER 
imapdriver,   mboxdriver, bezerkdriver,
tenexdriver,    nntpdriver, 
newsdriver,    dummydriver,
pop3driver,    mtxdriver,
mhdriver,      mmdfdriver,
philedriver;

Menu main_file_menu[] = {

  { NULL, "open", NUL_TERM, 
      main_open, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "save_mail_init", NUL_TERM,
      save_mlinit, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "address_book", NUL_TERM,
      main_address_book, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "note_book", NUL_TERM,
      main_note_book, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "compose", NUL_TERM,
      main_compose, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "buttons", NUL_TERM,
      main_buttons, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "close", NUL_TERM, 
      main_close, NULL, 0, NULL, NULL, BTN_NOMAILBOXESOPEN },
  { NULL, "exit", NUL_TERM, 
      bye,   NULL, 0, NULL, NULL, BTN_ON },
};

Menu main_option_menu[] = {
  { NULL, "preferences", NUL_TERM,
      main_preferences, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "more_preferences", NUL_TERM,
      main_more_preferences, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "raise_view", NUL_TERM, 
      main_raise_view, NULL, 0, NULL, NULL, BTN_NOMAILBOXESOPEN },
  { NULL, "raise_read", NUL_TERM, 
      main_raise_read, NULL, 0, NULL, NULL, BTN_NOREADOPEN },
  { NULL, "raise_compose", NUL_TERM, 
      main_raise_compose, NULL, 0, NULL, NULL, BTN_NOCOMPOSEOPEN },
  { NULL, "prune_log", NUL_TERM, 
      main_prune_log, NULL, 0, NULL, NULL, BTN_ON },

};


Menu main_help_menu[] = {
  { NULL, "tophelp", NUL_TERM,
      main_help, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "keyboard_help", NUL_TERM,
      main_keyboard_help, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "config_help", NUL_TERM,
      main_config_help, NULL, 0, NULL, NULL, NUL_TERM },
  { NULL, "resources_help", NUL_TERM,
      main_resources_help, NULL, 0, NULL, NULL, NUL_TERM },
  { NULL, "intl_help", NUL_TERM,
      main_intl_help, NULL, 0, NULL, NULL, NUL_TERM },
  { NULL, "version_help", NUL_TERM,
      main_version_help, NULL, 0, NULL, NULL, BTN_ON },
};

Menu main_menu[] = {
  { NULL, "file_menu", NUL_TERM, 
      NULL, main_file_menu, XtNumber(main_file_menu), 
      NULL, NULL, BTN_ON },
  { NULL, "option_menu", NUL_TERM,
      NULL, main_option_menu, XtNumber(main_option_menu),
      NULL, NULL, BTN_ON },
  { NULL, "HELP_menu", NUL_TERM, 
      NULL, main_help_menu, XtNumber(main_help_menu),
      NULL, NULL, BTN_ON },
};

/* Begin Program */

#ifdef __STDC__
void main(int argc,char **argv, char **envp)
#else
void main(argc,argv,envp)
     int argc;
     char ** argv;
     char ** envp;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  char buffer[FILEBUFFLEN];
  char envstring[256];


  if(getenv(ML_SETLANG))
#if XtSpecificationRelease > 4
    XtSetLanguageProc(NULL, NULL, NULL); 
#else
    setlocale(LC_ALL, EMPTYSTR); 
#endif


  /***********************************************************************
   *              Create top level application shell - start X           *
   ***********************************************************************/

  top = XtAppInitialize((XtAppContext*) &context,
			PROGRAM,
			NULL, 0, &argc,argv, 
#ifdef FALLBACK
			fallback_resources, 
#else /* FALLBACK */
			NULL,
#endif /* FALLBACK */
			args, n);      
  n = 0;

  setup_editres(top);

  XtAppAddActions(context, global_actions, XtNumber(global_actions));

  (void) start_up();

  (void) create_main_window();

  XtRealizeWidget(top); 
  display = XtDisplay(top);

  (void) create_cursors(display);

  (void) make_icons();

  if(ml_icon != (Pixmap) None)
    XtVaSetValues(top, XmNiconPixmap, ml_icon, NULL);

  /* 
   * For the poor folks on X terminals and PC's who start the
   * program as "ml -display {something}" and wonder why their 
   * attachment handlers and other sub-processes don't work.
   */

  sprintf(envstring,"DISPLAY=%s",XDisplayString(display));
  (void) putenv(envstring);
					     
  sprintf(buffer,"Welcome to %s, Version %s\n",PROGRAM,MLVERSION);
  AppendText(status,buffer);
  
  trap_stderr();

  XFlush(display);

  /* Debug a quirky locale */

  if(getenv("MLDEBUG")) {
    fprintf(stderr,"Locale: %s\n",setlocale(LC_ALL, NULL));
    fflush(stderr);
  }

  /* Check for obsolete or missing app-defaults. */

  if((preferences.local_defs == FALSE) 
     && (preferences.defaults_installed == FALSE))
    mm_log(NO_DEFAULTS,WARN);
  else {
    if((strcmp(preferences.defaults_version,DEFAULTS_VERSION)) != STRMATCH) {
      mm_log(WRONG_DEFAULTS, WARN);
      sprintf(buffer,DEFAULTS_VERSION_MSG,
	      DEFAULTS_VERSION,preferences.defaults_version);
      mm_log(buffer,WARN);
#ifdef FALLBACK
      mm_log(FALLBACK_FOUND1,NIL);
      mm_log(FALLBACK_FOUND2,NIL);
#endif
    }
  }

  if((strcmp(MESSAGE_CONSISTENCY_CHECK,MESSAGE_CONSISTENCY_STR)) != STRMATCH)
    mm_log(MESSAGES_CORRUPTED,ERROR);
  
  /* First time user. Take them right to the preferences screen. */

  if(no_prefs == TRUE)
    main_preferences(NULL, NULL, NULL);

  if(preferences.auto_open == TRUE)
    if((load_mlinit() != SYSCALL_SUCCESS))
      auto_open();


  /*************************************************************************
   *           Main application loop - get and process input forever       *
   *************************************************************************/

  main_loop();
  /* NOTREACHED */
}

#ifdef __STDC__
void start_up(void)
#else
void start_up()
#endif
{
  char buffer[FILEBUFFLEN];
  char *ptr;

  /* find out who is running the program. */

  if((get_user()) == SYSCALL_FAILURE)  {
    fprintf(stderr,"Cannot determine user identity. Program terminated.");
    exit(0);
  }

  (void) chdir(local_auth.homedir);

  mail_link(&imapdriver);
  mail_link(&mboxdriver);
  mail_link(&bezerkdriver);
  mail_link(&tenexdriver);
  mail_link(&nntpdriver);
  mail_link(&newsdriver);
  mail_link(&pop3driver);
  mail_link(&mtxdriver);
  mail_link(&mhdriver);
  mail_link(&mmdfdriver);
  mail_link(&philedriver);
  mail_link(&dummydriver);

  session.address_book = NULL;
  session.filters = NULL;
  session.session_filters = NULL;
  session.mime_handlers = NULL;
  session.busy = FALSE;
  session.button_state = BTN_NOMAILBOXESOPEN|BTN_NOCOMPOSEOPEN|BTN_NOREADOPEN;
  session.authorizations = NULL;
  session.mailboxlist = NULL;
  session.compose = NULL;
  session.view = NULL;
  session.read = NULL;
  session.prefst = NULL;
  session.pref2st = NULL;
  session.gentext = NULL;
  session.server_list = NULL;
  session.update_needed = 0;
  session.filter_map = NULL;
  session.note = NULL;
  session.btn = NULL;
  session.header_set = NULL;
  session.header_str = NULL;
  session.last_command = NULL;
  session.last_dir = NULL;

  session.buttons[VIEW_BTN_MENU] = NULL;
  session.buttons[READ_BTN_MENU] = NULL;
  session.buttons[COMPOSE_BTN_MENU] = NULL;
  session.buttons[MAILBOX_BTN_MENU] = NULL;

  (void) get_resources(top);

  /* Honor the NNTPSERVER variable unless told otherwise. */

  if((ptr = getenv("NNTPSERVER")) != NULL) 
      preferences.nntp_server = cpystr(ptr);

  if((load_defaults()) == SYSCALL_FAILURE)
    no_prefs = TRUE;
  (void) sane_defaults();


  /* Setup the default "custom buttons" */

  view_default_buttons();
  read_default_buttons();
  compose_default_buttons();
  mailboxes_default_buttons();

  /* Load these after the preferences, so we know where to look. */

  (void) load_mime_types();
  (void) load_mime_extension_map();
  (void) load_mime_handlers();
  load_message_catalog();
  init_representation();
  stuff_action_list();
  /* Initialize the c-client's sense of identity. Ignore results. */

  (void) myusername();
  return;
}


#ifdef __STDC__
void create_main_window(void)
#else
void create_main_window()
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;

  Widget form, menubar;
  XtTranslations translations;


  /**********************************************************************
   *                     Create status/telemetry widget                 *
   **********************************************************************/

  XtSetArg(args[n], XmNtitle, PROGRAM);                  n ++;
  form = XmCreateForm(top, "form", args, n);             n = 0;
  
  XtSetArg(args[n], XmNtopAttachment,   XmATTACH_FORM);  n ++;
  XtSetArg(args[n], XmNleftAttachment,  XmATTACH_FORM);  n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM);  n ++;
  menubar = XmCreateMenuBar(form,"menubar", args, n);    n = 0;

  create_buttons(NULL, menubar, main_menu, XtNumber(main_menu), 
		      session.button_state, NULL, ROOTMENULEVEL);

  XtSetArg(args[n], XmNeditable,         TRUE);                  n ++;
  XtSetArg(args[n], XmNeditMode,         XmMULTI_LINE_EDIT);     n ++;
  XtSetArg(args[n], XmNwordWrap,         TRUE);                  n ++;
  XtSetArg(args[n], XmNrows,             STATUSWIN_HEIGHT);      n ++;
  XtSetArg(args[n], XmNcolumns,          STATUSWIN_WIDTH);       n ++;
  XtSetArg(args[n], XmNtopAttachment,    XmATTACH_WIDGET);       n ++;
  XtSetArg(args[n], XmNtopWidget,        menubar);               n ++;
  XtSetArg(args[n], XmNleftAttachment,   XmATTACH_FORM);         n ++;
  XtSetArg(args[n], XmNrightAttachment,  XmATTACH_FORM);         n ++;
  XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM);         n ++;
  XtSetArg(args[n], XmNscrollVertical,   TRUE);                  n ++;
  XtSetArg(args[n], XmNscrollHorizontal, FALSE );                n ++;
  status = XmCreateScrolledText(form, "status", args, n);        n = 0;

  translations = XtParseTranslationTable(GLOBAL_text_translations);
  XtOverrideTranslations(status, translations);
  
  XtManageChild(form);
  XtManageChild(menubar);
  XtManageChild(status);
  XmProcessTraversal(status, XmTRAVERSE_CURRENT);
  return;
}



#ifdef __STDC__
MAILSTREAM *server_to_mailstream(char *server)
#else
MAILSTREAM *server_to_mailstream(server)
     char *server;
#endif
{
  Mailbox *mailbox;
  if((server == NULL) || (*server == NUL_TERM))
    return(NULL);

  for(mailbox = session.mailboxes; mailbox; mailbox = mailbox->next)
    if((strcasecmp(mailbox->host,server)) == STRMATCH)
      return(mailbox->mailstream);
  return(NULL);
}


#ifdef __STDC__
Server_Config ** find_servers(char *service, int *num_servers)
#else
Server_Config ** find_servers(service, num_servers)
     char *service;
     int *num_servers;
#endif
{
  Server_Config ** ret = NULL;
  Server_Config *server_config = NULL;
  int n = 0;

  if(session.server_config != NULL) {
    for(server_config = session.server_config; server_config;
	server_config = server_config->next)
      if((strcmp(server_config->service, service)) == STRMATCH)
	n ++;

    ret = (Server_Config **) fs_get((n + 1) * sizeof(Server_Config *));
    ret[n] = NULL;
    
    n = 0;
    for(server_config = session.server_config; server_config;
	server_config = server_config->next) {
      if((strcmp(server_config->service, service)) == STRMATCH) {
	ret[n] = server_config;
	n ++;
      }
    }
  }
  *num_servers = n;
  return(ret);
}

#ifdef __STDC__
Boolean add_server_to_session(char *str, int type)
#else
Boolean add_server_to_session(str,type)
     char *str;
     int type;
#endif
{
  Server_List *server_list;

  if(str == NULL)
    return(FALSE);

  for(server_list = session.server_list; 
      server_list; server_list = server_list->prev) {
    if((strcasecmp(server_list->name,str) == STRMATCH) 
       && (server_list->type == type))
      return(FALSE);
  }

  server_list = new_server_list();
  server_list->name = cpystr(str);
  server_list->type = type;
  if(session.server_list)
    session.server_list->next = server_list;
  server_list->prev = session.server_list;
  session.server_list = server_list;
  return(TRUE);
}

/* 
 * If an auto-open operation failed or was cancelled, we haven't obtained 
 * any mailbox list info. Here we check for that condition, and if that's 
 * the case, remove the server from session.server_list, so that the next 
 * open will load it correctly. 
 */

#ifdef __STDC__
void invalidate_server_list(Server_List *server_list)
#else
void invalidate_server_list(server_list)
     Server_List *server_list;
#endif
{

  if(server_list == session.server_list)
    session.server_list = server_list->prev;
  if(server_list->prev)
    server_list->prev->next = server_list->next;
  if(server_list->next)
    server_list->next->prev = server_list->prev;


  if(server_list->last_pattern)
    fs_give((void **) &server_list->last_pattern);

  if(server_list->name)
    fs_give((void **) &server_list->name);

  free_mailbox_list(server_list->mailbox_list);
  free_mailbox_list(server_list->mailbox_all_list);

  fs_give((void **) &server_list);
  return;
}

/*
 * load_mlinit should parse a file in the user's directory to open whichever 
 * mailboxes the user wishes to open.
 */
 
#ifdef __STDC__
int load_mlinit(void)
#else
int load_mlinit()
#endif
{
  FILE *fp;
  char *filename;
  char buf[FILEBUFFLEN];
  char *ptr;
  char *ptr2;
  
  filename = (char *) fs_get(strlen(local_auth.homedir)
               + strlen(INITFILE) + strlen(local_auth.hostname) + 8);
               
  sprintf(filename, "%s/%s.%s", local_auth.homedir, INITFILE,
    local_auth.hostname);
    
  if((fp = fopen(filename,"r")) == NULL) {
    sprintf(filename,"%s/%s", local_auth.homedir, INITFILE);
    if((fp = fopen(filename,"r")) == NULL) {
      fs_give((void **) &filename);
      return(SYSCALL_FAILURE);
    }
  }
  
  while((fgets(buf,sizeof(buf),fp)) != NULL) {
    buf[strlen(buf)-1] = NUL_TERM;
    
    if(*buf == '#' || *buf == '!' | *buf == NUL_TERM)
      continue;
    
    ptr = strchr(buf, TABCHAR);
    if(!ptr)
      continue;
    *ptr = NUL_TERM;
    ptr ++;
    ptr2 = strchr(ptr, TABCHAR);
    if(!ptr2)
      continue;
    *ptr2 = NUL_TERM;
    ptr2 ++;
    if(*ptr2 == NUL_TERM)
      continue;
    
    if((strcasecmp(buf,MAILTAG)) == STRMATCH) {
      init_open_mail(ptr, ptr2);
    } 
    else { 
      if((strcasecmp(buf,NEWSTAG)) == STRMATCH) {
	init_open_news(ptr, ptr2);
      }
    }
  }
  fclose(fp);
  fs_give((void **) &filename);
  return(SYSCALL_SUCCESS);
}


#ifdef __STDC__
void init_open_mail(char *server, char *mbox)
#else
void init_open_mail( server, mbox )
  char *server;
  char *mbox;
#endif
{
  Mailbox *mailbox = new_mailbox();
  Boolean new_server = FALSE;

  if((strcmp(server,LOCALTAG)) != STRMATCH)
    mailbox->host = get_qualified_host_name(server);
  mailbox->mailboxname = fix_mailboxpath(mbox);
  mailbox->imapname = 
    format_mailbox_name(mailbox->host,mailbox->mailboxname,MAILBOX_TYPE_MAIL);
  mailbox->type = MAILBOX_TYPE_MAIL;
  
  if((mailbox->imapname == NULL) || (*mailbox->imapname == NUL_TERM)) {
    free_mailbox(mailbox);
    return;
  }
 
  if(add_server_to_session(mailbox->host, MAILBOX_TYPE_MAIL))
    new_server = TRUE;
  
  open_mailbox(mailbox, new_server);
  return;
}
 
 
#ifdef __STDC__
void init_open_news(char *server, char *mbox)
#else
void init_open_news(server, mbox)
   char *server;
   char *mbox;
#endif
{
  Mailbox *mailbox = new_mailbox();
  Boolean new_server = FALSE;

  mailbox->host = get_qualified_host_name(server);
  mailbox->mailboxname = fix_mailboxpath(mbox);
  mailbox->imapname = 
    format_mailbox_name(mailbox->host,mailbox->mailboxname,MAILBOX_TYPE_NEWS);
  mailbox->type = MAILBOX_TYPE_NEWS;
  
  if((mailbox->imapname == NULL) || (*mailbox->imapname == NUL_TERM)) {
    free_mailbox(mailbox);
    return;
  }
  
  if(add_server_to_session(mailbox->host, MAILBOX_TYPE_NEWS))
    new_server = TRUE;
  
  open_mailbox(mailbox, new_server);
  return;
}
 
#ifdef __STDC__
void save_mlinit( void )
#else
void save_mlinit()
#endif
{
  FILE *fp;
  Mailbox *mailbox, *next, *prev;
  char *filename, *backup;
  int mbox[100];
  int i;
  
  filename = (char *) fs_get(strlen(local_auth.homedir) + strlen(INITFILE)
			     + strlen(local_auth.hostname) + 8);
  
  sprintf(filename, "%s/%s.%s", local_auth.homedir,
	  INITFILE, local_auth.hostname);
  
  if(access(filename,W_OK)) {
    sprintf(filename, "%s/%s", local_auth.homedir, INITFILE);
  }
  
  backup = (char *) fs_get(strlen(filename) + strlen(".bak") + 8);
  strcpy(backup,filename);
  strcat(backup,".bak");
  
  rename(filename,backup);
  
  fs_give((void **) &backup);
  
  if((fp = fopen(filename,"w")) == NULL) {
    mm_log(INITSAVE_ERROR,WARN);
    fs_give((void **) &filename);
    return;
  }
  
  (void) chmod(filename,S_IRWXU);
  
  fprintf(fp,MLINITFILEHEADER);  

  for(mailbox = session.mailboxes; mailbox && mailbox->next; 
      mailbox = mailbox->next)
    ;

  if(mailbox) {
    for(mailbox = mailbox; mailbox; mailbox = mailbox->prev) 
      fprintf(fp,"%s\t%s\t%s\n",
	      (mailbox->type == MAILBOX_TYPE_MAIL) ? MAILTAG : NEWSTAG,
	      (mailbox->host != NULL && *mailbox->host != NUL_TERM) 
	      ? mailbox->host : LOCALTAG,
	      mailbox->mailboxname);
  }
  if((fclose(fp)) == SYSCALL_SUCCESS)
    mm_log(INITSAVE_OK, NIL);
  else
    mm_log(INITSAVE_ERROR, WARN);
  fs_give((void **) &filename);
  return;
}



#ifdef __STDC__
void main_open(Widget w, XtPointer cb, XtPointer xp)
#else
void main_open(w,cb,xp)
     Widget w;
     XtPointer cb;
     XtPointer xp;
#endif
{
  mailboxes(FALSE);
  return;
}


/* 
 * auto_open has to initialize a few things in the mailbox structure
 * which would have otherwise been initialized by the mailbox list widget.
 */


#ifdef __STDC__
void auto_open(void)
#else
void auto_open()
#endif
{
  Mailbox *mailbox = new_mailbox();
  Boolean new_server = FALSE;

  mailbox->host = get_qualified_host_name(preferences.imap_server);
  mailbox->mailboxname = fix_mailboxpath(preferences.default_mailbox);
  mailbox->imapname = 
    format_mailbox_name(mailbox->host,mailbox->mailboxname,MAILBOX_TYPE_MAIL);
  mailbox->type = MAILBOX_TYPE_MAIL;

  if((mailbox->imapname == NULL) || (*mailbox->imapname == NUL_TERM)) {
    free_mailbox(mailbox);
    return;
  }

  if(add_server_to_session(mailbox->host, MAILBOX_TYPE_MAIL))
    new_server = TRUE;

  open_mailbox(mailbox, new_server);
  return;
}


#ifdef __STDC__
void open_mailbox(Mailbox *mailbox, Boolean new_server)
#else
void open_mailbox(mailbox, new_server)
     Mailbox *mailbox;
     Boolean new_server;
#endif
{
  char log[FILEBUFFLEN];
  Server_List *server_list;
  long openflags = 0L;

  push_cursor(WATCH_CURSOR);

  if((mailbox->imapname == NULL) || (*mailbox->imapname == NUL_TERM)) {
    free_mailbox(mailbox);
    pop_cursor();
    return;
  }

  /* 
   * Just in case the mailbox is already open, we'll be nice and close it.
   */

  close_invalid_mailbox(mailbox->imapname);

  sprintf(log,OPENING_MAILBOX,mailbox->imapname);
  mm_log(log,NIL);
  XmUpdateDisplay(top);

  if(((mailbox->mailstream = 
      mail_open(mailbox->mailstream,mailbox->imapname,openflags)) == NIL) 
     || (mailbox->mailstream->halfopen)) {
    server_list = find_server_list(mailbox->host,mailbox->type);
    invalidate_server_list(server_list);
    if(mailbox->mailstream)
      mail_close(mailbox->mailstream);
    free_mailbox(mailbox);
    mm_log(OPEN_FAILED,ERROR);
    pop_cursor();
    return;
  }
  else 
    session.button_state &= (~BTN_NOMAILBOXESOPEN);

  if(new_server) {
    server_list = find_server_list(mailbox->host,mailbox->type);
    server_list->active = TRUE;
    if(server_list->last_pattern)
      fs_give((void **) &server_list->last_pattern);
    server_list->last_pattern = cpystr(MAILBOXALLPATTERN);
    mail_find(mailbox->mailstream,server_list->last_pattern);
    server_list->active = FALSE;
  }

  mailbox->nmsgs = mailbox->mailstream->nmsgs;
  if(session.mailboxes)
    session.mailboxes->prev = mailbox;
  mailbox->next = session.mailboxes;
  session.mailboxes = mailbox;

  mailbox->timer = 
    XtAppAddTimeOut(context,
		    ((unsigned long) preferences.check_interval * 1000L),
		    (XtTimerCallbackProc) timed_check, mailbox);

  create_all_lview(mailbox);

  check_buttons(main_menu, XtNumber(main_menu), NULL, session.button_state);
  pop_cursor();
  return;
}

#ifdef __STDC__
void check_mailbox(Mailbox *mailbox)
#else
void check_mailbox(mailbox)
     Mailbox *mailbox;
#endif
{
  char msg[FILEBUFFLEN];

  if(session.busy == TRUE)
    return;

  push_cursor(WATCH_CURSOR);

  if(mail_ping(mailbox->mailstream)) {
    if(mailbox->has_new_mail == TRUE) {
      get_new_message_info(mailbox);
      mailbox->has_new_mail = FALSE;
      XBell(display,1000);
    }
    if(mailbox->flags_changed != 0L) {
      mail_check(mailbox->mailstream);
      mailbox->flags_changed = 0L;
    }
  }
  else {
    sprintf(msg,LOST_CONNECTION,mailbox->host); 
    mm_log(msg,ERROR);
  }
  pop_cursor();
  return;
}

#ifdef __STDC__
void timed_check(Mailbox *mailbox, XtIntervalId *id)
#else
void timed_check(mailbox,id)
     Mailbox *mailbox;
     XtIntervalId *id;
#endif
{

  unsigned long next_timer;
 
  /* 
   * We've been called by a timer (or user) to check the mailbox.
   * Unfortunately, we might be doing something critical.
   * If we are, set the timer interval to a short period (15 seconds)
   * so we'll keep coming back until things are free again.
   * Otherwise, check the mailbox, and reset to the normal timer interval.
   */

  if(session.busy == FALSE) { 
    check_mailbox(mailbox);
    next_timer = (unsigned long) preferences.check_interval * 1000L;
  }
  else
    next_timer = (unsigned long) 15000L;

  mailbox->timer = XtAppAddTimeOut(context,next_timer,
				   (XtTimerCallbackProc) timed_check,
				   mailbox);
  return;
}


/* Empty the status window. Clean, quick, painless. */

#ifdef __STDC__
void main_prune_log(Widget w,XtPointer data,XtPointer xp)
#else
void main_prune_log(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  XmTextSetString(status,EMPTYSTR);
  return;
}



#ifdef __STDC__
void main_close(Widget w,XtPointer data,XtPointer xp)
#else
void main_close(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  Mailbox *mailbox, *next;
  mm_log(CLOSINGEVERYTHING, NIL);

  for(mailbox = session.mailboxes; mailbox ; mailbox = next) {
    next = mailbox->next;
    close_mailbox(mailbox);
  }
  session.mailboxes = NULL;
  return;
}


#ifdef __STDC__
void close_mailbox(Mailbox *mailbox)
#else
void close_mailbox(mailbox)
     Mailbox *mailbox;
#endif
{
  Lview *lview;
  Read_Info *read_info;
  Boolean found = FALSE;
  char log[FILEBUFFLEN];
  char sequence[64];

  sprintf(log,CLOSING_MAILBOX_MSG,mailbox->imapname);
  mm_log(log,NIL);

  XtRemoveTimeOut(mailbox->timer);
  mailbox->timer = (XtIntervalId) 0;
  if(mailbox->bg_timer)
    XtRemoveTimeOut(mailbox->bg_timer);
  mailbox->bg_timer = (XtIntervalId) 0;

  if(mail_ping(mailbox->mailstream)) {
    if((mailbox->type == MAILBOX_TYPE_NEWS) 
       && (mailbox->nmsgs > 0L)
       && (mailbox->no_update == FALSE)) {
      if(mailbox->nmsgs == 1L)
	strcpy(sequence,"1");
      else
	sprintf(sequence,"1:%lu",(unsigned long) mailbox->nmsgs);
      mm_log(UPDATING_NEWSRC,NIL);
      mail_setflag(mailbox->mailstream,sequence,DELETED_FLAG);
      XmUpdateDisplay(top);
    }
    mail_check(mailbox->mailstream);
    mail_close(mailbox->mailstream);
  }
  
  if(session.view) {
    do {
      for(lview = session.view->lview; lview ; lview = lview->prev)
	if(lview->mailstream == mailbox->mailstream) {
	  found = TRUE;
	  view_unlink_destroy_lview(session.view,lview,TRUE);
	}
	else
	  found = FALSE;
    } while((lview) && (found == TRUE));
  }
  
  if(session.read) {
    do {
      found = FALSE;
      for(read_info = session.read->read_info; read_info; 
	  read_info = read_info->prev) {
	if(read_info->mailstream == mailbox->mailstream) {
	  if(read_info == session.read->current)
	    read_close_kill_current(session.read->shell,session.read,NULL);
	  else
	    kill_read_info(read_info);
	  found = TRUE;
	}
      }
    } while((read_info) && (found == TRUE));
    
    stuff_read_window_list(session.read, TRUE);
    if(session.read->is_realized)
      XmUpdateDisplay(session.read->shell);
  }

  compose_mailbox_close(mailbox->mailstream);

  if(mailbox == session.mailboxes)
    session.mailboxes = mailbox->next;
  
  if(mailbox->prev)
    mailbox->prev->next = mailbox->next;
  if(mailbox->next)
    mailbox->next->prev = mailbox->prev;
  
  free_mailbox(mailbox);
  
  if(session.mailboxes == NULL) {
    if(session.view && session.view->list_is_realized)
      view_lview_dismiss(NULL, session.view, NULL);
    if(session.edfilter && session.edfilter->is_realized)
      edfilter_dismiss(NULL, session.edfilter, NULL);
    if(session.read && session.read->list_is_realized)
      read_list_cancel(NULL, session.read, NULL);
    session.button_state |= (BTN_NOMAILBOXESOPEN);
  }  
  update_main_buttons();
  XFlush(display);
  return;
}



#ifdef __STDC__
void main_compose(Widget w, XtPointer data, XtPointer xp)
#else
void main_compose(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  (void) compose(NEW_MESSAGE_STR, MAIL_BEHAVIOUR, COMPOSE_NEW, 
		 NULL, NULL, NULL);
  return;
}

#ifdef __STDC__
void main_address_book(Widget w, XtPointer data, XtPointer xp)
#else
void main_address_book(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  (void) create_address_book_window(top);
  return;
}


#ifdef __STDC__
void main_note_book(Widget w, XtPointer data, XtPointer xp)
#else
void main_note_book(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  (void) create_note_window();
  return;
}

#ifdef __STDC__
void main_raise_view(Widget w, XtPointer data, XtPointer xp)
#else
void main_raise_view(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  if(session.view && session.view->lview) {
    create_view_window();
    view_context_switch(session.view,NULL,session.view->lview);
    de_iconify(session.view->shell);
  }
  return;
}


#ifdef __STDC__
void main_raise_read(Widget w, XtPointer data, XtPointer xp)
#else
void main_raise_read(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  if(session.read && session.read->read_info) {
    create_read_window();
    read_context_switch(session.read,NULL,session.read->read_info);
  }
  return;
}

#ifdef __STDC__
void main_raise_compose(Widget w, XtPointer data, XtPointer xp)
#else
void main_raise_compose(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  if(session.compose && session.compose->compose_info) {
    if(session.compose->is_realized != TRUE) 
      create_compose_window(session.compose,MAIL_BEHAVIOUR, CC_BEHAVIOUR);
    de_iconify(session.compose->shell);
    compose_context_switch(session.compose,NULL,session.compose->compose_info);
  }
  return;
}

#ifdef __STDC__
void no_op(Widget w, XtPointer data, XtPointer xp)
#else
void no_op(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  /* globally accessible absolutely useless function */
  return;
}


#ifdef __STDC__
void main_buttons(Widget w, XtPointer data, XtPointer xp)
#else
void main_buttons(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  create_btn_window();
  return;
}



#ifdef __STDC__
void main_help(Widget w, XtPointer data, XtPointer xp)
#else
void main_help(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  help(top, MAINHELPTITLE, MAINHELPFILE, DOCTYPE_ENRICHED);
  return;
}


#ifdef __STDC__
void main_keyboard_help(Widget w, XtPointer data, XtPointer xp)
#else
void main_keyboard_help(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  help(top, KEYBOARDHELPTITLE, KEYBOARDHELPFILE, DOCTYPE_ENRICHED);
  return;
}

#ifdef __STDC__
void main_config_help(Widget w, XtPointer data, XtPointer xp)
#else
void main_config_help(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  help(top, MIMEINFOHELPTITLE, MIMEINFOHELPFILE, DOCTYPE_ENRICHED);
  return;
}

#ifdef __STDC__
void main_resources_help(Widget w, XtPointer data, XtPointer xp)
#else
void main_resources_help(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  help(top, RESOURCESHELPTITLE, RESOURCESHELPFILE, DOCTYPE_PLAIN);
  return;
}

#ifdef __STDC__
void main_intl_help(Widget w, XtPointer data, XtPointer xp)
#else
void main_intl_help(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  help(top, INTLHELPTITLE, INTLHELPFILE, DOCTYPE_ENRICHED);
  return;
}


#ifdef __STDC__
void main_version_help(Widget w, XtPointer data, XtPointer xp)
#else
void main_version_help(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  help(top, VERSIONHELPTITLE, VERSIONHELPFILE, DOCTYPE_ENRICHED);
  return;
}

/* 
 * End the program. Unless you kill it, or it (shudder) dumps core,
 * this is the only way out. 
 */

#ifdef __STDC__
void bye(Widget w, XtPointer data, XtPointer xp)
#else
void bye(w,data,xp)
     Widget w;
     XtPointer data;
     XtPointer xp;
#endif
{
  if(session.mailboxes)
    main_close(w,NULL,NULL);
  exit(0);
}

/*
 * This is called out of our main loop when something changed during
 * a callback. refresh_needed means that the current contents of
 * the view window changed; we have to do something because the 
 * change is visible. update_needed means we have to re-evaluate
 * our logical view filters. They may or may not be changed as a 
 * result.
 */  


#ifdef __STDC__
void update_views(void)
#else
void update_views()
#endif
{
  Mailbox *mailbox;
  for(mailbox = session.mailboxes; mailbox; mailbox = mailbox->next) {
    if(mailbox->refresh_needed) {
      update_current_view(mailbox);
      mailbox->refresh_needed = 0;
    }
    if(mailbox->update_needed) {
      update_logical_views(mailbox,FALSE);
      mailbox->update_needed = 0;
    }
  }
  return;
}


/* 
 * Grab a chunk of messages envelopes in the "background".
 * We'll keep doing this at a settable interval until the mailbox
 * is completely loaded, or a newsgroup has been pruned.
 */


#ifdef __STDC__
void background_fetch(Mailbox *mailbox, XtIntervalId *id)
#else
void background_fetch(mailbox,id)
     Mailbox *mailbox;
     XtIntervalId *id;
#endif
{
  Boolean result;
  unsigned long count;

  if(session.busy == FALSE) { 
    if(mailbox->first_fetch <= 1L) {    /* no more to fetch */
      update_logical_views(mailbox,TRUE);
      mailbox->update_needed = 0;
      XtRemoveTimeOut(mailbox->bg_timer);
      mailbox->bg_timer = (XtIntervalId) 0;
      mm_log(BACKGROUND_DONE_MSG,MLNOTIFY);
      return;
    }
    else {                             /* more to fetch */
      push_cursor(WATCH_CURSOR);
      if(mailbox->type == MAILBOX_TYPE_MAIL) {
	if(mailbox->first_fetch <= preferences.mail_fetch)
	  mailbox->first_fetch = 1L;
	else
	  mailbox->first_fetch -= preferences.mail_fetch;
      }
      else {  /* MAILBOX_TYPE_NEWS */
	if(mailbox->first_fetch <= preferences.news_fetch)
	  mailbox->first_fetch = 1L;
	else
	  mailbox->first_fetch -= preferences.news_fetch;
      }
      count = mailbox->first_fetch;
      do {
	result = background_fetch_message(mailbox,count, FALSE);
	count ++;
      } while(result == TRUE);

      update_current_view(mailbox);
      mailbox->refresh_needed = 0;
      if(mailbox->first_fetch <= 1L) {   /* We've just gotten the last one */
	update_logical_views(mailbox,TRUE);
	mailbox->update_needed = 0;
	XtRemoveTimeOut(mailbox->bg_timer);
	mailbox->bg_timer = (XtIntervalId) 0;
	mm_log(BACKGROUND_DONE_MSG,MLNOTIFY);
	pop_cursor();
	return;
      }
      pop_cursor();
    }
  }

  /* Adios for now, but we'll be back. */

  mailbox->bg_timer = XtAppAddTimeOut(context,
				      (mailbox->type == MAILBOX_TYPE_NEWS) 
				      ? (preferences.news_background * 1000)
				      : (preferences.mail_background * 1000),
				      (XtTimerCallbackProc) background_fetch,
				      mailbox);
  return;
}


/* 
 * This loop is called from our modal pop-ups -- those that wait
 * for you to do something before continuing. The loop is exited
 * when the condition (usually an int) becomes non-zero. This usually
 * happens when a modal widget gets what it needs and relinquishes
 * exclusive control of the keyboard and mouse by killing itself.
 * In general we set "pirate" cursors for other windows to indicate
 * that they are non-functional until the active window is dismissed.
 * Yes, it's awful design strategy, but necessary in this program,
 * since something you are doing elsewhere (like expunging messages 
 * or closing mailboxes) could invalidate the entire state of the 
 * current data structures.
 */

#ifdef __STDC__
void modal_main_loop (int *exit_condition)
#else
void modal_main_loop (exit_condition)
     int *exit_condition;
#endif
{
  XEvent event;

  while(!(*exit_condition)) {
    if(XPending(display)) {
      XtAppNextEvent(context, &event);
      XtDispatchEvent(&event);
    }
  }
  return;
}


/*
 * This is the normal event processing loop. We trigger a logical view update
 * through external means, so that functions which are making several changes
 * don't have to flash the screen for every change. When the functions exit,
 * we'll fall back into this loop, and the update will be triggered. This
 * way it's only called once.
 */

#ifdef __STDC__
void main_loop(void)
#else
void main_loop ()
#endif
{
  XEvent event;

  while(1) {
    XtAppNextEvent(context, &event);
    XtDispatchEvent(&event);
    if(session.update_needed) {
      update_views();
      session.update_needed = 0;
    }
  }
  /* NOTREACHED */
}

/* 
 * A function to make our main menu buttons tweakable from other files,
 * since they don't know the size of our main_menu. 
 */

#ifdef __STDC__
void update_main_buttons(void)
#else
void update_main_buttons()
#endif
{
  check_buttons(main_menu, XtNumber(main_menu), NULL, session.button_state);
  return;
}

/*
 * Functions to capture the stderr stream from sub-processes
 * (or X window messages) and send them to the ML status (main) window.
 */

#ifdef __STDC__
void redirect_stderr(XtPointer client_data, int *fd, XtInputId *id)
#else
void redirect_stderr(client_data, fd, id)
     XtPointer client_data;
     int *fd;
     XtInputId *id;
#endif
{
  char buf[FILEBUFFLEN];
  int nbytes;

  if((nbytes = read(*fd,buf,FILEBUFFLEN - 1)) == (EOF))
    return;
  buf[nbytes] = NUL_TERM;
  if(nbytes) {
    de_iconify(top);
    AppendText(status,buf);
  }
  return;
}


#ifdef __STDC__
void trap_stderr(void)
#else
void trap_stderr()
#endif
{
  (void) signal(SIGPIPE,SIG_IGN);

  if(pipe(errfd))
    return;

  dup2(errfd[1], fileno(stderr));
  close(errfd[1]);

  XtAppAddInput(context, errfd[0], (caddr_t) XtInputReadMask, 
		(XtInputCallbackProc) redirect_stderr, NULL);

  return;
}


#ifdef __STDC__
void setup_editres(Widget shell)
#else
void setup_editres(shell)
     Widget shell;
#endif
{
#ifndef NO_EDITRES
  XtAddEventHandler(shell, (EventMask) 0, 
		    True, _XEditResCheckMessages, NULL );
#endif
}


/*
 * Pops the window to the front. De-iconifies it if neccessary first.
 */

#ifdef __STDC__
void de_iconify(Widget shell)
#else
void de_iconify(shell)
     Widget shell;
#endif
{
  XMapWindow(display,XtWindow(shell));
  XRaiseWindow(display,XtWindow(shell));
}

