/* notefncs.c
 *
 *
 * $Author: mtm $
 * $Date: 1995/05/09 01:27:15 $
 * $Id: notefncs.c,v 1.1 1995/05/09 01:27:15 mtm Exp mtm $
 *
 * Mike Macgirvin <Mike_Macgirvin@CAMIS.Stanford.EDU>
 *
 * Copyright 1995 by The Leland Stanford Junior University.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notices appear in all copies and that both the
 * above copyright notices and this permission notice appear in supporting
 * documentation, and that the name of The Leland Stanford Junior University 
 * not be used in advertising or publicity pertaining to distribution of the 
 * software without specific, written prior permission.  This software is made 
 * available "as is", and THE LELAND STANFORD JUNIOR UNIVERSITY DISCLAIMS ALL 
 * WARRANTIES, EXPRESS OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING 
 * WITHOUT LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE LELAND STANFORD JUNIOR 
 * UNIVERSITY BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR 
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, 
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
 * SOFTWARE.
 *
 */

/*
 * $Log: notefncs.c,v $
 * Revision 1.1  1995/05/09  01:27:15  mtm
 * Initial revision
 *
 * 
 */

#include "ml.h"

Menu note_file_menu[] = {
  { NULL, "save", NUL_TERM,
      note_save, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "close_window", NUL_TERM,
      note_close, NULL, 0, NULL, NULL, BTN_ON },
};

Menu note_menu[] = {
  { NULL, "file_menu", NUL_TERM,
      NULL, note_file_menu, XtNumber(note_file_menu), 
      NULL, NULL, BTN_ON },
  { NULL, "HELP", NUL_TERM,
      note_help, NULL, 0, 
      NULL, NULL, BTN_ON },
};



#ifdef __STDC__
void create_note_window(void)
#else
void create_note_window()
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  Widget form, menubar;
  XtTranslations translations;
  Note *note;

  if(session.note == NULL) {
    note = new_note();
    session.note = note;
  }
  else
    note = session.note;

  note->autosave = (XtIntervalId) 0;

  if(note->is_realized == TRUE) {
    de_iconify(note->shell);
    return;
  }

  push_cursor(WATCH_CURSOR);

  note->shell = XtAppCreateShell("MLnotebook", "MLnotebook",
			       topLevelShellWidgetClass,
			       display, args, n ); 
  n = 0;

  setup_editres(note->shell);

  if(note_icon != (Pixmap) None)
    XtVaSetValues(note->shell,
		  XmNiconPixmap,note_icon,
		  NULL);

  form = XmCreateForm(note->shell, "form", args, n);            n = 0;
  XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM);           n ++;
  XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM);          n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM);         n ++;
  menubar = XmCreateMenuBar(form,"menubar", args, n);           n = 0;
  XtManageChild(menubar);
  create_buttons(NULL, menubar, note_menu, XtNumber(note_menu), 
		      BTN_ON, (XtPointer) note, ROOTMENULEVEL);

  XtSetArg(args[n], XmNeditable, TRUE);                         n ++;
  XtSetArg(args[n], XmNeditMode, XmMULTI_LINE_EDIT);            n ++;
  XtSetArg(args[n], XmNwordWrap, TRUE);                         n ++;
  XtSetArg(args[n], XmNcolumns,  COMPOSEWIDTH - 1);             n ++;
  XtSetArg(args[n], XmNscrollVertical, TRUE);                   n ++; 
  XtSetArg(args[n], XmNscrollHorizontal, FALSE );               n ++;

  note->text =
    XmCreateScrolledText(form, "text", args, n);        n = 0;
  
  translations = XtParseTranslationTable(GLOBAL_text_translations);
  XtOverrideTranslations(note->text, translations);

  XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET);         n ++;
  XtSetArg(args[n], XmNtopWidget, menubar);                     n ++;
  XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM);        n ++;
  XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM);          n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM);         n ++;
  XtSetValues(XtParent(note->text), args, n);        n = 0;
  XtManageChild(note->text);

  XtAddCallback(note->shell, XmNdestroyCallback, 
		(XtCallbackProc) note_destroy, note);

  if(preferences.note_save)
    note->autosave =  XtAppAddTimeOut(context,preferences.note_save * 1000L,
				      (XtTimerCallbackProc) note_autosave,
				      note);

  XtManageChild(form);
  XtRealizeWidget(note->shell);
  note->is_realized = TRUE;
  note_load(note);
  pop_cursor();
  return;
}



#ifdef __STDC__
void note_load(Note *note)
#else
void note_load(note)
     Note *note;
#endif
{
  char filename[MAXPATHLEN];
  Binary_Buffer *binary_buffer = NULL;

  sprintf(filename,"%s/%s",local_auth.homedir,NOTEFILE);

  binary_buffer = load_binary_file(filename);
  if(binary_buffer) {
    text_blast(note->text,(char *) binary_buffer->data);
    free_binary_buffer(binary_buffer);
  }
  return;
}

#ifdef __STDC__
void note_autosave(Note *note, XtIntervalId *id)
#else
void note_autosave(note,id)
     Note *note;
     XtIntervalId *id;
#endif
{
  if(note->is_realized == TRUE) {
    if(session.busy == FALSE)
      note_save(NULL, note, NULL);
    if(preferences.note_save) 
      note->autosave =  XtAppAddTimeOut(context,preferences.note_save * 1000L,
					(XtTimerCallbackProc) note_autosave,
					note);
    else
      note->autosave = (XtIntervalId) 0;
  }
  return;
}


#ifdef __STDC__
void note_save(Widget w, Note *note, XtPointer xp)
#else
void note_save(w,note,xp)
     Widget w;
     Note *note;
     XtPointer xp;
#endif
{
  char filename[MAXPATHLEN];
  char buffer[FILEBUFFLEN];
  FILE *fp;
  char *str;
  int errors = 0;

  push_cursor(WATCH_CURSOR);

  str = XmTextGetString(note->text);

  if((str == NULL) || (*str == NUL_TERM))
    if(str != NULL)
      fs_give((void **) &str);
    
  if(str) {
    sprintf(filename,"%s/%s",local_auth.homedir,NOTEFILE);
    if((fp = fopen(filename,"w")) == NULL) {
      sprintf(buffer,CANNOT_WRITE_FILE, filename);
      mm_log(buffer,WARN);
      pop_cursor();
      return;
    }
    (void) chmod(filename,S_IRWXU);

    if((fwrite(str,strlen(str),1,fp)) != 1) 
      errors ++;
    fs_give((void **) &str);
  
    if(fclose(fp))
      errors ++;

    if(errors)
      mm_log(NOTESAVEFAIL,WARN);
    else
      mm_log(NOTESAVESUCCESS, NIL);
  }

  pop_cursor();
  return;
}

#ifdef __STDC__
void note_help(Widget w, Note *note, XtPointer xp)
#else
void note_help(w,note,xp)
     Widget w;
     Note *note;
     XtPointer xp;
#endif
{
  help(note->shell, NOTEHELPTITLE, NOTEHELPFILE, DOCTYPE_ENRICHED);
  return;
}

#ifdef __STDC__
void note_close(Widget w, Note *note, XtPointer xp)
#else
void note_close(w,note,xp)
     Widget w;
     Note *note;
     XtPointer xp;
#endif
{
  note_save(w,note,xp);
  if(note->autosave)
    XtRemoveTimeOut(note->autosave);
  note->autosave = (XtIntervalId) 0;
  XtRemoveCallback(note->shell, XmNdestroyCallback, 
		(XtCallbackProc) note_destroy, note);
  XtDestroyWidget(note->shell);
  note->is_realized = FALSE;
  return;
}

#ifdef __STDC__
void note_destroy(Widget w, Note *note, XtPointer xp)
#else
void note_destroy(w,note,xp)
     Widget w;
     Note *note;
     XtPointer xp;
#endif
{
  if(note->autosave)
    XtRemoveTimeOut(note->autosave);
  note->autosave = (XtIntervalId) 0;
  note->is_realized = FALSE;
  return;
}

