/* view.c
 *
 *
 * $Author: mtm $
 * $Date: 1995/05/09 01:27:15 $
 * $Id: view.c,v 1.1 1995/05/09 01:27:15 mtm Exp mtm $
 *
 * Mike Macgirvin <Mike_Macgirvin@CAMIS.Stanford.EDU>
 *
 * Copyright 1995 by The Leland Stanford Junior University.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notices appear in all copies and that both the
 * above copyright notices and this permission notice appear in supporting
 * documentation, and that the name of The Leland Stanford Junior University 
 * not be used in advertising or publicity pertaining to distribution of the 
 * software without specific, written prior permission.  This software is made 
 * available "as is", and THE LELAND STANFORD JUNIOR UNIVERSITY DISCLAIMS ALL 
 * WARRANTIES, EXPRESS OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING 
 * WITHOUT LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE LELAND STANFORD JUNIOR 
 * UNIVERSITY BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR 
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, 
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
 * SOFTWARE.
 *
 */

/*
 * $Log: view.c,v $
 * Revision 1.1  1995/05/09  01:27:15  mtm
 * Initial revision
 *
 * 
 */

#include "ml.h"

/* 
 *  These MUST correspond with the Representation_Type 
 *  enum in structures.h 
 */
 
char *REPRESENTATION[4];

#ifdef __STDC__
void init_representation(void)
#else
void init_representation()
#endif
{ 
  REPRESENTATION[0] = NORMAL_REPSTR;
  REPRESENTATION[1] = REVERS_REPSTR;
  REPRESENTATION[2] = SENDER_REPSTR;
  REPRESENTATION[3] = SUBJEC_REPSTR;
  return;
}


char *SYS_SETTABLE_FLAGS[] = {
  DELETED_FLAG,
  SEEN_FLAG,
  FLAGGED_FLAG,
  ANSWERED_FLAG,
  NULL,
};
  

Menu keyword_menu[] = {
  { NULL, "keyword_accept", NUL_TERM,
      keyword_accept, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "keyword_cancel", NUL_TERM,
      keyword_cancel, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "keyword_HELP", NUL_TERM,
      keyword_help, NULL, 0, NULL, NULL, BTN_ON },
};

Menu view_lview_file_menu[] = {
  { NULL, "detach", NUL_TERM,
      view_lview_detach, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "save", NUL_TERM,
      view_lview_save, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "close_window", NUL_TERM,
      view_lview_dismiss, NULL, 0, NULL, NULL, BTN_ON },
};

Menu view_lview_option_menu[] = {
  { NULL, "hide", NUL_TERM,
      view_lview_hide, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "show", NUL_TERM,
      view_lview_show, NULL, 0, NULL, NULL, BTN_ON },
};


Menu view_lview_menu[] = {
  { NULL, "file_menu", NUL_TERM,
      NULL, view_lview_file_menu, XtNumber(view_lview_file_menu),
      NULL, NULL, BTN_ON },
  { NULL, "option_menu", NUL_TERM,
      NULL, view_lview_option_menu, XtNumber(view_lview_option_menu),
      NULL, NULL, BTN_ON },
  { NULL, "HELP", NUL_TERM,
      view_lview_help, NULL, 0, NULL, NULL, BTN_ON },
};

Menu view_save_menu[] = {
  { NULL, "fullhead", NUL_TERM,
      view_savefull, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "parthead", NUL_TERM,
      view_savepart, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "nonehead", NUL_TERM,
      view_savenone, NULL, 0, NULL, NULL, BTN_NOSELECTION },
};


Menu view_pipe_menu[] = {
  { NULL, "fullhead", NUL_TERM,
      view_execfull, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "parthead", NUL_TERM,
      view_execpart, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "nonehead", NUL_TERM,
      view_execnone, NULL, 0, NULL, NULL, BTN_NOSELECTION },
};

Menu view_print_menu[] = {
  { NULL, "fullhead", NUL_TERM,
      view_printfull, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "parthead", NUL_TERM,
      view_printpart, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "nonehead", NUL_TERM,
      view_printnone, NULL, 0, NULL, NULL, BTN_NOSELECTION },
};

Menu view_file_menu[] = {
  { NULL, "read_selected", NUL_TERM,
      view_read_selected, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "copy", NUL_TERM,
      view_copy, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "move", NUL_TERM,
      view_move, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "print_menu", NUL_TERM,
      NULL, view_print_menu, XtNumber(view_print_menu), 
      NULL, NULL, BTN_NOSELECTION },
  { NULL, "save_menu", NUL_TERM,
      NULL, view_save_menu, XtNumber(view_save_menu), 
      NULL, NULL, BTN_NOSELECTION },
  { NULL, "pipe_menu", NUL_TERM,
      NULL, view_pipe_menu, XtNumber(view_pipe_menu), 
      NULL, NULL, BTN_NOSELECTION },
  { NULL, "address_book", NUL_TERM,
      main_address_book, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "note_book", NUL_TERM,
      main_note_book, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "close_window", NUL_TERM,
      view_close_window, NULL, 0, NULL, NULL, BTN_ON },
};

Menu view_setflags_menu[] = {
  { NULL, "set_deleted", NUL_TERM,
      view_set_deleted, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "set_seen", NUL_TERM,
      view_set_seen, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "set_flagged", NUL_TERM,
      view_set_flagged, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "set_answered", NUL_TERM,
      view_set_answered, NULL, 0, NULL, NULL, BTN_NOSELECTION },
};

Menu view_clearflags_menu[] = {
  { NULL, "clear_deleted", NUL_TERM,
      view_clear_deleted, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "clear_seen", NUL_TERM,
      view_clear_seen, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "clear_flagged", NUL_TERM,
      view_clear_flagged, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "clear_answered", NUL_TERM,
      view_clear_answered, NULL, 0, NULL, NULL, BTN_NOSELECTION },
};

Menu view_edit_menu[] = {
  { NULL, "select_messages", NUL_TERM,
      view_select_messages, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "select_new_messages", NUL_TERM,
      view_select_new_messages, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "select_unseen_messages", NUL_TERM,
      view_select_unseen_messages, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "unselect_messages", NUL_TERM, 
      view_unselect_messages, NULL, 0, NULL, NULL, BTN_NOSELECTION },
  { NULL, "find_messages", NUL_TERM, 
      view_find, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "setflags_menu",  NUL_TERM,
      NULL, view_setflags_menu, XtNumber(view_setflags_menu), 
      NULL, NULL, BTN_NOSELECTION },
  { NULL, "clearflags_menu",  NUL_TERM,
      NULL, view_clearflags_menu, XtNumber(view_clearflags_menu),
      NULL, NULL, BTN_NOSELECTION },
  { NULL, "set_keyword",  NUL_TERM,
      view_set_keyword, NULL, 0, NULL, NULL, 
      BTN_NOKEYWORDS|BTN_NOSELECTION },
  { NULL, "unset_keyword", NUL_TERM, 
      view_unset_keyword, NULL, 0, 
      NULL, NULL, BTN_NOKEYWORDS|BTN_NOSELECTION },
};

Menu view_options_menu[] = {
  { NULL, "filter", NUL_TERM, 
      view_filter, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "sort_normal",  NUL_TERM,
      view_sort_normal, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "sort_reverse", NUL_TERM,
      view_sort_reverse, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "sort_subject", NUL_TERM,
      view_sort_subject, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "sort_sender", NUL_TERM,
      view_sort_sender, NULL, 0, NULL, NULL, BTN_ON },
};

Menu view_reply_submenu[] = {
  { NULL, "reply_sender", NUL_TERM, 
      view_reply_sender, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "reply_all", NUL_TERM,
      view_reply_all, NULL, 0, NULL, NULL, BTN_ON },
};

Menu view_forward_submenu[] = {
  { NULL, "forward", NUL_TERM, 
      view_forward, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "forward_attach",  NUL_TERM,
      view_forward_attach, NULL, 0, NULL, NULL, BTN_ON },
};

Menu view_send_menu[] = {
  { NULL, "compose_new", NUL_TERM, 
      main_compose, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "reply_menu", NUL_TERM, 
      NULL, view_reply_submenu, XtNumber(view_reply_submenu), 
      NULL, NULL, BTN_NOSELECTION },
  { NULL, "forward_menu", NUL_TERM,
      NULL, view_forward_submenu, XtNumber(view_forward_submenu),
      NULL, NULL, BTN_NOSELECTION },
  { NULL, "remail", NUL_TERM,
      view_remail, NULL, 0, NULL, NULL, BTN_NOSELECTION },
};

Menu view_mailbox_menu[] = {
  { NULL, "check_mailbox", NUL_TERM,
     view_check_mailbox, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "expunge_messages", NUL_TERM, 
      view_expunge_messages, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "open_mailbox", NUL_TERM,
      main_open, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "close_mailbox", NUL_TERM, 
      view_close_mailbox,   NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "no_update",   NUL_TERM,
      view_no_update, NULL, 0, NULL, NULL, BTN_ISMAIL },
  { NULL, "stopload_mailbox", NUL_TERM,
      view_stopload_mailbox, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "contload_mailbox", NUL_TERM,
      view_contload_mailbox, NULL, 0, NULL, NULL, BTN_ON },
};

Menu view_help_menu[] = {
  { NULL, "window_help", NUL_TERM,
      view_window_help, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "action_help", NUL_TERM, 
      view_action_help, NULL, 0, NULL, NULL, BTN_ON },
};

Menu view_menu[] = {
  { NULL, "file_menu", NUL_TERM, 
      NULL, view_file_menu, XtNumber(view_file_menu), 
      NULL, NULL, BTN_ON },
  { NULL, "edit_menu", NUL_TERM, 
      NULL, view_edit_menu, XtNumber(view_edit_menu), 
      NULL, NULL, BTN_ON },
  { NULL, "option_menu", NUL_TERM,
      NULL, view_options_menu, XtNumber(view_options_menu), 
      NULL, NULL, BTN_ON },
  { NULL, "send_menu", NUL_TERM,
      NULL, view_send_menu, XtNumber(view_send_menu), 
      NULL, NULL, BTN_ON },
  { NULL, "mailbox_menu", NUL_TERM,
      NULL, view_mailbox_menu, XtNumber(view_mailbox_menu),
      NULL, NULL, BTN_ON },
  { NULL, "prevwin", NUL_TERM,
      view_prev, NULL, 0, 
      NULL, NULL, BTN_NOPREV },
  { NULL, "listwin", NUL_TERM,
      view_lview_window, NULL, 0, 
      NULL, NULL, BTN_ON },
  { NULL, "nextwin", NUL_TERM,
      view_next, NULL, 0, 
      NULL, NULL, BTN_NONEXT },
  { NULL, "HELP_menu", NUL_TERM, 
      NULL, view_help_menu, XtNumber(view_help_menu), 
      NULL, NULL, BTN_ON },
};


Button_Table MLview_button_table[] = {


  { "Save", "view-save-fullheader",
      view_savefull, BTN_NOSELECTION },
  { "Save", "view-save-partheader",
      view_savepart, BTN_NOSELECTION },
  { "Save", "view-save-noheader",
      view_savenone, BTN_NOSELECTION },
  { "Shell", "view-pipe-fullheader",
      view_execfull, BTN_NOSELECTION },
  { "Shell", "view-pipe-partheader",
      view_execpart, BTN_NOSELECTION },
  { "Shell", "view-pipe-noheader",
      view_execnone, BTN_NOSELECTION },
  { "Print", "view-print-fullheader",
      view_printfull, BTN_NOSELECTION },
  { "Print", "view-print-partheader",
      view_printpart, BTN_NOSELECTION },
  { "Print", "view-print-noheader",
      view_printnone, BTN_NOSELECTION },
  { "Read", "view-read-selected",
      view_read_selected, BTN_NOSELECTION },
  { "Copy", "view-copy-messages",
      view_copy, BTN_NOSELECTION },
  { "Move", "view-move-messages",
      view_move, BTN_NOSELECTION },
  { "Address", "view-addressbook",
      main_address_book, BTN_ON },
  { "Notebook", "view-notebook",
      main_note_book, BTN_ON },
  { "Dismiss", "view-close-window",
      view_close_window, BTN_ON },
  { "Delete", "view-set-deleted",
      view_set_deleted, BTN_NOSELECTION },
  { "See", "view-set-seen",
      view_set_seen, BTN_NOSELECTION },
  { "Flag", "view-set-flagged",
      view_set_flagged, BTN_NOSELECTION },
  { "Answered", "view-set-answered",
      view_set_answered, BTN_NOSELECTION },
  { "Undelete", "view-clear-deleted",
      view_clear_deleted, BTN_NOSELECTION },
  { "Unsee", "view-clear-seen",
      view_clear_seen, BTN_NOSELECTION },
  { "Unflag", "view-clear-flagged",
      view_clear_flagged, BTN_NOSELECTION },
  { "Unanswered","view-clear-answered",
      view_clear_answered, BTN_NOSELECTION },
  { "Select", "view-select-all",
      view_select_messages, BTN_ON },
  { "New", "view-select-new",
      view_select_new_messages, BTN_ON },
  { "Unseen", "view-select-unseen",
      view_select_unseen_messages, BTN_ON },
  { "Clear", "view-unselect-messages",
      view_unselect_messages, BTN_NOSELECTION },
  { "Find", "view-find-messages",
      view_find, BTN_ON },
  { "Keyword", "view-set-keyword",
      view_set_keyword, BTN_NOKEYWORDS|BTN_NOSELECTION },
  { "Unkeyword", "view-unset-keyword",
      view_unset_keyword, BTN_NOKEYWORDS|BTN_NOSELECTION },
  { "Filters", "view-filters",
      view_filter, BTN_ON },
  { "Normal", "view-sort-normal",
      view_sort_normal, BTN_ON },
  { "Reverse", "view-sort-reverse",
      view_sort_reverse, BTN_ON },
  { "Subject", "view-sort-subject",
      view_sort_subject, BTN_ON },
  { "Sender", "view-sort-sender",
      view_sort_sender, BTN_ON },
  { "Reply", "view-reply-sender",
      view_reply_sender, BTN_ON },
  { "ReplyAll", "view-reply-all",
      view_reply_all, BTN_ON },
  { "Forward", "view-forward-insert",
      view_forward, BTN_ON },
  { "Forward", "view-forward-attach",
      view_forward_attach, BTN_ON },
  { "Compose", "view-compose-new",
      main_compose, BTN_ON },
  { "Remail", "view-remail",
      view_remail, BTN_NOSELECTION },
  { "Check", "view-check-mailbox",
     view_check_mailbox, BTN_ON },
  { "Expunge", "view-expunge-messages",
      view_expunge_messages, BTN_ON },
  { "OpenMailbox", "view-open-mailbox",
      main_open, BTN_ON },
  { "CloseMailbox", "view-close-mailbox",
      view_close_mailbox,   BTN_ON },
  { "NoNewsUpdate", "view-no-update-newsrc",
      view_no_update, BTN_ISMAIL },
  { "Pause", "view-pause-background-fetch",
      view_stopload_mailbox, BTN_ON },
  { "Resume","view-resume-background-fetch",
      view_contload_mailbox, BTN_ON },

};

Button_Menu MLview_default_buttons[] = {
  { "New",          view_select_new_messages, NULL, BTN_ON,          NULL },
  { "View",         view_read_selected,       NULL, BTN_NOSELECTION, NULL },
  { "Delete",       view_set_deleted,         NULL, BTN_NOSELECTION, NULL },
  { "Copy",         view_copy,                NULL, BTN_NOSELECTION, NULL },
  { "Move",         view_move,                NULL, BTN_NOSELECTION, NULL },
  { "Save",         view_savepart,            NULL, BTN_NOSELECTION, NULL },
  { "Reply",        view_reply_sender,        NULL, BTN_NOSELECTION, NULL },
  { "ReplyAll",     view_reply_all,           NULL, BTN_NOSELECTION, NULL },
  { "CloseMailbox", view_close_mailbox,       NULL, BTN_ON,          NULL },
  { "Expunge",      view_expunge_messages,    NULL, BTN_ON,          NULL },
};


#ifdef __STDC__
int view_number_buttons(void)
#else
int view_number_buttons()
#endif
{
  return(XtNumber(MLview_button_table));
}

#ifdef __STDC__
Button_Table *view_get_table(void)
#else
Button_Table *view_get_table()
#endif
{
  return(MLview_button_table);
}

#ifdef __STDC__
void view_default_buttons(void)
#else
void view_default_buttons()
#endif
{
  if(session.buttons[VIEW_BTN_MENU] == NULL)
    setup_default_buttons(&session.buttons[VIEW_BTN_MENU],
			  MLview_default_buttons,
			  XtNumber(MLview_default_buttons));
  return;
}


#ifdef __STDC__
void view_reset_buttons(void)
#else
void view_reset_buttons()
#endif
{
  Button_Menu *button_menu;

  if((session.view == NULL) 
     || (session.view->current == NULL)
     || (session.view->is_realized == FALSE))
    return;
  for(button_menu = session.buttons[VIEW_BTN_MENU];
      button_menu; button_menu = button_menu->next)
    if(button_menu->widget != NULL)
      XtDestroyWidget(button_menu->widget);

  view_default_buttons();

  make_pbuttons(session.view->rowcol, session.buttons[VIEW_BTN_MENU], 
		session.view->current->button_state, 
		(XtPointer) session.view);

  check_buttons(view_menu,XtNumber(view_menu), 
		session.buttons[VIEW_BTN_MENU], 
		session.view->current->button_state);
  return;
}




#ifdef __STDC__
void create_view_window(void)
#else
void create_view_window()
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  View *view;
  Widget menubar, action_button;
  XtTranslations translations;

  if(session.view == NULL) {
    view = new_view();
    view->current = NULL;
    view->lview = NULL;
    session.view = view;
  }
  else
    view = session.view;

  if(view->is_realized == TRUE) {
    de_iconify(view->shell);
    XmProcessTraversal(view->list, XmTRAVERSE_CURRENT);
    return;
  }

  view->shell = XtAppCreateShell("MLview", "MLview",
			       topLevelShellWidgetClass,
			       display, args, n ); 
  n = 0;

  setup_editres(view->shell);

  if(view_icon != (Pixmap) None)
    XtVaSetValues(view->shell,
		  XmNiconPixmap,view_icon,
		  NULL);

  view->form = XmCreateForm(view->shell,"form",   args, n); n = 0;

  XtSetArg(args[n], XmNtopAttachment,   XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNleftAttachment,  XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n ++;

  menubar = XmCreateMenuBar(view->form,"menubar", args, n); n = 0;
  XtManageChild(menubar);
  create_buttons(NULL, menubar, view_menu, XtNumber(view_menu), 
		      BTN_ON|BTN_NOSELECTION, (XtPointer) view, ROOTMENULEVEL);

  XtSetArg(args[n], XmNtopAttachment,   XmATTACH_WIDGET); n ++;
  XtSetArg(args[n], XmNtopWidget, menubar);  n ++;
  XtSetArg(args[n], XmNleftAttachment,  XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n ++;
  XtSetArg(args[n], XmNpacking, XmPACK_TIGHT); n ++;
  view->rowcol = XmCreateRowColumn(view->form, "rowcol", args, n); n = 0;

  view_default_buttons();

  make_pbuttons(view->rowcol, session.buttons[VIEW_BTN_MENU], 
		BTN_ON|BTN_NOSELECTION, (XtPointer) view);

  XtManageChild(view->rowcol);


  XtSetArg(args[n], XmNtraversalOn,    FALSE);                  n ++;
  XtSetArg(args[n], XmNtopAttachment,  XmATTACH_WIDGET);        n ++;
  XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM);          n ++;
  XtSetArg(args[n], XmNtopWidget,      view->rowcol);           n ++;

  action_button = XmCreatePushButton(view->form,"action_button", 
				     args, n); 
  n = 0;

  XtAddCallback(action_button,XmNactivateCallback,
		(XtCallbackProc) view_do_actions, view);
  XtManageChild(action_button);

  XtSetArg(args[n], XmNtopAttachment,   XmATTACH_WIDGET ); n ++;
  XtSetArg(args[n], XmNtopWidget,       view->rowcol    ); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM   ); n ++;
  XtSetArg(args[n], XmNleftAttachment,  XmATTACH_WIDGET ); n ++;
  XtSetArg(args[n], XmNleftWidget,      action_button   ); n ++;

  view->action = XmCreateTextField(view->form, "action_textfld",
				   args, n); 
  n = 0;

  XtManageChild(view->action);
  translations = XtParseTranslationTable(GLOBAL_text_field_translations);
  XtOverrideTranslations(view->action,translations);

  translations = 
    XtParseTranslationTable(GLOBAL_terminal_text_field_translations);
  XtOverrideTranslations(view->action,translations);  
  XtAddCallback(view->action, XmNactivateCallback, 
		(XtCallbackProc) view_do_actions, view);

  XtAddCallback(view->action, XmNmodifyVerifyCallback, 
		(XtCallbackProc) text_field_edit, NULL);

  XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n ++;
  XtSetArg(args[n], XmNtopWidget, view->action); n ++;
  XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNscrollBarDisplayPolicy, XmSTATIC); n ++;
  if(preferences.view_height) {
    XtSetArg(args[n], XmNvisibleItemCount, preferences.view_height ); n ++; 
  }
  XtSetArg(args[n], XmNselectionPolicy,XmEXTENDED_SELECT); n ++;
  XtSetArg(args[n], XmNlistSizePolicy,XmCONSTANT); n ++;


  view->list = XmCreateScrolledList(view->form,"list",args,n); n = 0;

  XtAddCallback(view->list, XmNextendedSelectionCallback,
		(XtCallbackProc) view_set_select_messages, view);

  XtAddCallback(view->list, XmNdefaultActionCallback,
		(XtCallbackProc) view_read_message, view);


  XtAddCallback(view->shell, XmNdestroyCallback,
		(XtCallbackProc) view_destroy, view);

  XtManageChild(view->list);

  XtManageChild(view->form);
  XtRealizeWidget(view->shell);
  XmProcessTraversal(view->list, XmTRAVERSE_CURRENT);

  view->is_realized = TRUE;
  return;
}


/* 
 *  Used in the do_action function for scripting flag settings.
 * IMAP flags have a preceding backslash. User flags don't.
 * Here we see if it's an IMAP flag, and return a pointer to
 * the flag name (which has the backslash). Otherwise, return
 * the original pointer; which is then assumed to be a user flag.
 */

#ifdef __STDC__
char *sys_flagname(char *flag)
#else
char *sys_flagname(flag)
     char *flag;
#endif
{
  int i;
  for(i = 0; SYS_SETTABLE_FLAGS[i] != NULL; i ++)
    if((strcasecmp(flag,(SYS_SETTABLE_FLAGS[i]) + 1)) == STRMATCH)
      return(SYS_SETTABLE_FLAGS[i]);
  return(flag);
}



#ifdef __STDC__
Widget new_list_widget(View *view)
#else
Widget new_list_widget(view) 
     View *view;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  
  Widget list;

  XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n ++;
  XtSetArg(args[n], XmNtopWidget, view->action); n ++;
  XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNscrollBarDisplayPolicy, XmSTATIC); n ++;
  XtSetArg(args[n], XmNvisibleItemCount, preferences.view_height ); n ++; 
  XtSetArg(args[n], XmNselectionPolicy,XmEXTENDED_SELECT); n ++;
  XtSetArg(args[n], XmNlistSizePolicy,XmCONSTANT); n ++;

  list = XmCreateScrolledList(view->form,"list",args,n); n = 0;

  XtAddCallback(list, XmNextendedSelectionCallback,
		(XtCallbackProc) view_set_select_messages, view);

  XtAddCallback(list, XmNdefaultActionCallback,
		(XtCallbackProc) view_read_message, view);

  return(list);
}


#ifdef __STDC__
void view_action_help(Widget w, View *view, XtPointer xp)
#else
void view_action_help(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  help(view->shell, VIEWACTIONHELPTITLE, VIEWACTIONHELPFILE, DOCTYPE_ENRICHED);
  return;
}

#ifdef __STDC__
void view_window_help(Widget w, View *view, XtPointer xp)
#else
void view_window_help(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  help(view->shell, VIEWWINDOWHELPTITLE, VIEWWINDOWHELPFILE, DOCTYPE_ENRICHED);
  return;
}

#ifdef __STDC__
void view_do_actions(Widget w, View *view, XtPointer xp)
#else
void view_do_actions(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  char **parsed_actions;
  char *raw_action = NULL;
  char *clean_action = NULL;
  int n;
  char * ptr;
  ACTIONREC * a;
  HEADERREC * h;
  Lview *lview;

  char warn[FILEBUFFLEN];

  lview = view->current;

  raw_action = GetTextField(view->action);

  if((lview == NULL) || (raw_action == NULL))
    return;

  if(*raw_action == NUL_TERM) {
    fs_give((void **) &raw_action);
    return;
  }

  clean_action = clean_and_parse_action(raw_action);
  raw_action = NULL;

  parsed_actions = get_actions(clean_action);
  fs_give((void **) &clean_action);

  if(parsed_actions == NULL)
    return;

  for(n = 0; parsed_actions[n]; n ++ ) {

    if(lview->script_arg)
      fs_give((void **) &lview->script_arg);

    lview->script_arg = NULL;

    a = find_action(parsed_actions[n]);

    if(a == NULL) {
      sprintf(warn,NO_SUCH_ACTION,parsed_actions[n]);
      mm_log(warn,WARN);
      break;
    }

    ptr = &parsed_actions[n][strlen(a->name) + 1];
    if(a->option == TRUE) {
      h = get_hdrlist(ptr);
      if(h != NULL) {
	lview->script_header = h->hdrtype;
	ptr += strlen(h->name) + 1;
      }
    }

    if(a->arg == TRUE)
      lview->script_arg = cpystr(ptr);

    lview->scripting = TRUE;

    switch(a->action) {

    case DELETE_ACT:
      view_set_deleted(w,view,xp);
      break;
    case EXPUNGE_ACT:
      view_expunge_messages(w,view,xp);
      break;
    case COPY_ACT:
      view_copy(w,view,xp);
      break;
    case MOVE_ACT:
      view_move(w,view,xp);
      break;
    case SAVE_ACT:
      view_save_dispatch(w,view,HEADER_UNKNOWN);
      break;
    case PRINT_ACT:
      view_print_dispatch(w,view,HEADER_UNKNOWN);
      break;
    case SHELL_ACT:
      view_exec_dispatch(w,view,HEADER_UNKNOWN);
      break;
    case SELECT_ACT:
      view_select_messages(w,view,xp);
      break;
    case NEW_ACT:
      view_select_new_messages(w,view,xp);
      break;
    case UNSEEN_ACT:
      view_select_unseen_messages(w,view,xp);
      break;
    case UNSELECT_ACT:
      view_unselect_messages(w,view,xp);
      break;
    case FLAG_ACT:
      view_set_flagged(w,view,xp);
      break;
    case SET_ACT:
      view_setflag(w,view,sys_flagname(lview->script_arg));
      break;
    case UNSET_ACT:
      view_clearflag(w,view,sys_flagname(lview->script_arg));
      break;
    case READ_ACT:
      view_read_selected(w,view,(XmListCallbackStruct *) xp);
      break;
      
    default:
      sprintf(warn,NO_SUCH_ACTION,a->name);
      mm_log(warn,WARN);

      break;
    }

    lview->scripting = FALSE;
    continue;
  }

  free_actions(parsed_actions);
  return;
}

#ifdef __STDC__
void view_set_select_messages(Widget w, View *view, XmListCallbackStruct *xp)
#else
void view_set_select_messages(w,view,xp)
     Widget w;
     View *view;
     XmListCallbackStruct *xp;
#endif
{
  unsigned long n = 0;
  Message_List *message_list;

  for(message_list = view->current->message_list;
      message_list; message_list = message_list->next) {
    if((is_selected(xp, message_list->number)) == TRUE)
      message_list->selected = TRUE;
    else
      message_list->selected = FALSE;
    if(message_list->selected == TRUE)
      n ++;
  }

  view->current->num_selected = n;
  if(n) 
    view->current->button_state &= ~(BTN_NOSELECTION);
  else
    view->current->button_state |= BTN_NOSELECTION;

  check_buttons(view_menu,XtNumber(view_menu), 
		session.buttons[VIEW_BTN_MENU], view->current->button_state);
  return;
}


#ifdef __STDC__
void view_check_mailbox(Widget w, View *view, XtPointer xp)
#else
void view_check_mailbox(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Mailbox *mailbox;

  for(mailbox = session.mailboxes; mailbox; mailbox = mailbox->next) {
    if(mailbox->mailstream == view->current->mailstream) {
      check_mailbox(mailbox);
      break;
    }
  }
  return;
}

#ifdef __STDC__
void view_stopload_mailbox(Widget w, View *view, XtPointer xp)
#else
void view_stopload_mailbox(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Mailbox *mailbox;

  for(mailbox = session.mailboxes; mailbox; mailbox = mailbox->next) {
    if(mailbox->mailstream == view->current->mailstream) {
      if(mailbox->bg_timer) {
	update_logical_views(mailbox,TRUE);
	mailbox->update_needed = 0;
	XtRemoveTimeOut(mailbox->bg_timer);
	mailbox->bg_timer = (XtIntervalId) 0;
	mm_log(BACKGROUND_STOP_MSG,MLNOTIFY);
	break;
      }
      else
	break;
    }
  }
  return;
}

#ifdef __STDC__
void view_contload_mailbox(Widget w, View *view, XtPointer xp)
#else
void view_contload_mailbox(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Mailbox *mailbox;

  for(mailbox = session.mailboxes; mailbox; mailbox = mailbox->next) {
    if(mailbox->mailstream == view->current->mailstream) {
      if(mailbox->bg_timer != (XtIntervalId) 0)
	return;
      mailbox->bg_timer 
	= XtAppAddTimeOut(context,
			  (mailbox->type == MAILBOX_TYPE_NEWS) 
			  ? (preferences.news_background * 1000)
			  : (preferences.mail_background * 1000),
			  (XtTimerCallbackProc) background_fetch,
			  mailbox);
      return;
    }
  }
  return;
}

#ifdef __STDC__
void view_find(Widget w, View *view, XtPointer xp)
#else
void view_find(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Lview *lview;
  Filter *filter;
  char *text;
  parse_errors errors;
  Message_List *new, *curr;
  Message_List *message_list;
  unsigned long count = 0L;
  char errmsg[FILEBUFFLEN];

  text = input_string(view->shell,SEARCH_FILT_STR,NULL,EDFILTERHELPFILE);
  if(text == NULL)
    return;
  if(*text == NUL_TERM) {
    fs_give((void **) &text);
    return;
  }

  filter = new_filter();
  filter->name = cpystr("Search");
  errors = lv_parse_filter(filter->name,text,
			   (LEAF **) &filter->search_tree,
			   &filter->text,
			   &filter->search_types,
			   errmsg);

  if(errors != parse_success) {
    free_filter(filter);
    fs_give((void **) &text);
    mm_log(errmsg,WARN);
    return;
  }

  view_unselect_messages(w,view,xp);

  new = do_imap_search(view->current->mailbox,view->current,
		       filter->search_tree);

  for(message_list = new; message_list; message_list = message_list->next)
    count ++;

  sprintf(errmsg,N_MESSAGES_FOUND,count,
	  (count == 1) ? MESSAGE_SINGULAR : MESSAGE_PLURAL);
  mm_log(errmsg,MLNOTIFY);
  if(count) {
    XtVaSetValues(view->list, XmNselectionPolicy, XmMULTIPLE_SELECT, NULL);
    for(curr = view->current->message_list; curr; curr = curr->next) {
      for(message_list = new; message_list; message_list = message_list->next){
	if(message_list->message == curr->message) {
	  XmListSelectPos(view->list,curr->number,FALSE);
	  curr->selected = TRUE;
	}
      }
    }
    XtVaSetValues(view->list, XmNselectionPolicy, XmEXTENDED_SELECT, NULL);
  }

  view->current->num_selected = count;
  if(count) 
    view->current->button_state &= ~(BTN_NOSELECTION);
  else
    view->current->button_state |= BTN_NOSELECTION;

  check_buttons(view_menu,XtNumber(view_menu), 
		session.buttons[VIEW_BTN_MENU], view->current->button_state);

  free_message_list(new,TRUE,FALSE);
  free_filter(filter);
  fs_give((void **) &text);
  return;
}


#ifdef __STDC__
void view_read_message(Widget w, View *view, XmListCallbackStruct *xp)
#else
void view_read_message(w,view,xp)
     Widget w;
     View *view;
     XmListCallbackStruct *xp;
#endif
{

  Message_List *message_list;

  view->current->num_selected = 0L;
  for(message_list = view->current->message_list;
      message_list; message_list = message_list->next) {
    message_list->selected = FALSE;
    if(message_list->number == xp->item_position) {
      message_list->selected = TRUE;
      view->current->num_selected = 1L;
      view->current->button_state &= ~(BTN_NOSELECTION);
      check_buttons(view_menu, XtNumber(view_menu), 
		    session.buttons[VIEW_BTN_MENU], view->current->button_state);
      view_read_selected(w,view,xp);
      break;
    }
  }
  return;
}

#ifdef __STDC__
void view_read_selected(Widget w, View *view, XmListCallbackStruct *xp)
#else
void view_read_selected(w,view,xp)
     Widget w;
     View *view;
     XmListCallbackStruct *xp;
#endif
{
  if(view->current->num_selected) {
    push_cursor(WATCH_CURSOR);
    read_messagelist(view->current->message_list);
    pop_cursor();
  }
  return;
}

#ifdef __STDC__
void view_next(Widget w, View *view, XtPointer xp)
#else
void view_next(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  if(view->current->next)
    view_context_switch(view,view->current,view->current->next);
  return;
}

#ifdef __STDC__
void view_prev(Widget w, View *view, XtPointer xp)
#else
void view_prev(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  if(view->current->prev)
    view_context_switch(view,view->current,view->current->prev);
  return;
}

#ifdef __STDC__
void view_expunge_messages(Widget w, View *view, XtPointer xp)
#else
void view_expunge_messages(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  push_cursor(WATCH_CURSOR);
  mail_expunge(view->current->mailstream);
  view_after_expunge();
  pop_cursor();
  return;
}

#ifdef __STDC__
char *rep_to_sortstr(Representation_Type representation)
#else
char *rep_to_sortstr(representation)
     Representation_Type representation;
#endif
{
  switch(representation) {
  case REP_DESCENDING:
    return(REVERS_PREFSTR);
    break;
  case REP_SORT_SENDER:
    return(SENDER_PREFSTR);
    break;
  case REP_SORT_SUBJECT:
    return(SUBJEC_PREFSTR);
    break;
  case REP_ASCENDING:
    default:
    return(NORMAL_PREFSTR);
    break;
    
  }
  /* NOTREACHED */
}


#ifdef __STDC__
Representation_Type sortstr_to_rep(char * sortstr)
#else
Representation_Type sortstr_to_rep(sortstr)
     char *sortstr;
#endif
{
  if(sortstr != NULL) {
    if((strncasecmp(sortstr,REVERS_PREFSTR,strlen(REVERS_PREFSTR))) 
       == STRMATCH)
      return(REP_DESCENDING);
    if((strncasecmp(sortstr,SENDER_PREFSTR,strlen(SENDER_PREFSTR))) 
       == STRMATCH)
      return(REP_SORT_SENDER);
    if((strncasecmp(sortstr,SUBJEC_PREFSTR,strlen(SUBJEC_PREFSTR))) 
       == STRMATCH)
      return(REP_SORT_SUBJECT);
  }
  return(REP_ASCENDING);
}

#ifdef __STDC__
void view_sort_reverse(Widget w, View *view, XtPointer xp)
#else
void view_sort_reverse(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Message_List *old, *new;

  if((view->current == NULL)
     || (view->current->filter == NULL)
     || (view->current->message_list == NULL))
    return;
  
  push_cursor(WATCH_CURSOR);

  view->current->representation = REP_DESCENDING;
  old = view->current->message_list;

  new = sort_message_list(view->current->message_list, 
			  view->current->representation);

  update_lview_message_list(view->current,old,new, FALSE);

  pop_cursor();
  return;
}

#ifdef __STDC__
void view_sort_normal(Widget w, View *view, XtPointer xp)
#else
void view_sort_normal(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Message_List *old, *new;
  
  if((view->current == NULL)
     || (view->current->filter == NULL)
     || (view->current->message_list == NULL))
    return;

  push_cursor(WATCH_CURSOR);

  view->current->representation = REP_ASCENDING;
  old = view->current->message_list;

  new = sort_message_list(view->current->message_list, 
			  view->current->representation);

  update_lview_message_list(view->current,old,new, FALSE);

  pop_cursor();
  return;
}

#ifdef __STDC__
void view_sort_subject(Widget w, View *view, XtPointer xp)
#else
void view_sort_subject(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Message_List *old, *new;
  
  if((view->current == NULL)
     || (view->current->filter == NULL)
     || (view->current->message_list == NULL))
    return;

  push_cursor(WATCH_CURSOR);

  view->current->representation = REP_SORT_SUBJECT;
  old = view->current->message_list;

  new = sort_message_list(view->current->message_list, 
			  view->current->representation);

  update_lview_message_list(view->current,old,new, FALSE);

  pop_cursor();
  return;
}

#ifdef __STDC__
void view_sort_sender(Widget w, View *view, XtPointer xp)
#else
void view_sort_sender(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Message_List *old, *new;
  
  if((view->current == NULL)
     || (view->current->filter == NULL)
     || (view->current->message_list == NULL))
    return;

  push_cursor(WATCH_CURSOR);

  view->current->representation = REP_SORT_SENDER;
  old = view->current->message_list;

  new = sort_message_list(view->current->message_list, 
			  view->current->representation);

  update_lview_message_list(view->current,old,new,FALSE);

  pop_cursor();
  return;
}


#ifdef __STDC__
void view_select_messages(Widget w, View *view, XtPointer xp)
#else
void view_select_messages(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  Message_List *message_list;
  unsigned long selected = 0L;

  view_unselect_messages(w,view,xp);

  push_cursor(WATCH_CURSOR);

  XtSetArg(args[n], XmNselectionPolicy, XmMULTIPLE_SELECT);    n ++;
  XtSetValues(view->list,args,n );                             n = 0;

  for(message_list = view->current->message_list;
      message_list; message_list = message_list->next) {
    XmListSelectPos(view->list,message_list->number,FALSE);
    message_list->selected = TRUE;
    selected ++;
  }

  XtSetArg(args[n], XmNselectionPolicy, XmEXTENDED_SELECT);    n ++;
  XtSetValues(view->list,args,n );                             n = 0;

  view->current->num_selected = selected;
  if(selected > 0) 
    view->current->button_state &= ~(BTN_NOSELECTION);
  else
    view->current->button_state |= BTN_NOSELECTION;

  check_buttons(view_menu,XtNumber(view_menu), 
		session.buttons[VIEW_BTN_MENU], view->current->button_state);
  pop_cursor();
  return;
}

#ifdef __STDC__
void view_unselect_messages(Widget w, View *view, XtPointer xp)
#else
void view_unselect_messages(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Message_List *message_list;

  push_cursor(WATCH_CURSOR);

  for(message_list = view->current->message_list;
      message_list; message_list = message_list->next) {
    XmListDeselectPos(view->list,message_list->number);
    message_list->selected = FALSE;
  }

  view->current->num_selected = 0L;
  view->current->button_state |= BTN_NOSELECTION;

  check_buttons(view_menu,XtNumber(view_menu), 
		session.buttons[VIEW_BTN_MENU], view->current->button_state);
  pop_cursor();
  return;
}

#ifdef __STDC__
void view_select_new_messages(Widget w, View *view, XtPointer xp)
#else
void view_select_new_messages(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  Message_List *message_list;
  MESSAGECACHE *elt;
  unsigned long selected = 0L;

  view_unselect_messages(w,view,xp);

  push_cursor(WATCH_CURSOR);

  XtSetArg(args[n], XmNselectionPolicy, XmMULTIPLE_SELECT);     n ++;
  XtSetValues(view->list,args,n );                              n = 0;

  for(message_list = view->current->message_list;
      message_list; message_list = message_list->next) {
    if((message_list->message == NULL) 
       || (message_list->message->fetched == FALSE)
       || (message_list->message->longcache == NULL))
      continue;
    elt = &(message_list->message->longcache->elt);
    if((elt) && (elt->recent) && (!elt->seen)) {
      XmListSelectPos(view->list,message_list->number,FALSE);
      message_list->selected = TRUE;
      selected ++;
    }
  }

  XtSetArg(args[n], XmNselectionPolicy, XmEXTENDED_SELECT);     n ++;
  XtSetValues(view->list,args,n );                              n = 0;

  view->current->num_selected = selected;
  if(selected) 
    view->current->button_state &= ~(BTN_NOSELECTION);
  else
    view->current->button_state |= BTN_NOSELECTION;

  check_buttons(view_menu,XtNumber(view_menu), 
		session.buttons[VIEW_BTN_MENU], view->current->button_state);
  pop_cursor();
  return;
}


#ifdef __STDC__
void view_select_unseen_messages(Widget w, View *view, XtPointer xp)
#else
void view_select_unseen_messages(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  Message_List *message_list;
  MESSAGECACHE *elt;
  unsigned long selected = 0L;

  view_unselect_messages(w,view,xp);

  push_cursor(WATCH_CURSOR);

  XtSetArg(args[n], XmNselectionPolicy, XmMULTIPLE_SELECT);     n ++;
  XtSetValues(view->list,args,n );                              n = 0;

  for(message_list = view->current->message_list;
      message_list; message_list = message_list->next) {
    if((message_list->message == NULL) 
       || (message_list->message->fetched == FALSE)
       || (message_list->message->longcache == NULL))
      continue;
    elt = &(message_list->message->longcache->elt);
    if((elt) && (!elt->seen)) {
      XmListSelectPos(view->list,message_list->number,FALSE);
      message_list->selected = TRUE;
      selected ++;
    }
  }

  XtSetArg(args[n], XmNselectionPolicy, XmEXTENDED_SELECT);     n ++;
  XtSetValues(view->list,args,n );                              n = 0;

  view->current->num_selected = selected;
  if(selected) 
    view->current->button_state &= ~(BTN_NOSELECTION);
  else
    view->current->button_state |= BTN_NOSELECTION;

  check_buttons(view_menu,XtNumber(view_menu), 
		session.buttons[VIEW_BTN_MENU], view->current->button_state);
  pop_cursor();
  return;
}


#ifdef __STDC__
void view_close_window(Widget w, View *view, XtPointer xp)
#else
void view_close_window(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  XtRemoveCallback(view->shell, XmNdestroyCallback,
		   (XtCallbackProc) view_destroy, view);
  XtDestroyWidget(view->shell);
  view->is_realized = FALSE;
  return;
}

#ifdef __STDC__
void view_destroy(Widget w, View *view, XtPointer xp)
#else
void view_destroy(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view->is_realized = FALSE;
  return;
}

#ifdef __STDC__
void view_copy_dispatch(Widget w, View *view, Boolean move)
#else
void view_copy_dispatch(w,view,move)
     Widget w;
     View *view;
     Boolean move;
#endif
{
  extern STRINGDRIVER mail_string;
  STRING bs;
  char *ptr;
  Boolean changed = FALSE;
  char *sequence = NULL;
  char *mailbox_name = NULL;
  Message_List *message_list;
  Message_List *new_message_list = NULL;
  int result = NIL;

  if(view->current->scripting && view->current->script_arg)
    mailbox_name = fix_mailboxpath(view->current->script_arg);

  if(mailbox_name == NULL)
    mailbox_name = 
      get_mailbox_name(view->shell, 
		       (view->current->mailbox->host)
		       ? view->current->mailbox->host
		       : EMPTYSTR,
		       (view->current->mailbox->type == MAILBOX_TYPE_NEWS)
		       ? preferences.imap_server
		       : view->current->mailbox->host, 
		       preferences.default_savebox,
		       &changed);

  if(mailbox_name == NULL)
    return;

  push_cursor(WATCH_CURSOR);

  if(changed == FALSE) {
    sequence = generate_sequence(view->current);

    if(sequence && *sequence && *mailbox_name) 
      result = mail_copy(view->current->mailstream, sequence, mailbox_name);
    if(result == T) {
      if(move) {
	mm_log(MOVE_OK, NIL);
	view_setflag(w,view, DELETED_FLAG);
	for(message_list = view->current->message_list; message_list; 
	    message_list = message_list->next) {
	  if(message_list->selected == TRUE) 
	    update_view_line(view->current->mailbox, message_list->message);
	}
      }
      else
	mm_log(COPY_OK, NIL);
    }
  }
  else {
    free_message_list(view->current->mailbox->found, TRUE, FALSE);
    view->current->mailbox->found = NULL;
    for(message_list = view->current->message_list; message_list;
	message_list = message_list->next) {
      if(message_list->selected == TRUE)
	mm_searched(message_list->message->mailstream,
		    message_list->message->msgno);
    }
    new_message_list = sort_message_list(view->current->mailbox->found, 
				     REP_ASCENDING);
    view->current->mailbox->found = NULL;
    result = SYSCALL_SUCCESS;
    if(new_message_list) {
      for(message_list = new_message_list; 
	  message_list; message_list = message_list->next) {
	if(message_list->message) {
	  result = copy_message_to_mailbox(message_list->message->mailstream,
					   message_list->message->msgno,
					   mailbox_name);
	  if(result != SYSCALL_SUCCESS)
	    break;
	}
      }
      if(result == SYSCALL_SUCCESS) {
	if(move) {
	  view_setflag(w,view, DELETED_FLAG);
	  for(message_list = new_message_list; 
	      message_list; message_list = message_list->next) 
	    update_view_line(view->current->mailbox, message_list->message);
	}
	mm_log((move) ? MOVE_OK : COPY_OK, NIL);
      }
      else
	mm_log((move) ? MOVE_FAILED : COPY_FAILED, NIL );
      free_message_list(new_message_list, TRUE, FALSE);
    }
  }
  
  if(sequence)
    fs_give((void **) &sequence);
  fs_give((void **) &mailbox_name);
  
  pop_cursor();
  return;
}




#ifdef __STDC__
void view_copy(Widget w, View *view, XtPointer xp)
#else
void view_copy(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_copy_dispatch(w, view, FALSE);
  return;
}


#ifdef __STDC__
void view_move(Widget w, View *view, XtPointer xp)
#else
void view_move(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_copy_dispatch(w, view, TRUE);
  return;
}

/*
 * To avoid excessive scrolling and flicker when we change context,
 * we grab a new list widget, load it up, and then change the old
 * one out from under the user. There's still a screen flash, but
 * that's life...
 */

#ifdef __STDC__
void view_context_switch(View *view,Lview *old,Lview *new)
#else
void view_context_switch(view,old,new)
     View *view;
     Lview *old;
     Lview *new;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  Dimension width, height;
  char *new_title = NULL;
  Widget list, oldlist;

  unsigned long total = 0L;
  unsigned long highlighted = 0L;
  Message_List *message_list;

  if(! view || view->is_realized == FALSE)
    return;

  XtSetArg(args[n], XmNwidth, &width);       n ++;
  XtSetArg(args[n], XmNheight, &height);     n ++;
  XtGetValues(view->shell, args, n );         n = 0;

  view->current = new;

  list = new_list_widget(view);

  if(new->title) {
    new_title = (char *) fs_get(strlen(new->title) + 64);
    strcpy(new_title,new->title);
    strcat(new_title, REPRESENTATION[new->representation]);
    XtSetArg(args[n], XmNtitle, new_title); n ++;
    XtSetValues(view->shell, args, n); n = 0;
    fs_give((void **) &new_title);
  }

  if((new->filter) && (new->filter->action))
    XmTextSetString(view->action,new->filter->action);
  else
    XmTextSetString(view->action,preferences.default_action);

  XtSetArg(args[n], XmNselectionPolicy, XmMULTIPLE_SELECT);     n ++;
  XtSetValues(list,args,n );                                    n = 0;

  for(message_list = new->message_list;
      message_list; message_list = message_list->next) {
    if(message_list->message) {
      XmListAddItemUnselected(list,message_list->message->viewline,0);
      message_list->number = (++total);
      if(message_list->selected) {
	XmListSelectPos(list,message_list->number, FALSE);
	highlighted ++;
      }
    }
  }

  XtSetArg(args[n], XmNselectionPolicy, XmEXTENDED_SELECT);     n ++;
  XtSetValues(list,args,n );                                    n = 0;

  new->num_selected = highlighted;

  if(new->representation == REP_ASCENDING)
    XmListSetBottomPos(list, new->count);
  else
    XmListSetBottomPos(list,1);

  oldlist = view->list;

  XtUnmanageChild(oldlist);

  view->list = list;
  XtManageChild(list);
  XtSetArg(args[n], XmNwidth, width );                          n ++;
  XtSetArg(args[n], XmNheight, height );                        n ++;
  XtSetValues(view->shell, args, n );                           n = 0;

  XtDestroyWidget(oldlist);
  if(highlighted != 0L)
    new->button_state &= ~(BTN_NOSELECTION);
  else
    new->button_state |= BTN_NOSELECTION;
  check_buttons(view_menu,XtNumber(view_menu), 
		session.buttons[VIEW_BTN_MENU], new->button_state);
  return;
}

/* 
 * Called when we are going to close a mailbox, and we need to 
 * destroy the current lview. Since any other views of this mailbox
 * are going away, we can't use them. This finds a view unrelated
 * to this mailbox or else NULL. It's safe to context switch to
 * another mailbox.
 */

#ifdef __STDC__
Lview *find_lview_another_mailbox(Mailbox *mailbox)
#else
Lview *find_lview_another_mailbox(mailbox)
     Mailbox *mailbox;
#endif
{
  Lview *lview = NULL;
  
  for(lview = session.view->lview; lview; lview = lview->prev) {
    if(lview->mailbox != mailbox)
      return(lview);
  }
  return(NULL);
}

#ifdef __STDC__
void view_unlink_destroy_lview(View *view,Lview *lview,Boolean closing)
#else
void view_unlink_destroy_lview(view,lview,closing)
     View *view;
     Lview *lview;
     Boolean closing;
#endif
{
  Lview *curr = NULL;
  Lview *next_view = NULL;

  /* First reparent any orphans we might make */

  if((closing == FALSE) && (lview->filter != NULL) 
     && (lview->filter->filter_type != FILTER_TYPE_ALL)) {
    for(curr = view->lview; curr; curr = curr->prev) {
      if((curr->mailstream == lview->mailstream) && (curr->parent == lview))
	  curr->parent = lview->parent;
    }
    for(curr = view->lview; curr; curr = curr->prev) {
      if((curr->parent != NULL) && (curr->level != (curr->parent->level + 1)))
	curr->level --;
    }
  }



  if(view->lview == lview)
    view->lview = lview->prev;

  if(lview->prev) {
    lview->prev->next = lview->next;
    if(lview->next)
      lview->prev->button_state &= ~(BTN_NONEXT);
    else
      lview->prev->button_state |= BTN_NONEXT;
  }

  if(lview->next) {
    lview->next->prev = lview->prev;
    if(lview->prev)
      lview->next->button_state &= ~(BTN_NOPREV);
    else
      lview->next->button_state |= BTN_NOPREV;
  }

  if(lview->filter && lview->filter->filter_type == FILTER_TYPE_ALL) 
    free_message_list(lview->message_list,TRUE,TRUE);
  else
    free_message_list(lview->message_list,TRUE,FALSE);

  free_filter(lview->filter);

  if(view->is_realized) {
    if(lview == view->current) {

      if(closing) {
	next_view = find_lview_another_mailbox(lview->mailbox);
      }
      else {
	if(lview->prev)
	  next_view = lview->prev;
	else
	  if(lview->next)
	    next_view = lview->next;
      }
      if(next_view)
	view_context_switch(view,NULL,next_view);
      else 
	view_close_window(view->shell,view,NULL);
    }
  }

  stuff_lview_list(view,TRUE,closing);

  free_lview(lview);
  return;
}

/* Returns the "ALL" lview (main mailbox) for a particular mailstream */

#ifdef __STDC__
Lview *get_all_lview(MAILSTREAM *mailstream)
#else
Lview *get_all_lview(mailstream)
     MAILSTREAM *mailstream;
#endif
{
  Lview *lview;

  if(session.view && session.view->lview)
    for(lview = session.view->lview; lview ; lview = lview->prev) 
      if((lview->mailstream == mailstream) && (lview->filter) 
	 && (lview->filter->filter_type == FILTER_TYPE_ALL))
	return(lview);

  return(NULL);
}

/* Returns the "UNFILTERED" lview for a particular mailstream */

#ifdef __STDC__
Lview *get_bucket_lview(MAILSTREAM *mailstream)
#else
Lview *get_bucket_lview(mailstream)
     MAILSTREAM *mailstream;
#endif
{
  Lview *lview;

  if(session.view && session.view->lview)
    for(lview = session.view->lview; lview ; lview = lview->prev) 
      if((lview->mailstream == mailstream) && (lview->filter) 
	 && (lview->filter->filter_type == FILTER_TYPE_BUCKET))
	return(lview);

  return(NULL);
}

#ifdef __STDC__
void load_default_filters( Lview *lview,Filter_Map *filter_map)
#else
void load_default_filters(lview,filter_map)
     Lview *lview;
     Filter_Map *filter_map;
#endif
{
  FILE *fp;
  Boolean found = FALSE;
  char buf[FILEBUFFLEN];
  char pname[FILEBUFFLEN];
  char *name;
  int level;
  char *parent_name;
  char *ptr;
  Lview *our_lview;
  Lview *parent_lview;
  Lview *pptr;
  Filter *our_filter;
  Filter *curr_filter;
  parse_errors errors;
  char *temp_txt;
  char errmsg[FILEBUFFLEN];

  mm_log(LOADING_FILTERS, NIL);

  if((fp = fopen(filter_map->filename,"r")) == NULL) {
    return;
  }

  /* 
   * parse the filter map file. 
   * filter_name [SP] level [SP] parent_name ( [SP] parent_name ) [LF]
   */

  while(fgets(buf,sizeof(buf),fp)) {
    buf[strlen(buf) - 1] = NUL_TERM;
    name = buf;
    ptr = strchr(buf,SPACECHAR);
    if(ptr == NULL)
      continue;
    *ptr = NUL_TERM;
    ptr ++;
    level = atoi(ptr);
    if(level == 0)
      continue;
    parent_name = strchr(ptr,SPACECHAR);
    if(parent_name == NULL)
      continue;
    parent_name ++;

    for(curr_filter = session.filters; 
	curr_filter; curr_filter = curr_filter->next) {

      /* find requested filter */

      if(strcmp(curr_filter->name,name) == STRMATCH)
	break;
    }

    if(curr_filter == NULL)
      continue;
    
    /* find the (active) parent */

    for(parent_lview = session.view->lview; 
	parent_lview; parent_lview = parent_lview->prev) {

      if(parent_lview->mailstream != lview->mailstream)
	continue;

      if(parent_lview->level != (level - 1))
	 continue;

      if(parent_lview->filter == NULL)
	continue;

      *pname = NUL_TERM;

      for(pptr = parent_lview, found = FALSE; pptr; pptr = pptr->parent) {
	if(pptr->filter && pptr->filter->name) {
	  if(found)
	    strcat(pname,SPACESTR);
	  found = TRUE;
	  strcat(pname,pptr->filter->name);
	}
	if(pptr->filter && pptr->filter->filter_type == FILTER_TYPE_ALL)
	    break;
      }
      if((*pname != NUL_TERM) && ((strcmp(pname,parent_name)) == STRMATCH))
	break;
    }

    /* no parent found */
    if(parent_lview == NULL)
      continue;

    /* we have a parent -- now see if the filter parses */

    our_filter = new_filter();
    
    our_filter->filter_type = FILTER_TYPE_USER;

    our_filter->name = cpystr(curr_filter->name);
    temp_txt = cpystr(curr_filter->text);
    our_filter->action = cpystr(curr_filter->action);

    errors = lv_parse_filter(our_filter->name,temp_txt,
			     (LEAF **) &our_filter->search_tree,
			     &our_filter->text,
			     &our_filter->search_types,
			     errmsg);

    fs_give((void **) &temp_txt);

    if(errors != parse_success) {
      free_filter(our_filter);
      mm_log(errmsg,WARN);
      continue;
    }
    our_lview = new_lview();
    our_lview->filter = our_filter;
    our_lview->parent = parent_lview;
    our_lview->level = level;
    create_lview(our_lview,TRUE);
  }
  fclose(fp);

  update_logical_views(lview->mailbox,FALSE);
  check_buttons(view_menu,XtNumber(view_menu), 
		session.buttons[VIEW_BTN_MENU], session.view->current->button_state);
  return;
}

#ifdef __STDC__
void attach_session_filters(Lview *lview)
#else
void attach_session_filters(lview)
     Lview *lview;
#endif
{
  Lview *our_lview;
  Lview *parent_lview;
  Filter *our_filter;
  Filter *curr_filter;
  parse_errors errors;
  char *temp_txt;
  char errmsg[FILEBUFFLEN];


  if(session.session_filters == NULL)
    return;

  mm_log(ATTACHING_FILTERS, NIL);


  for(curr_filter = session.session_filters; 
      curr_filter; curr_filter = curr_filter->next) {
    
    our_filter = new_filter();
    
    our_filter->filter_type = FILTER_TYPE_ONESHOT;
    our_filter->name = cpystr(curr_filter->name);
    temp_txt = cpystr(curr_filter->text);
    our_filter->action = cpystr(curr_filter->action);

    errors = lv_parse_filter(our_filter->name,temp_txt,
			     (LEAF **) &our_filter->search_tree,
			     &our_filter->text,
			     &our_filter->search_types,
			     errmsg);

    fs_give((void **) &temp_txt);
    our_lview = new_lview();
    our_lview->filter = our_filter;
    our_lview->parent = lview;
    our_lview->level = lview->level + 1;
    create_lview(our_lview,TRUE);
  }
  update_logical_views(lview->mailbox,TRUE);
  view_context_switch(session.view, session.view->current, our_lview);
  return;
}




#ifdef __STDC__
void create_lview(Lview *lview, Boolean batch)
#else
void create_lview(lview, batch)
     Lview *lview;
     Boolean batch;
#endif
{
  char buffer[FILEBUFFLEN];
  char *fixedname;
  Mailbox *mailbox;
  Lview *parent;
  Lview *bucket;
  Filter *filter;
  
  push_cursor(WATCH_CURSOR);
  
  create_view_window();
  if(! session.view->current) {
    pop_cursor();
    return;
  }

  mailbox = session.view->current->mailbox;

  lview->mailstream   = mailbox->mailstream;
  lview->mailbox      = mailbox;

  if(((bucket = get_bucket_lview(mailbox->mailstream)) == NULL)
     && lview->filter && lview->filter->filter_type == FILTER_TYPE_USER) {
    
    /* create the "UNFILTERED" view */
    
    filter = new_filter();
    filter->name = cpystr(BUCKET_VIEW_NAME);
    filter->action = cpystr(preferences.default_action);
    filter->filter_type = FILTER_TYPE_BUCKET;
    
    bucket = new_lview();
    bucket->parent = get_all_lview(mailbox->mailstream);
    bucket->level = 1;
    bucket->mailstream = mailbox->mailstream;
    bucket->mailbox = mailbox;
    bucket->filter = filter;

    if(mailbox->type == MAILBOX_TYPE_NEWS)
      filter->representation = sortstr_to_rep(preferences.news_sort);
    else
      filter->representation = sortstr_to_rep(preferences.mail_sort);

    bucket->representation = filter->representation;

    filter->mailboxname = cpystr(mailbox->imapname);
    fixedname = unfix_mailboxpath(mailbox->mailboxname);
    
    sprintf(buffer,"[%s] %s%s%s",
	    filter->name,
	    (mailbox->host) ? mailbox->host : EMPTYSTR,
	    (mailbox->host) ? ":" : EMPTYSTR,
	    fixedname);
    
    fs_give((void **) &fixedname);
    
    bucket->title = cpystr(buffer);
    
    if((mailbox->mailstream != NULL) 
       && (*mailbox->mailstream->user_flags == NULL))
      bucket->button_state |= BTN_NOKEYWORDS;
    
    if(mailbox->type == MAILBOX_TYPE_NEWS)
      bucket->button_state |= BTN_ISNEWS;
    else
      bucket->button_state |= BTN_ISMAIL;

    
    for(parent = session.view->lview; parent; parent = parent->prev) {
      if(parent == bucket->parent)
	break;
    }
    
    if(parent) {
      bucket->next = parent;
      bucket->prev = parent->prev;
      parent->prev = bucket;
      parent->button_state &= ~(BTN_NOPREV);
      if(bucket->prev) {
	bucket->button_state &= ~(BTN_NOPREV);
	bucket->prev->next = bucket;
	bucket->prev->button_state &= ~(BTN_NONEXT);
      }
      else
	bucket->button_state |= BTN_NOPREV;
      
      bucket->button_state &= ~(BTN_NONEXT);
    }
    else
      mm_log(CANNOT_FIND_PARENT, ERROR);
  }
  
  lview->filter->mailboxname = cpystr(mailbox->imapname);
  
  fixedname = unfix_mailboxpath(mailbox->mailboxname);
  
  sprintf(buffer,"[%s] %s%s%s",
	  lview->filter->name,
	  (mailbox->host) ? mailbox->host : EMPTYSTR,
	  (mailbox->host && *mailbox->host) ? ":" : EMPTYSTR,
	  fixedname);
  
  fs_give((void **) &fixedname);
  
  lview->title = cpystr(buffer);
  lview->representation = lview->filter->representation;

  /* 
   *  Link into our structure beneath the parent. We should always
   *  have a parent. session.view->lview is ordered in reverse.
   *  (Everything is "prev" to the root node).
   */
  
  if(mailbox->type == MAILBOX_TYPE_NEWS)
    lview->button_state |= BTN_ISNEWS;
  else
    lview->button_state |= BTN_ISMAIL;

  if((mailbox->mailstream != NULL) 
     && (*mailbox->mailstream->user_flags == NULL))
    lview->button_state |= BTN_NOKEYWORDS;
  
  for(parent = session.view->lview; parent; parent = parent->prev) {
    if(parent == lview->parent)
      break;
  }
  
  if(parent) {
    lview->next = parent;
    lview->prev = parent->prev;
    parent->prev = lview;
    parent->button_state &= ~(BTN_NOPREV);
    if(lview->prev) {
      lview->button_state &= ~(BTN_NOPREV);
      lview->prev->next = lview;
      lview->prev->button_state &= ~(BTN_NONEXT);
    }
    else
      lview->button_state |= BTN_NOPREV;

    lview->button_state &= ~(BTN_NONEXT);
  }

  else 
    mm_log(CANNOT_FIND_PARENT, ERROR);

  /* 
   * If we're running in "batch" mode, there could be several filters
   * to attach. The caller has promised to update the logical views
   * when it's done attaching everything it needs. This avoids lengthy
   * (steadily increasing) delays as the number of filters is growing;
   * and it also avoids excessive flicker. Set batch to false if there's
   * only one filter to attach. Everything will be updated and sorted for 
   * you when this function returns.
   */

  if(batch == FALSE) {
    update_logical_views(mailbox,TRUE);
    view_context_switch(session.view, session.view->current, lview);
  }
  pop_cursor();
  return;
}

/* 
 * Creates a main mailbox ("ALL" lview) and loads 
 * its message info.
 */

#ifdef __STDC__
void create_all_lview(Mailbox *mailbox)
#else
void create_all_lview(mailbox)
     Mailbox *mailbox;
#endif
{
  Lview  *lview;
  Filter *filter;
  Filter_Map *filter_map;
  Message_List *old, *new;
  char buffer[FILEBUFFLEN];
  char *fixedname;

  push_cursor(WATCH_CURSOR);
  create_view_window();

  /* 
   * "ALL" is just another logical view except it has a few
   * special properties. It ALWAYS holds a message structure
   * for each message in the mailbox. So whenever mailbox changes 
   * occur, we update this view to guarantee that it stays consistent
   * with the server's representation. We then derive other logical
   * views from it. We also don't put any special meaning on the name 
   * "ALL". We set the filter_type to FILTER_TYPE_ALL instead, and
   * use a combination of the filter_type and either the mailstream or 
   * mailbox to absolutely identify it.
   */

  lview = new_lview();
  lview->mailstream   = mailbox->mailstream;
  lview->mailbox      = mailbox;
  lview->count        = mailbox->nmsgs;

  filter = new_filter();
  filter->mailboxname = cpystr(mailbox->imapname);
  filter->name = cpystr(ALL_VIEW_NAME);
  filter->filter_type = FILTER_TYPE_ALL;

  if(mailbox->type == MAILBOX_TYPE_NEWS)
    filter->representation = sortstr_to_rep(preferences.news_sort);
  else
    filter->representation = sortstr_to_rep(preferences.mail_sort);

  lview->filter = filter;
  lview->representation = filter->representation;

  fixedname = unfix_mailboxpath(mailbox->mailboxname);

  sprintf(buffer,"[%s] %s%s%s",
	  filter->name,
	  (mailbox->host) ? mailbox->host : EMPTYSTR,
	  (mailbox->host && *mailbox->host) ? ":" : EMPTYSTR,
	  fixedname);

  fs_give((void **) &fixedname);

  lview->title = cpystr(buffer);
  lview->next = NULL;
  lview->button_state |= BTN_NONEXT;

  if(session.view->lview) {
    lview->prev = session.view->lview;
    lview->button_state &= ~(BTN_NOPREV);
    session.view->lview->next = lview;
    session.view->lview->button_state &= ~(BTN_NONEXT);
  }
  else { 
    lview->button_state |= BTN_NOPREV;
    lview->prev = NULL;
  }

  if(mailbox->type == MAILBOX_TYPE_NEWS)
    lview->button_state |= BTN_ISNEWS;
  else
    lview->button_state |= BTN_ISMAIL;

  if((mailbox->mailstream != NULL) 
     && (*mailbox->mailstream->user_flags == NULL))
    lview->button_state |= BTN_NOKEYWORDS;

  lview->button_state |= BTN_ISALL;

  session.view->lview = lview;

  fast_fetch_message_info(mailbox,lview);

  lview->has_new = has_new_messages(lview,TRUE);

  old = lview->message_list;

  new = sort_message_list(lview->message_list,lview->representation);

  update_lview_message_list(lview,old,new,FALSE);

  stuff_lview_list(session.view,TRUE, FALSE);

  view_context_switch(session.view, session.view->current, lview);

  /* 
   * If there are any saved active filters for this mailbox, load them 
   * at this time.
   */

  if((filter_map = get_filter_map(lview)) != NULL)
    load_default_filters(lview,filter_map);
  attach_session_filters(lview);
  pop_cursor();
  return;
}

/*
 * Called from the c-client when a (single) message is expunged.
 * We go through this in distinct steps. First, remove the message
 * from the "ALL" view (main mailbox), but !!DON'T!! yet free the message.
 * We need to check it against all our views for this mailbox. Decrement 
 * any higher message numbers in the message structures we encounter in
 * this view by one. Once that's accomplished, we go through all our 
 * logical views for this mailbox and unlink any message_list structures 
 * which contain the expunged message. We can free these message_list 
 * structures at this point, but we still don't free the message itself.
 * The final act is to destroy the message structure itself once all of 
 * its references have been eliminated.
 * Since expunge is always initiated by us, when we perform the act,
 * we follow up by calling the view_after_expunge() procedure to update
 * the visuals. This takes care of all the highlighting stuff, and
 * avoids excessive flicker if we are expunging several messages.
 */

#ifdef __STDC__
void view_expunge(MAILSTREAM *mailstream,unsigned long msgno)
#else
void view_expunge(mailstream,msgno)
     MAILSTREAM *mailstream;
     unsigned long msgno;
#endif
{
  Boolean found_it = FALSE;
  Lview *lview;

  Message_List *message_list;
  Message_List *dead_message_list = NULL;
  unsigned long x = 1L;

  if(session.view == NULL)
    return;

  lview = get_all_lview(mailstream);

  if(! lview)
    return;

  found_it = FALSE;

  for(message_list = lview->message_list; 
      message_list; message_list = message_list->next) {

    if(message_list->message == NULL)
      continue;

    if(message_list->message->msgno == msgno) {
      found_it = TRUE;
      dead_message_list = message_list;

      if(lview->count)
	lview->count --;

      if((message_list->selected == TRUE) && (lview->num_selected > 0L)) {
	lview->num_selected --;
	if(lview->num_selected == 0L)
	  lview->button_state |= BTN_NOSELECTION;
      }

      if(message_list->next)
	message_list->next->prev = message_list->prev;
      if(message_list->prev) 
	message_list->prev->next = message_list->next;
      else
	lview->message_list = message_list->next;
    }  

    if(message_list->message->msgno > msgno) 
      message_list->message->msgno --;
  }

  /* 
   * Yes, we could have done this while we were running through but it 
   * complicates the logic. I'll probably work that in at some point, 
   * because this could cause delays on a huge mailbox.
   */

  for(message_list = lview->message_list; 
      message_list; message_list = message_list->next) {
    message_list->number = x;
    x ++;
  }

  lview->has_new = has_new_messages(lview,FALSE);
  
  if(found_it == FALSE)       /* shouldn't happen */
    return;

  /* Update any other lviews attached to this stream. */

  for(lview = session.view->lview; lview; lview = lview->prev) {
    if(lview->mailstream != mailstream)
      continue;
    if(lview->filter && lview->filter->filter_type == FILTER_TYPE_ALL)
      continue;  /* already got this guy */

    found_it = FALSE;

    for(message_list = lview->message_list; 
	message_list; message_list = message_list->next) {
      
      if(message_list->message == NULL)
	continue;
      
      if(message_list->message == dead_message_list->message) {
	found_it = TRUE;
	
	if(lview->count)
	  lview->count --;
	if((message_list->selected == TRUE) && (lview->num_selected > 0L)) {
	  lview->num_selected --;
	  if(lview->num_selected == 0L)
	    lview->button_state |= BTN_NOSELECTION;
	}
	
	if(message_list->next)
	  message_list->next->prev = message_list->prev;
	if(message_list->prev) 
	  message_list->prev->next = message_list->next;
	else
	  lview->message_list = message_list->next;
	
	free_message_list(message_list,FALSE,FALSE);
      }
    }

    lview->has_new = has_new_messages(lview,FALSE);

    x = 1L;
    for(message_list = lview->message_list; 
	message_list; message_list = message_list->next) {
      message_list->number = x;
      x ++;
    }
  }

  free_message_list(dead_message_list,FALSE,TRUE);
  return;
}

#ifdef __STDC__
void view_after_expunge(void)
#else
void view_after_expunge()
#endif
{

  read_after_expunge();
  if(session.view->is_realized) {
    view_context_switch(session.view,
			session.view->current,session.view->current);
  }
  stuff_lview_list(session.view,TRUE,FALSE);
  return;
}

#ifdef __STDC__
void view_close_mailbox(Widget w, View *view, XtPointer xp)
#else
void view_close_mailbox(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Mailbox *mailbox, *next;

  for(mailbox = session.mailboxes; mailbox ; mailbox = next) {

    next = mailbox->next;
    if(mailbox->mailstream != view->current->mailstream)
      continue;
    close_mailbox(mailbox);
    break;

  }
  return;
}


#ifdef __STDC__
void view_no_update(Widget w, View *view, XtPointer xp)
#else
void view_no_update(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  if(view && view->current && view->current->mailbox)
    view->current->mailbox->no_update = TRUE;
  return;
}




#ifdef __STDC__
void view_lview_window(Widget w, View *view, XtPointer xp)
#else
void view_lview_window(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  XtTranslations translations;
  Widget form, menubar, dummy;

  if(view->list_is_realized == TRUE) {
    de_iconify(view->list_shell);
    return;
  }

  view->list_shell = XtAppCreateShell("MLlview","MLlview",
				    topLevelShellWidgetClass,
				    display, args, n ); 
  n = 0;

  setup_editres(view->list_shell);

  if(list_icon != (Pixmap) None)
    XtVaSetValues(view->list_shell,
		  XmNiconPixmap,list_icon,
		  NULL);

  form = XmCreateForm(view->list_shell, "form", args, n); n = 0;
  XtSetArg(args[n], XmNtopAttachment,   XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNleftAttachment,  XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n ++;

  menubar = XmCreateMenuBar(form,"menubar", args, n);     n = 0;
  XtManageChild(menubar);
  create_buttons(NULL, menubar,view_lview_menu, 
		 XtNumber(view_lview_menu), 
		 BTN_ON, (XtPointer) view, ROOTMENULEVEL);

  /* 
   * An invisible text field. This is used only to set the window 
   * width, since the list widget living under the form can change 
   * size otherwise, depending only on its contents.
   */

  XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n ++;
  dummy = XmCreateTextField(form,"dummy", args, n); n = 0;
  XtManageChild(dummy);

  XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n ++;
  XtSetArg(args[n], XmNtopWidget, menubar); n ++;
  XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n ++;
  
  XtSetArg(args[n], XmNscrollBarDisplayPolicy, XmSTATIC);        n ++;
  XtSetArg(args[n], XmNlistSizePolicy,         XmCONSTANT);      n ++;
  XtSetArg(args[n], XmNselectionPolicy,        XmSINGLE_SELECT); n ++;
  view->list_list = 
    XmCreateScrolledList(form,"list",args,n); 
  n = 0;

  XtAddCallback(view->list_list,
		XmNsingleSelectionCallback, 
		(XtCallbackProc) view_lview_select, view);
  XtAddCallback(view->list_list,
		XmNdefaultActionCallback, 
		(XtCallbackProc) view_lview_accept, view);
  XtAddCallback(view->list_shell,
		XmNdestroyCallback, 
		(XtCallbackProc) view_lview_destroy, view);

  stuff_lview_list(view,FALSE,FALSE);

  XtManageChild(view->list_list); 
  XtManageChild(form);
  XtRealizeWidget(view->list_shell);
  view->list_is_realized = TRUE;
  return;
}

#ifdef __STDC__
void view_lview_save(Widget w, View *view, XtPointer xp)
#else
void view_lview_save(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  FILE *fp;
  Lview *lview;
  Lview *current;
  Filter_Map *filter_map;
  char log[MAXPATHLEN + 128];
  Lview *pptr;

  for(current = view->lview; current; current = current->prev) {
    if((current->filter == NULL) 
       || (current->filter->filter_type != FILTER_TYPE_ALL))
      continue;
    filter_map = get_filter_map(current);
    if(filter_map == NULL)
      filter_map = generate_filter_map(current);
    
    push_cursor(WATCH_CURSOR);

    if((fp = fopen(filter_map->filename,"w")) == NULL) {
      sprintf(log,CANNOT_WRITE_FILE, filter_map->filename);
      mm_log(log,WARN);
      pop_cursor();
      continue;
    }
    (void) chmod(filter_map->filename,S_IRWXU);

    for(lview = current->prev; lview; lview = lview->prev) {
      if(lview->filter == NULL)
	continue;
      if(lview->filter->filter_type == FILTER_TYPE_ALL) {
	fclose(fp);
	break;
      }
      if(lview->filter->filter_type != FILTER_TYPE_USER)
	continue;
      if((lview->parent == NULL) 
	 || (lview->parent->filter == NULL)
	 || (lview->parent->filter->name == NULL))
	continue;

      fprintf(fp,"%s %d",
	      lview->filter->name,
	      lview->level);
      pptr = lview->parent;

      for(pptr = lview->parent; pptr; pptr = pptr->parent) {
	fprintf(fp," %s",(pptr->filter && pptr->filter->name) ? 
		  pptr->filter->name : SPACESTR);
	if(pptr->filter 
	   && pptr->filter->filter_type == FILTER_TYPE_ALL)
	  break;
      }
      fprintf(fp,"\n");
    }
    fclose(fp);
  }
  pop_cursor();
  return;
}


#ifdef __STDC__
void view_lview_select(Widget w, View *view, XmListCallbackStruct *xp)
#else
void view_lview_select(w,view,xp)
     Widget w;
     View *view;
     XmListCallbackStruct *xp;
#endif
{
  Lview *lview;

  for(lview = view->lview; lview; lview = lview->prev) {
    if(lview->number == xp->item_position) {
      if(lview->selected == FALSE) {
	lview->selected = TRUE;
	create_view_window();
	view_context_switch(view,view->current,lview);
	de_iconify(view->shell);
      }
      else
	XmListSelectPos(view->list_list,xp->item_position, FALSE);
    }
    else
      lview->selected = FALSE;
  }
  return;
}

#ifdef __STDC__
void view_lview_accept(Widget w, View *view, XmListCallbackStruct *xp)
#else
void view_lview_accept(w,view,xp)
     Widget w;
     View *view;
     XmListCallbackStruct *xp;
#endif
{
  return;
}

#ifdef __STDC__
void view_lview_hide(Widget w, View *view, XtPointer xp)
#else
void view_lview_hide(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view->hide_empty = TRUE;
  stuff_lview_list(view,TRUE,FALSE);
  return;
}

#ifdef __STDC__
void view_lview_show(Widget w, View *view, XtPointer xp)
#else
void view_lview_show(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view->hide_empty = FALSE;
  stuff_lview_list(view,TRUE,FALSE);
  return;
}

#ifdef __STDC__
void view_lview_detach(Widget w, View *view, XtPointer xp)
#else
void view_lview_detach(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Lview *lview;
  Filter *filter, *session_filter;
  Boolean found = FALSE;

  for(lview = view->lview; lview; lview = lview->prev) {
    if(lview->selected == TRUE) {

      /* 
       * Annoying, but we have to make the lview we want
       * to detach the ``current'' lview. We might get some update 
       * flicker as it gets switched in, and then immediately
       * switched out again. That's life. "view_close_mailbox()"
       * and "view_unlink_destroy_lview()" operate on the assumption
       * that the current lview is the target.
       */

      view_context_switch(view,view->current,lview);
      break;
    }
  }

  if(! lview)
    return;

  for(session_filter = session.session_filters;
      session_filter; session_filter = session_filter->next) {
    if(session_filter->name && lview->filter && lview->filter->name)
      if((strcmp(session_filter->name,lview->filter->name)) == STRMATCH)
	break;
  }

  if(session_filter) {
    if((ml_confirm(view->list_shell,DETACHFROMSESSION)) == TRUE) {
      if(session_filter->next)
	session_filter->next->prev = session_filter->prev;
      if(session_filter->prev) 
	session_filter->prev->next = session_filter->next;
      else
	session.session_filters = session_filter->next;
      free_filter(session_filter);
    }
  }

  if(lview->filter && lview->filter->filter_type == FILTER_TYPE_ALL)
    view_close_mailbox(w,view,xp);
  else  
    view_unlink_destroy_lview(view,lview,FALSE);
  return;
}

#ifdef __STDC__
void view_lview_dismiss(Widget w, View *view, XtPointer xp)
#else
void view_lview_dismiss(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Lview *lview;

  for(lview = view->lview; lview; lview = lview->prev)
    if(lview->selected == TRUE)
      lview->selected = FALSE;

  XtRemoveCallback(view->list_shell,
		   XmNdestroyCallback, 
		   (XtCallbackProc) view_lview_destroy, view);

  XtDestroyWidget(view->list_shell);
  view->list_is_realized = FALSE;
  return;
}

#ifdef __STDC__
void view_lview_destroy(Widget w, View *view, XtPointer xp)
#else
void view_lview_destroy(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Lview *lview;

  for(lview = view->lview; lview; lview = lview->prev)
    if(lview->selected == TRUE)
      lview->selected = FALSE;

  view->list_is_realized = FALSE;
  return;
}

#ifdef __STDC__
void view_lview_help(Widget w, View *view, XtPointer xp)
#else
void view_lview_help(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  help(view->list_shell, VIEWLISTHELPTITLE, VIEWLISTHELPFILE, 
       DOCTYPE_ENRICHED);
  return;
}


/* 
 * Stuff the view selector window. Update is set to false to stuff
 * it the first time around. Thereafter, it is always called with
 * update set to true, in order to refresh the contents. The closing
 * flag means to not check the new message status, since the message cache
 * structures might be bogus at that point. It will just go on prior
 * knowledge which should still be accurate.
 */


#ifdef __STDC__
void stuff_lview_list(View *view, Boolean update, Boolean closing)
#else
void stuff_lview_list(view, update, closing)
     View *view;
     Boolean update;
     Boolean closing;
#endif
{
  Lview *lview;
  int n = 1;
  int level = 0;
  char buffer[FILEBUFFLEN];
  int spaces;
  char *fixed_name = NULL;
  int font_type = 0;
  XmString xstr;


  if(view == NULL) 
    return;

  if(update) {
    if(view->list_is_realized == FALSE)
      return;
    XmListDeselectAllItems(view->list_list);
    XmListDeleteAllItems(view->list_list);
  }

  for(lview = view->lview; lview; lview = lview->prev) {

    if(closing != TRUE)
      lview->has_new = has_new_messages(lview,FALSE);

    font_type = 0;

    if((view->hide_empty == TRUE) && (lview->count == 0L)) {
      lview->number = 0;
      lview->selected = FALSE;
      continue;
    }

    lview->number = n;
    lview->selected = FALSE;
    buffer[0] = NUL_TERM;

    sprintf(buffer,"%6lu ",lview->count);

    if((lview->filter) 
       && (lview->filter->filter_type == FILTER_TYPE_ONESHOT))
      font_type += 2;

    if(lview->has_new)
      font_type += 1;

    for(level = 0; level < lview->level; level ++)
      strcat(buffer,">");
    if(lview->level)
      strcat(buffer," ");
    strcat(buffer,lview->filter->name);
    for(spaces = strlen(buffer); spaces < 30; spaces ++)
      strcat(buffer,SPACESTR);

    if(lview->mailbox && lview->mailbox->mailboxname)
      fixed_name = unfix_mailboxpath(lview->mailbox->mailboxname);
    if(fixed_name) {
      strcat(buffer,fixed_name);
      fs_give((void **) &fixed_name);
    }
    else
      strcat(buffer,EMPTYSTR);

    if(lview->mailbox && lview->mailbox->host) {
      strcat(buffer,"(");
      strcat(buffer,lview->mailbox->host);
      strcat(buffer,")");
    }
      
    switch(font_type) {
    case 1:
      xstr = XmStringCreate(buffer, "BOLD");
      break;
    case 2:
      xstr = XmStringCreate(buffer, "ITALIC");
      break;
    case 3:
      xstr = XmStringCreate(buffer, "BOLD_ITALIC");
      break;
    case 0:
    default:
      xstr = XmStringCreate(buffer, "NORMAL");
      break;
    }

    XmListAddItemUnselected(view->list_list,xstr,0);
    XmStringFree(xstr);
    n ++;
  }
  return;
}

#ifdef __STDC__
char *view_fetch(Message *message,Header_Mode header_mode)
#else
char *view_fetch(message,header_mode)
     Message *message;
     Header_Mode header_mode;
#endif
{
  char *mail_header = NULL;
  char *header = NULL;
  char *mail_text = NULL;
  char *everything;

  if((message == NULL) 
     || (message->mailstream == NULL)
     || (message->msgno == 0L))
    return(NULL);

  mail_header = cpystr(mail_fetchheader(message->mailstream,message->msgno));
  stripcr(mail_header);
  if(mail_header[strlen(mail_header) - 1] == LFCHAR)
    mail_header[strlen(mail_header) - 1] = NUL_TERM;

  if(header_mode == HEADER_FULL) 
    header = cpystr(mail_header);

  if(header_mode == HEADER_PART)
    header = get_short_header(mail_header);

  if(header == NULL)
    header = cpystr(EMPTYSTR);

  mail_text = cpystr(mail_fetchtext(message->mailstream,message->msgno));
  stripcr(mail_text);

  everything = (char *) fs_get(strlen(header) + strlen(mail_text) + 8);
  strcpy(everything,header);
  if(*header != NUL_TERM)
    strcat(everything,LFSTR);
  strcat(everything,mail_text);

  fs_give((void **) &mail_header);
  fs_give((void **) &header);
  fs_give((void **) &mail_text);

  return(everything);
}

#ifdef __STDC__
void view_printfull(Widget w, View *view, XtPointer xp)
#else
void view_printfull(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_print_dispatch(w,view,HEADER_FULL);
  return;
}

#ifdef __STDC__
void view_printpart(Widget w, View *view, XtPointer xp)
#else
void view_printpart(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_print_dispatch(w,view,HEADER_PART);
  return;
}

#ifdef __STDC__
void view_printnone(Widget w, View *view, XtPointer xp)
#else
void view_printnone(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_print_dispatch(w,view,HEADER_NONE);
  return;
}

#ifdef __STDC__
void view_print_dispatch(Widget w,View *view,Header_Mode header_mode)
#else
void view_print_dispatch(w,view,header_mode)
     Widget w;
     View *view;
     Header_Mode header_mode;
#endif
{
  Message_List *message_list;
  char *contents;
  int errors = 0;

  if(*preferences.print_command == NUL_TERM) {
    mm_log(NO_PRINT_COMMAND, WARN);
    return;
  }

  push_cursor(WATCH_CURSOR);

  for(message_list = view->current->message_list; message_list;
      message_list = message_list->next ) {

    if(message_list->selected) {
      contents = view_fetch(message_list->message,
			    (view->current->scripting)
			    ? view->current->script_header : header_mode);
      if(! contents)
	continue;
      
      if(write_to_pipe(preferences.print_command,
		    NULL, contents, strlen(contents)))
	errors ++;
      fs_give((void **) &contents);
    }
  }
  if(errors)
    mm_log(PRINT_FAILED, WARN);
  else
    mm_log(PRINT_OK, NIL);

  pop_cursor();
  return;
}

#ifdef __STDC__
void view_execfull(Widget w, View *view, XtPointer xp)
#else
void view_execfull(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_exec_dispatch(w,view,HEADER_FULL);
  return;
}

#ifdef __STDC__
void view_execpart(Widget w, View *view, XtPointer xp)
#else
void view_execpart(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_exec_dispatch(w,view,HEADER_PART);
  return;
}

#ifdef __STDC__
void view_execnone(Widget w, View *view, XtPointer xp)
#else
void view_execnone(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_exec_dispatch(w,view,HEADER_NONE);
  return;
}


#ifdef __STDC__
void view_exec_dispatch(Widget w,View *view,Header_Mode header_mode)
#else
void view_exec_dispatch(w,view,header_mode)
     Widget w;
     View *view;
     Header_Mode header_mode;
#endif
{
  Message_List *message_list;
  char *contents;
  int errors = 0;
  char *command = NULL;

  if(view->current->scripting && view->current->script_arg)
    command = cpystr(view->current->script_arg);

  if(command == NULL)
    command = input_string(view->shell,INPUTCOMMANDSTR, 
			   session.last_command, SHELLCOMMANDHELPFILE);
  if(command == NULL)
    return;

  if(*command == NUL_TERM) {
    fs_give((void **) &command);
    return;
  }

  push_cursor(WATCH_CURSOR);

  for(message_list = view->current->message_list; message_list;
      message_list = message_list->next ) {

    if(message_list->selected) {
      contents = view_fetch(message_list->message,header_mode);
      if(! contents)
	continue;

      
      if(write_to_pipe(command,
		    NULL, contents, strlen(contents)))
	errors ++;
      fs_give((void **) &contents);
    }
  }
  if(errors)
    mm_log(PIPE_FAILED, WARN);

  if(session.last_command != NULL)
    fs_give((void **) &session.last_command);
  session.last_command = command;

  pop_cursor();
  return;
}


#ifdef __STDC__
void view_savefull(Widget w, View *view, XtPointer xp)
#else
void view_savefull(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_save_dispatch(w,view,HEADER_FULL);
  return;
}

#ifdef __STDC__
void view_savepart(Widget w, View *view, XtPointer xp)
#else
void view_savepart(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_save_dispatch(w,view,HEADER_PART);
  return;
}

#ifdef __STDC__
void view_savenone(Widget w, View *view, XtPointer xp)
#else
void view_savenone(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_save_dispatch(w,view,HEADER_NONE);
  return;
}

#ifdef __STDC__
void view_save_dispatch(Widget w, View *view, Header_Mode header_mode)
#else
void view_save_dispatch(w,view,header_mode)
     Widget w;
     View *view;
     Header_Mode header_mode;
#endif
{
  FILE *fp;
  char *filename = NULL;
  Message_List *message_list;
  char *contents;
  int errors = 0;
  Boolean append_it;

  if(view->current->scripting && view->current->script_arg) {
    filename = cpystr(view->current->script_arg);
    append_it = TRUE;
  }

  if(! filename)
    filename = file_select(view->shell,NULL, NULL, NULL, TRUE, &append_it);
  if(filename == NULL)
    return;

  push_cursor(WATCH_CURSOR);

  if((fp = fopen(filename,(append_it == TRUE) ? "a" : "w")) != NULL) {
    (void) chmod(filename,S_IRWXU);

    for(message_list = view->current->message_list; message_list;
	message_list = message_list->next ) {

      if(message_list->selected) {
	contents = view_fetch(message_list->message,
			      (view->current->scripting) 
			      ? view->current->script_header : header_mode);
	if(! contents)
	  continue;
	if((fwrite(contents,strlen(contents),1,fp)) != 1)
	  errors ++;
	fs_give((void **) &contents);
      }
      if(errors)
	break;
    }
    if((fclose(fp)) != SYSCALL_SUCCESS)
      errors ++;
  }

  if(errors)
    mm_log(SAVE_FAILED, WARN);
  else
    mm_log(SAVE_OK, NIL);

  pop_cursor();
  return;
}


#ifdef __STDC__
void view_set_deleted(Widget w, View *view, XtPointer xp)
#else
void view_set_deleted(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_setflag(w,view,DELETED_FLAG);
  return;
}

#ifdef __STDC__
void view_set_seen(Widget w, View *view, XtPointer xp)
#else
void view_set_seen(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_setflag(w,view,SEEN_FLAG);
  return;
}

#ifdef __STDC__
void view_set_flagged(Widget w, View *view, XtPointer xp)
#else
void view_set_flagged(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_setflag(w,view,FLAGGED_FLAG);
  return;
}

#ifdef __STDC__
void view_set_answered(Widget w, View *view, XtPointer xp)
#else
void view_set_answered(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_setflag(w,view,ANSWERED_FLAG);
  return;
}

#ifdef __STDC__
void view_clear_deleted(Widget w, View *view, XtPointer xp)
#else
void view_clear_deleted(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_clearflag(w,view,DELETED_FLAG);
  return;
}

#ifdef __STDC__
void view_clear_seen(Widget w, View *view, XtPointer xp)
#else
void view_clear_seen(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_clearflag(w,view,SEEN_FLAG);
  return;
}

#ifdef __STDC__
void view_clear_flagged(Widget w, View *view, XtPointer xp)
#else
void view_clear_flagged(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_clearflag(w,view,FLAGGED_FLAG);
  return;
}

#ifdef __STDC__
void view_clear_answered(Widget w, View *view, XtPointer xp)
#else
void view_clear_answered(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  view_clearflag(w,view,ANSWERED_FLAG);
  return;
}

#ifdef __STDC__
void update_flags(Lview *lview)
#else
void update_flags(lview)
     Lview *lview;
#endif
{
  Message_List *message_list;

  if((lview == NULL) || (lview->num_selected == 0L))
    return;

  for(message_list = lview->message_list; message_list;
      message_list = message_list->next) {
    if((message_list->selected == TRUE)
       && (message_list->message != NULL))
      mm_flags(message_list->message->mailstream, 
	       message_list->message->msgno);
  }
  return;
}

#ifdef __STDC__
void view_setflag(Widget w, View *view, char *flag)
#else
void view_setflag(w,view,flag)
     Widget w;
     View *view;
     char *flag;
#endif
{
  char *sequence = NULL;

  sequence = generate_sequence(view->current);
  if(sequence == NULL)
    return;

  push_cursor(WATCH_CURSOR);

  mail_setflag(view->current->mailstream,sequence,flag);
  update_flags(view->current);

  fs_give((void **) &sequence);
  pop_cursor();
  return;
}

#ifdef __STDC__
void view_clearflag(Widget w, View *view, char *flag)
#else
void view_clearflag(w,view,flag)
     Widget w;
     View *view;
     char *flag;
#endif
{
  char *sequence = NULL;

  sequence = generate_sequence(view->current);
  if(sequence == NULL)
    return;
  push_cursor(WATCH_CURSOR);

  mail_clearflag(view->current->mailstream,sequence,flag);
  update_flags(view->current);

  fs_give((void **) &sequence);
  pop_cursor();
  return;
}

#ifdef __STDC__
void view_filter(Widget w, View *view, XtPointer xp)
#else
void view_filter(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  create_edfilter_window(view->shell);
  return;
}

#ifdef __STDC__
void view_reply_sender(Widget w, View *view, XtPointer xp)
#else
void view_reply_sender(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Message_List *message_list;
  
  for(message_list = view->current->message_list;
      message_list; message_list = message_list->next) {
    if(message_list->selected)
      compose(REPLYSENDERSTR, MAIL_BEHAVIOUR, COMPOSE_REPLY,
	      NULL, NULL, message_list->message);
  }
  return;
}

#ifdef __STDC__
void view_reply_all(Widget w, View *view, XtPointer xp)
#else
void view_reply_all(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Message_List *message_list;
  
  for(message_list = view->current->message_list;
      message_list; message_list = message_list->next) {
    if(message_list->selected)
      compose(REPLYALLSTR, MAIL_BEHAVIOUR, COMPOSE_REPLYALL,
	      NULL, NULL, message_list->message);
  }
  return;
}

#ifdef __STDC__
void view_remail(Widget w, View *view, XtPointer xp)
#else
void view_remail(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Message_List *message_list;
  
  for(message_list = view->current->message_list;
      message_list; message_list = message_list->next) {
    if(message_list->selected)
      compose(REMAILSTR, MAIL_BEHAVIOUR, COMPOSE_REMAIL,
	      NULL, NULL, message_list->message);
  }
  return;
}

#ifdef __STDC__
void view_forward(Widget w, View *view, XtPointer xp)
#else
void view_forward(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  (void) compose(FORWARDSTR, MAIL_BEHAVIOUR, COMPOSE_FORWARD,
		 NULL, view->current, NULL);
  return;
}

#ifdef __STDC__
void view_forward_attach(Widget w, View *view, XtPointer xp)
#else
void view_forward_attach(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  (void) compose(FORWARDATTACHSTR, MAIL_BEHAVIOUR, 
		 COMPOSE_FORWARDATTACH,
		 NULL, view->current, NULL);
  return;
}

#ifdef __STDC__
int message_list_compare(Message_List *list1,Message_List *list2)
#else
int message_list_compare(list1,list2)
     Message_List *list1;
     Message_List *list2;
#endif
{
  Message_List *ptr1 = list1;
  Message_List *ptr2 = list2;

  for( ; ; ) {
    if((ptr1 != NULL) && (ptr2 != NULL)) {
      if(ptr1->message != ptr2->message)
	return(LISTNOMATCH);
      else {
	ptr1 = ptr1->next;
	ptr2 = ptr2->next;
	continue;
      }
    }
    else {
      if(ptr1 != ptr2)
	return(LISTNOMATCH);
      else
	return(LISTMATCH);
    }
  }
  /* NOTREACHED */
}

#ifdef __STDC__
void update_current_view(Mailbox *mailbox)
#else
void update_current_view(mailbox)
     Mailbox *mailbox;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  Lview *lview;
  Message_List *message_list;
  char ** strtab;

  if(session.view == NULL || session.view->is_realized == FALSE)
    return;
  if(session.view->current == NULL)
    return;

  lview = session.view->current;
  if(lview->count == 0L)
    return;

  /* 
   * If we're pruning the view, zero it and start over. Then
   * set prune_happened so we don't ever do it again. 
   */
    
  XmListDeselectAllItems(session.view->list);

  if(lview->pruned && (lview->prune_happened == FALSE))
    XmListDeleteAllItems(session.view->list);

  strtab = (char **) fs_get(lview->count * (sizeof(char *)));
  for(message_list = lview->message_list, n = 0 ; message_list;
      message_list = message_list->next) {
    strtab[n] = (char *) message_list->message->viewline;
    n ++;
  }

  if(lview->pruned && (lview->prune_happened == FALSE)) {
    XmListAddItems(session.view->list,
		   (XmStringTable) strtab,
		   lview->count, 1);
    if(lview->representation == REP_ASCENDING)
      XmListSetBottomPos(session.view->list, lview->count);
    else
      XmListSetBottomPos(session.view->list,1);
  }

  else
    XmListReplaceItemsPos(session.view->list,
			  (XmStringTable) strtab, 
			  lview->count,1);


  if(lview->pruned)
    lview->prune_happened = TRUE;

  free(strtab);
  n = 0;

  XtSetArg(args[n], XmNselectionPolicy, XmMULTIPLE_SELECT);   n ++;
  XtSetValues(session.view->list, args, n);                   n = 0;

  for(message_list = lview->message_list; message_list;
      message_list = message_list->next) {
    if(message_list->selected == TRUE) 
      XmListSelectPos(session.view->list, message_list->number, FALSE);
  }

  XtSetArg(args[n], XmNselectionPolicy, XmEXTENDED_SELECT);   n ++;
  XtSetValues(session.view->list, args, n);                   n = 0;

  stuff_lview_list(session.view,TRUE,FALSE);

  return;
}

#ifdef __STDC__
void update_bucket(Mailbox *mailbox)
#else
void update_bucket(mailbox)
     Mailbox *mailbox;
#endif
{
  Lview *all;
  Lview *bucket;
  Message_List *message_list;
  Message_List *old;
  Message_List *new;

  all = get_all_lview(mailbox->mailstream);
  bucket = get_bucket_lview(mailbox->mailstream);

  if(all == NULL || bucket == NULL)
    return;

  free_message_list(mailbox->found,TRUE,FALSE);
  mailbox->found = NULL;

  for(message_list = all->message_list; message_list; 
      message_list = message_list->next) {
    if((message_list->message) && (message_list->message->in_view == FALSE))
      mm_searched(mailbox->mailstream,message_list->message->msgno);
  }

  new = mailbox->found;
  mailbox->found = NULL;
  old = bucket->message_list;

  if((message_list_compare(old,new)) == LISTMATCH)
    free_message_list(new,TRUE,FALSE);
  else {
    update_lview_message_list(bucket,old,new,FALSE);
    free_message_list(old,TRUE,FALSE);
  }
  return;
}

#ifdef __STDC__
void set_found_flags(Message_List *message_list)
#else
void set_found_flags(message_list)
     Message_List *message_list;
#endif
{
  Message_List *curr;

  for(curr = message_list; curr; curr = curr->next)
    if(curr->message)
      curr->message->in_view = TRUE;
  return;
}

/* 
 *  The initialize flag says to update all views.
 *  We set it to true when creating a filter, and after an expunge
 *  to catch the one-shot filters.
 */

#ifdef __STDC__
void update_logical_views(Mailbox *mailbox,Boolean initialize)
#else
void update_logical_views(mailbox,initialize)
     Mailbox *mailbox;
     Boolean initialize;
#endif
{
  Lview *lview;
  Message_List *old;
  Message_List *new;

  if(session.view == NULL)
    return;
  
  for(lview = session.view->lview; lview ; lview = lview->prev) {
    if(lview->filter == NULL || lview->mailbox != mailbox)
      continue;

    if(lview->filter->filter_type == FILTER_TYPE_ALL) {
      lview->has_new = has_new_messages(lview,TRUE); /* and wipe "in_view" */
      continue;
    }

    if(lview->filter->filter_type == FILTER_TYPE_BUCKET)
      continue;

    if((lview->filter->filter_type == FILTER_TYPE_ONESHOT)
       && (initialize == FALSE))
      continue;

    new = do_imap_search(lview->mailbox,lview->parent,
			 lview->filter->search_tree);

    new = sort_message_list(new, lview->representation);

    old = lview->message_list;

    if((message_list_compare(old,new)) == LISTMATCH) {
      free_message_list(new,TRUE,FALSE);
      set_found_flags(old);
      continue;
    }
    
    update_lview_message_list(lview,old,new, TRUE);
    free_message_list(old,TRUE,FALSE);
  }

  update_bucket(mailbox);

  stuff_lview_list(session.view,TRUE,FALSE);
  mailbox->update_needed  = 0;
  return;
}

#ifdef __STDC__
int sort_ascending(const void *list1,const void *list2)
#else
int sort_ascending(list1,list2) 
     const void *list1;
     const void *list2;
#endif
{
  Message_List ** l1 = (Message_List **) list1;
  Message_List ** l2 = (Message_List **) list2;

  return((*l1)->message->msgno - (*l2)->message->msgno);
}

#ifdef __STDC__
int sort_reverse(const void *list1, const void *list2)
#else
int sort_reverse(list1,list2) 
     const void *list1;
     const void *list2;
#endif
{
  Message_List ** l1 = (Message_List **) list1;
  Message_List ** l2 = (Message_List **) list2;

  return((*l2)->message->msgno - (*l1)->message->msgno);
}

#ifdef __STDC__
int sort_subject(const void *list1,const void *list2)
#else
int sort_subject(list1,list2) 
     const void *list1;
     const void *list2;
#endif
{
  Message_List ** l1 = (Message_List **) list1;
  Message_List ** l2 = (Message_List **) list2;

  int result;
  unsigned char *ptr1;
  unsigned char *ptr2;


  if((*l1)->message->subject == NULL) 
    ptr1 = (unsigned char *) EMPTYSTR;
  else
    ptr1 = (unsigned char *) (*l1)->message->subject;

  if((*l2)->message->subject == NULL)
    ptr2 = (unsigned char *) EMPTYSTR;
  else
    ptr2 = (unsigned char *) (*l2)->message->subject;

  do {
    if((result = strncasecmp((char *) ptr1,"re",2)) == STRMATCH) {
      if((*(ptr1 + 2) == ':') || (*(ptr1 + 2) == '[')) { 
	while(*ptr1 && ! isspace(*ptr1))
	  ptr1 ++;
	while(*ptr1 && isspace(*ptr1))
	  ptr1 ++;
      }
      else
	result = ~STRMATCH;
    }
  } while(result == STRMATCH);

  do {
    if((result = strncasecmp((char *) ptr2,"re",2)) == STRMATCH) {
      if((*(ptr2 + 2) == ':') || (*(ptr2 + 2) == '[')) { 
	while(*ptr2 && ! isspace(*ptr2))
	  ptr2 ++;
	while(*ptr2 && isspace(*ptr2))
	  ptr2 ++;
      }
      else
	result = ~STRMATCH;
    }
  } while(result == STRMATCH);

  while((*ptr1) && (isspace(*ptr1)))
    ptr1 ++;

  while((*ptr2) && (isspace(*ptr2)))
    ptr2 ++;

  result = strcasecmp((char *) ptr1,(char *) ptr2);
  if(result == STRMATCH)
    return((*l1)->message->msgno - (*l2)->message->msgno);
  else
    return(result);
}


#ifdef __STDC__
int sort_sender(const void *list1, const void *list2)
#else
int sort_sender(list1,list2) 
     const void *list1;
     const void *list2;
#endif
{
  Message_List ** l1 = (Message_List **) list1;
  Message_List ** l2 = (Message_List **) list2;

  int result;
  char * ptr1, *ptr2;

  if((*l1)->message->sender)
    ptr1 = (*l1)->message->sender;
  else
    ptr1 = EMPTYSTR;

  if((*l2)->message->sender)
    ptr2 = (*l2)->message->sender;
  else
    ptr2 = EMPTYSTR;

  result = strcasecmp(ptr1, ptr2);
  if(result == STRMATCH)
    return((*l1)->message->msgno - (*l2)->message->msgno);
  else
    return(result);

}	       

#ifdef __STDC__
Message_List * sort_message_list(Message_List *message_list, 
				 Representation_Type representation)
#else
Message_List * sort_message_list(message_list, representation)
     Message_List *message_list;
     Representation_Type representation;
#endif
{
  Message_List **sort;
  Message_List *ml_ptr;
  Message_List *result = NULL;
  int n = 0;
  int total = 0;

  if(message_list == NULL)
    return(message_list);

  for(ml_ptr = message_list; ml_ptr; ml_ptr = ml_ptr->next)
    total ++;

  if(total < 2)
    return(message_list);

  sort = (Message_List **) fs_get(total * sizeof(Message_List *));

  for(ml_ptr = message_list; ml_ptr; ml_ptr = ml_ptr->next) {
    if(ml_ptr->message == NULL)
      continue;
    sort[n] = ml_ptr; 
    n ++;
  }

  switch (representation) {
  case REP_DESCENDING:
    qsort((char *) sort,total,sizeof(Message_List **),sort_reverse);
    break;
  case REP_SORT_SUBJECT:
    qsort((char *) sort,total,sizeof(Message_List **),sort_subject);
    break;
  case REP_SORT_SENDER:
    qsort((char *) sort,total,sizeof(Message_List **),sort_sender);
    break;
  case REP_ASCENDING:
  default:
    qsort((char *) sort,total,sizeof(Message_List **),sort_ascending);
    break;
  }

  for(n = 0; n < total; n ++) {
    ml_ptr = sort[n];
    if(n == 0) {
      result = ml_ptr;
      ml_ptr->next = sort[n + 1];
      ml_ptr->prev = NULL;
      ml_ptr->number = n + 1;
    }
    else {
      ml_ptr->prev = sort[n - 1];
      ml_ptr->number = n + 1;
      if(n == (total - 1))
	ml_ptr->next = NULL;
      else
	ml_ptr->next = sort[n + 1];
    }
  }
  fs_give((void **) &sort);
  return(result);
}

/* 
 *  This function has been overloaded a bit to save running through the list
 * a second time to clear the "in_view" flag.
 */

#ifdef __STDC__
Boolean has_new_messages(Lview *lview,Boolean wipe)
#else
Boolean has_new_messages(lview,wipe)
     Lview *lview;
     Boolean wipe;
#endif
{
  Message_List *message_list;
  MESSAGECACHE *elt;
  Boolean has_new = FALSE;

  for(message_list = lview->message_list; message_list;
      message_list = message_list->next) {

    if(! message_list->message)
      continue;

    if(wipe)
      message_list->message->in_view = FALSE;

    if((message_list->message->fetched == TRUE)
       && (message_list->message->longcache)) {
      elt = &(message_list->message->longcache->elt);
      if((! elt->seen) && elt->recent) {
	has_new = TRUE;
	if(wipe == FALSE)
	  return(has_new);
      }
    }
  }
  return(has_new);
}

#ifdef __STDC__
void update_lview_message_list(Lview *lview,
			       Message_List *old, Message_List *new,
			       Boolean update_in_view_flag)
#else
void update_lview_message_list(lview,old,new,update_in_view_flag)
     Lview *lview;
     Message_List *old;
     Message_List *new;
     Boolean update_in_view_flag;
#endif
{
  int n = 0;
  Message_List *list1, *list2;

  for(list1 = old; list1; list1 = list1->next) {
    for(list2 = new; list2; list2 = list2->next) {
      if(list1->message != list2->message)
	continue;
      if(list1->selected == TRUE)
	list2->selected = TRUE;
    }
  }

  for(list2 = new; list2; list2 = list2->next) {
    n ++;
    if((list2->message) && (update_in_view_flag == TRUE))
      list2->message->in_view = TRUE;
  }

  lview->count = n;
  lview->message_list = new;

  lview->has_new = has_new_messages(lview,FALSE);

  if(lview == session.view->current)
    view_context_switch(session.view,session.view->current,lview);
  return;
}


#ifdef __STDC__
void view_set_keyword(Widget w, View *view, XtPointer xp)
#else
void view_set_keyword(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Message_List *message_list;
  char *keyword = get_keyword(view->shell,view->current->mailstream);
  
  if(keyword) {
    view_setflag(w,view,keyword);
    fs_give((void **) &keyword);
    for(message_list = view->current->message_list; message_list;
	message_list = message_list->next) {
      if((message_list->selected == TRUE) &&
	 (message_list->message != NULL)) {
	update_view_line_stream_msgno(message_list->message->mailstream,
				      message_list->message->msgno);
      }
    }
  }
  return;
}

#ifdef __STDC__
void view_unset_keyword(Widget w, View *view, XtPointer xp)
#else
void view_unset_keyword(w,view,xp)
     Widget w;
     View *view;
     XtPointer xp;
#endif
{
  Message_List *message_list;
  char *keyword = get_keyword(view->shell,view->current->mailstream);
  
  if(keyword) {
    view_clearflag(w,view,keyword);
    fs_give((void **) &keyword);
    for(message_list = view->current->message_list; message_list;
	message_list = message_list->next) {
      if((message_list->selected == TRUE) &&
	 (message_list->message != NULL)) {
	update_view_line_stream_msgno(message_list->message->mailstream,
				      message_list->message->msgno);
      }
    }
  }
  return;
}


#ifdef __STDC__
char *get_keyword(Widget w, MAILSTREAM *mailstream)
#else
char *get_keyword(w,mailstream)
     Widget w;
     MAILSTREAM *mailstream;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  Widget form, menubar;
  char *keyword = NULL;
  int num = 0;
  Keyword_Win *keyword_win = NULL;
  XmString xstr;


  if((mailstream == NULL) || (*mailstream->user_flags == NULL))
    return(NULL);

  push_cursor(PIRATE_CURSOR);

  keyword_win = (Keyword_Win *) fs_get(sizeof(Keyword_Win));
  keyword_win->done = 0;
  keyword_win->keyword = NULL;

  XtSetArg(args[n], XtNx, -1000); n ++;
  XtSetArg(args[n], XtNy, -1000); n ++;

  keyword_win->shell = XtCreatePopupShell("keyword_shell",
			 topLevelShellWidgetClass, w,
			 args, n); n = 0;

  setup_editres(keyword_win->shell);

  if(pirate_icon != (Pixmap) None)
    XtVaSetValues(keyword_win->shell,
		  XmNiconPixmap,pirate_icon,
		  NULL);
	
  form = XmCreateForm(keyword_win->shell, "form", args, n ); n = 0;

  XtSetArg(args[n], XmNtopAttachment,   XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNleftAttachment,  XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n ++;
  menubar = XmCreateMenuBar(form, "menubar", args, n); n = 0;
  XtManageChild(menubar);

  create_buttons(NULL, menubar, keyword_menu, 
		      XtNumber(keyword_menu),
		      BTN_ON, (XtPointer) keyword_win, ROOTMENULEVEL);


  XtSetArg(args[n], XmNscrollBarDisplayPolicy, XmSTATIC);             n ++;
  XtSetArg(args[n], XmNlistSizePolicy,         XmCONSTANT);           n ++;
  XtSetArg(args[n], XmNvisibleItemCount,       5 );                  n ++; 
  XtSetArg(args[n], XmNselectionPolicy,        XmSINGLE_SELECT);      n ++;
  keyword_win->list = XmCreateScrolledList(form,"list",args,n);      n = 0;

  XtAddCallback(keyword_win->list, XmNsingleSelectionCallback, 
		(XtCallbackProc) keyword_select, keyword_win);

  XtAddCallback(keyword_win->list, XmNdefaultActionCallback, 
		(XtCallbackProc) keyword_accept, keyword_win);

  XtSetArg(args[n], XmNtopAttachment,   XmATTACH_WIDGET); n ++;
  XtSetArg(args[n], XmNtopWidget,       menubar); n ++;
  XtSetArg(args[n], XmNleftAttachment,  XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n ++;

  XtSetValues(XtParent(keyword_win->list), args, n); n = 0;

  for(num = 0; mailstream->user_flags[num] != NULL ; num ++) {
    xstr = XmStringCreateSimple(mailstream->user_flags[num]);
    XmListAddItemUnselected(keyword_win->list,xstr,0);
    XmStringFree(xstr);
  }

  XtAddCallback(keyword_win->shell, XmNdestroyCallback,
		(XtCallbackProc) keyword_destroy,keyword_win);

  XtManageChild(keyword_win->list);
  XtManageChild(form);
  XtManageChild(keyword_win->shell);

  XtPopup(keyword_win->shell, XtGrabExclusive);
  position_popup_widget(keyword_win->shell);
  modal_main_loop(&keyword_win->done);

  XtPopdown(keyword_win->shell);
  keyword = keyword_win->keyword;
  XtDestroyWidget(keyword_win->shell);

  if(keyword_win)
    fs_give((void **) &keyword_win);

  pop_cursor();
  return(keyword);
}

#ifdef __STDC__
void keyword_accept(Widget w,Keyword_Win *keyword_win,XtPointer xp)
#else
void keyword_accept(w,keyword_win,xp)
     Widget w;
     Keyword_Win *keyword_win;
     XtPointer xp;
#endif
{
  keyword_win->done = MODAL_LOOP_DONE;
  return;
}

#ifdef __STDC__
void keyword_destroy(Widget w, Keyword_Win *keyword_win,XtPointer xp)
#else
void keyword_destroy(w,keyword_win,xp)
     Widget w;
     Keyword_Win *keyword_win;
     XtPointer xp;
#endif
{
  if(keyword_win) {
    if(keyword_win->keyword)
      fs_give((void **) &keyword_win->keyword);
    fs_give((void **) &keyword_win);
  }
  pop_cursor();
  return;
}

#ifdef __STDC__
void keyword_select(Widget w, Keyword_Win *keyword_win,
		    XmListCallbackStruct *xp)
#else
void keyword_select(w,keyword_win,xp)
     Widget w;
     Keyword_Win *keyword_win;
     XmListCallbackStruct *xp;
#endif
{
  char *str;
  XmStringGetLtoR(xp->item,XmSTRING_DEFAULT_CHARSET, &str);
  if(keyword_win->keyword) {
    if((strcmp(keyword_win->keyword,str)) == STRMATCH) {
      fs_give((void **) &keyword_win->keyword);
      fs_give((void **) &str);
      return;
    }
    else
      fs_give((void **) &keyword_win->keyword);
  }
  keyword_win->keyword = str;
  return;
}

#ifdef __STDC__
void keyword_cancel(Widget w, Keyword_Win *keyword_win, XtPointer xp)
#else
void keyword_cancel(w,keyword_win,xp)
     Widget w;
     Keyword_Win *keyword_win;
     XtPointer xp;
#endif
{
  if(keyword_win->keyword)
    fs_give((void **) &keyword_win->keyword);
  keyword_win->keyword = NULL;
  keyword_win->done = MODAL_LOOP_DONE;
  return;
}

#ifdef __STDC__
void keyword_help(Widget w, Keyword_Win *keyword_win, XtPointer xp)
#else
void keyword_help(w,keyword_win,xp)
     Widget w;
     Keyword_Win *keyword_win;
     XtPointer xp;
#endif
{
  help(keyword_win->shell, KEYWORDHELPTITLE, 
       KEYWORDHELPFILE, DOCTYPE_ENRICHED);
  return;
}


