/* winfncs.c
 *
 *
 * $Author: mtm $
 * $Date: 1995/05/09 01:27:15 $
 * $Id: winfncs.c,v 1.1 1995/05/09 01:27:15 mtm Exp mtm $
 *
 * Mike Macgirvin <Mike_Macgirvin@CAMIS.Stanford.EDU>
 *
 * Copyright 1995 by The Leland Stanford Junior University.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notices appear in all copies and that both the
 * above copyright notices and this permission notice appear in supporting
 * documentation, and that the name of The Leland Stanford Junior University 
 * not be used in advertising or publicity pertaining to distribution of the 
 * software without specific, written prior permission.  This software is made 
 * available "as is", and THE LELAND STANFORD JUNIOR UNIVERSITY DISCLAIMS ALL 
 * WARRANTIES, EXPRESS OR IMPLIED, WITH REGARD TO THIS SOFTWARE, INCLUDING 
 * WITHOUT LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE, AND IN NO EVENT SHALL THE LELAND STANFORD JUNIOR 
 * UNIVERSITY BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR 
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER 
 * IN AN ACTION OF CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, 
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
 * SOFTWARE.
 *
 */

/*
 * $Log: winfncs.c,v $
 * Revision 1.1  1995/05/09  01:27:15  mtm
 * Initial revision
 *
 * 
 */

#include "ml.h"

#include <X11/cursorfont.h>

/* globals */

Cursor clock_cursor = 0;
Cursor top_left_cursor = 0;
Cursor pirate_cursor = 0;


Cursor create_left_cursor();
Cursor create_clock_cursor();
Cursor create_pirate_cursor();

Menu help_help_menu[] = {
  { NULL, "dismiss", NUL_TERM,
      help_dismiss, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "print", NUL_TERM,
      help_print, NULL, 0, NULL, NULL, BTN_ON },
  { NULL, "HELP", NUL_TERM,
      help_help, NULL, 0, NULL, NULL, BTN_ON },
};

typedef struct _CURSOR_STACK_ {
  int stack[64];
  int pointer;
} Cursor_Stack;

Cursor_Stack cursor_stack;

#ifdef __STDC__
void push_cursor(int cursor_type)
#else
void push_cursor(cursor_type)
     int cursor_type;
#endif
{
  cursor_stack.pointer ++;
  cursor_stack.stack[cursor_stack.pointer] = cursor_type;
  set_cursor(cursor_type);
  return;
}

#ifdef __STDC__
void pop_cursor(void)
#else
void pop_cursor()
#endif
{
  if(cursor_stack.pointer) {
    cursor_stack.pointer --;
    set_cursor(cursor_stack.stack[cursor_stack.pointer]);
  }
  else
    reset_cursors();
}

#ifdef __STDC__
void set_cursor(int cursor_type)
#else
void set_cursor(cursor_type)
     int cursor_type;
#endif
{
  switch (cursor_type) {
  case PIRATE_CURSOR:
    set_pirate_cursors();
    break;
  case WATCH_CURSOR:
    set_watch_cursors();
    break;
  case DEFAULT_CURSOR:
  default:
    reset_cursors();
    break;
  }
  return;
}

#ifdef __STDC__
void create_cursors(Display *display)
#else
void create_cursors(display)
     Display *display;
#endif
{
  clock_cursor    = create_clock_cursor(display);
  top_left_cursor = create_left_cursor(display);
  pirate_cursor   = create_pirate_cursor(display);
  cursor_stack.pointer = 0;
  return;
}

#ifdef __STDC__
Cursor create_left_cursor(Display *display)
#else
Cursor create_left_cursor(display)
     Display *display;
#endif
{
  return (XCreateFontCursor(display, XC_top_left_arrow));
}

#ifdef __STDC__
Cursor create_pirate_cursor(Display *display)
#else
Cursor create_pirate_cursor(display)
     Display *display;
#endif
{
  return(XCreateFontCursor(display, XC_pirate));
}

#ifdef __STDC__
Cursor create_clock_cursor(Display *display)
#else
Cursor create_clock_cursor(display)
     Display *display;
#endif
{
  return(XCreateFontCursor(display, XC_watch));
}

#ifdef __STDC__
void display_fontcursor(Display *display, Window w, Cursor cursor)
#else
void display_fontcursor(display, w, cursor)
     Display *display;
     Window w;
     Cursor cursor;
#endif
{
  if(cursor)
    XDefineCursor(display, w, cursor);
  return;
}

#ifdef __STDC__
void reset_cursor(Display *display, Window w)
#else
void reset_cursor(display, w)
     Display *display;
     Window w;
#endif
{
  XUndefineCursor(display, w);
  XFlush(display);
  return;
}

#ifdef __STDC__
void set_watch_cursors(void)
#else
void set_watch_cursors()
#endif
{
  Gentext *gentext;
  session.busy = TRUE;
  XDefineCursor(display, XtWindow(top), clock_cursor);
  XDefineCursor(display, XtWindow(status), clock_cursor);
  
  if(session.view && session.view->is_realized)
    XDefineCursor(display, XtWindow(session.view->shell), clock_cursor);
  if(session.mailboxlist && session.mailboxlist->is_realized)
    XDefineCursor(display, XtWindow(session.mailboxlist->shell), clock_cursor);
  if(session.view && session.view->list_is_realized)
    XDefineCursor(display,XtWindow(session.view->list_shell), clock_cursor);
  if(session.edfilter && session.edfilter->is_realized)
    XDefineCursor(display,XtWindow(session.edfilter->shell), clock_cursor);
  if(session.address_book && session.address_book->is_realized)
    XDefineCursor(display, XtWindow(session.address_book->shell),clock_cursor);
  if(session.compose && session.compose->is_realized)
    XDefineCursor(display, XtWindow(session.compose->shell), clock_cursor);
  if(session.compose && session.compose->edit && 
     session.compose->edit->is_realized)
    XDefineCursor(display,XtWindow(session.compose->edit->shell),clock_cursor);
  if(session.compose && session.compose->search_win && 
     session.compose->search_win->is_realized)
    XDefineCursor(display,
		  XtWindow(session.compose->search_win->shell), clock_cursor);
  if(session.compose && session.compose->list_is_realized)
    XDefineCursor(display,XtWindow(session.compose->list_shell), clock_cursor);
  if(session.read && session.read->is_realized)
    XDefineCursor(display, XtWindow(session.read->shell), clock_cursor);
  if(session.read && session.read->list_is_realized)
    XDefineCursor(display, XtWindow(session.read->list_shell), clock_cursor);
  if(session.prefst && session.prefst->is_realized)
    XDefineCursor(display, XtWindow(session.prefst->shell), clock_cursor);
    if(session.pref2st && session.pref2st->is_realized)
    XDefineCursor(display, XtWindow(session.pref2st->shell), clock_cursor);
  if(session.note && session.note->is_realized)
    XDefineCursor(display, XtWindow(session.note->shell), clock_cursor);

  for(gentext = session.gentext; gentext; gentext = gentext->next)
    XDefineCursor(display,XtWindow(gentext->shell), clock_cursor);

    
  XFlush(display);
  return;
}

#ifdef __STDC__
void set_pirate_cursors(void)
#else
void set_pirate_cursors()
#endif
{
  Gentext *gentext;
  session.busy = TRUE;
  XDefineCursor(display, XtWindow(top), pirate_cursor);
  XDefineCursor(display, XtWindow(status), pirate_cursor);
  
  if(session.view && session.view->is_realized) 
    XDefineCursor(display,  XtWindow(session.view->shell), pirate_cursor);
  if(session.mailboxlist && session.mailboxlist->is_realized)
    XDefineCursor(display,XtWindow(session.mailboxlist->shell), pirate_cursor);
  if(session.view && session.view->list_is_realized)
    XDefineCursor(display, XtWindow(session.view->list_shell), pirate_cursor);
  if(session.edfilter && session.edfilter->is_realized)
    XDefineCursor(display, XtWindow(session.edfilter->shell), pirate_cursor);
    if(session.address_book && session.address_book->is_realized)
    XDefineCursor(display,XtWindow(session.address_book->shell),pirate_cursor);
  if(session.compose && session.compose->is_realized) 
    XDefineCursor(display, XtWindow(session.compose->shell), pirate_cursor);
  if(session.compose && session.compose->edit 
     && session.compose->edit->is_realized)
    XDefineCursor(display,
		  XtWindow(session.compose->edit->shell),pirate_cursor);
  if(session.compose && session.compose->search_win 
     && session.compose->search_win->is_realized)
    XDefineCursor(display, XtWindow(session.compose->search_win->shell), 
		  pirate_cursor);
  if(session.compose && session.compose->list_is_realized)
    XDefineCursor(display,XtWindow(session.compose->list_shell),pirate_cursor);
  if(session.read && session.read->is_realized)
    XDefineCursor(display,XtWindow(session.read->shell), pirate_cursor);
  if(session.read && session.read->list_is_realized)
    XDefineCursor(display, XtWindow(session.read->list_shell), pirate_cursor);
  if(session.prefst && session.prefst->is_realized)
    XDefineCursor(display, XtWindow(session.prefst->shell), pirate_cursor);
  if(session.pref2st && session.pref2st->is_realized)
    XDefineCursor(display, XtWindow(session.pref2st->shell), pirate_cursor);
  if(session.note && session.note->is_realized)
    XDefineCursor(display, XtWindow(session.note->shell), pirate_cursor);

  for(gentext = session.gentext; gentext; gentext = gentext->next)
    XDefineCursor(display,XtWindow(gentext->shell), pirate_cursor);



  XFlush(display);
  return;
}

#ifdef __STDC__
void reset_cursors(void)
#else
void reset_cursors()
#endif
{
  Gentext *gentext;
  XUndefineCursor(display, XtWindow(top));
  XUndefineCursor(display, XtWindow(status));

  if(session.view && session.view->is_realized)
    XUndefineCursor(display, XtWindow(session.view->shell));
  if(session.mailboxlist && session.mailboxlist->is_realized)
    XUndefineCursor(display, XtWindow(session.mailboxlist->shell));
  if(session.view && session.view->list_is_realized)
    XUndefineCursor(display, XtWindow(session.view->list_shell));
  if(session.edfilter && session.edfilter->is_realized)
    XUndefineCursor(display,XtWindow(session.edfilter->shell));
  if(session.address_book && session.address_book->is_realized)
    XUndefineCursor(display,XtWindow(session.address_book->shell));
  if(session.compose && session.compose->is_realized) 
    XUndefineCursor(display,XtWindow(session.compose->shell));
  if(session.compose && session.compose->edit 
     && session.compose->edit->is_realized)
    XUndefineCursor(display,XtWindow(session.compose->edit->shell));
  if(session.compose && session.compose->search_win 
     && session.compose->search_win->is_realized)
    XUndefineCursor(display,XtWindow(session.compose->search_win->shell));
  if(session.compose && session.compose->list_is_realized)
    XUndefineCursor(display,XtWindow(session.compose->list_shell));
  if(session.read && session.read->is_realized) 
    XUndefineCursor(display,XtWindow(session.read->shell));
  if(session.read && session.read->list_is_realized)
    XUndefineCursor(display,XtWindow(session.read->list_shell)); 
  if(session.prefst && session.prefst->is_realized)
    XUndefineCursor(display,XtWindow(session.prefst->shell));
  if(session.pref2st && session.pref2st->is_realized)
    XUndefineCursor(display,XtWindow(session.pref2st->shell));
  if(session.note && session.note->is_realized)
    XUndefineCursor(display,XtWindow(session.note->shell));

  for(gentext = session.gentext; gentext; gentext = gentext->next)
    XUndefineCursor(display,XtWindow(gentext->shell));

  XFlush(display);
  session.busy = FALSE;
  return;
}

#ifdef __STDC__
void get_pointer_position(Widget w, Position *x, Position *y)
#else
void get_pointer_position(w,x,y)
     Widget w;
     Position *x;
     Position *y;
#endif
{
  Window junk1, junk2;
  int root_x, root_y, junkX, junkY;
  unsigned int junkM;
  
  if (XQueryPointer(XtDisplay(w), XtWindow(w), 
		    &junk1,  &junk2, 
		    &root_x, &root_y, 
		    &junkX,  &junkY, 
		    &junkM) == FALSE) 
    return;
  
  *x = (short) root_x;
  *y = (short) root_y;
  return;
}

#ifdef __STDC__
void get_screen_size(int *x, int *y)
#else
void get_screen_size(x,y)
     int *x;
     int *y;
#endif
{
  int screen_number;

  screen_number = XDefaultScreen(display);

  *x = XDisplayWidth(display,screen_number);
  *y = XDisplayHeight(display,screen_number);
  return;
}


#define BORDER_SLOP 60

#ifdef __STDC__
void position_popup_widget(Widget w)
#else
void position_popup_widget(w)
     Widget w;
#endif
{
  int rsize_x, rsize_y;
  Dimension wid_h, wid_w, wid_b;
  Position p_x, p_y;


  get_screen_size(&rsize_x, &rsize_y);
  get_pointer_position(w, &p_x, &p_y);

  XtRealizeWidget(w); 

  XtVaGetValues(w, 
		XmNheight, &wid_h, 
		XmNwidth, &wid_w, 
		XmNborderWidth, &wid_b,
		NULL);

  p_x -= (Position) (wid_w / 2) - wid_b;
  p_y -= (Position) (wid_h / 2) - wid_b;

  if(p_x < BORDER_SLOP)
    p_x = BORDER_SLOP;
  else 
    if(((int) (p_x + wid_w + wid_b)) > rsize_x)
      p_x = rsize_x - wid_w - wid_b - BORDER_SLOP;

  if(p_y < BORDER_SLOP)
    p_y = BORDER_SLOP;
  else
    if(((int) (p_y + wid_h + wid_b)) > rsize_y)
      p_y = rsize_y - wid_h - wid_b - BORDER_SLOP;

  XtVaSetValues(w, XtNx, p_x, XtNy, p_y, NULL);
}



/* To determine if an item in a list widget was selected or deselected */

#ifdef __STDC__
Boolean is_selected(XmListCallbackStruct *ptr, unsigned long num)
#else
Boolean is_selected(ptr,num)
     XmListCallbackStruct *ptr;
     unsigned long num;
#endif
{
  register int i;
  for(i = 0; i < ptr->selected_item_count; i ++) {
    if(ptr->selected_item_positions[i] == num)
      return(TRUE);
  }
  return(FALSE);
}

#ifdef __STDC__
Widget create_label(Widget form, Widget below, char *string)
#else
Widget create_label(form,below,string)
     Widget form;
     Widget below;
     char *string;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  Widget label;

  XtSetArg(args[n], XmNtopAttachment,   XmATTACH_WIDGET ); n ++;
  XtSetArg(args[n], XmNtopWidget,       below           ); n ++;
  XtSetArg(args[n], XmNleftAttachment,  XmATTACH_FORM   ); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM   ); n ++;
  XtSetArg(args[n], XmNborderWidth,     0               ); n ++;

  label = XmCreateLabel(form, string, args, n);            n = 0;
  XtManageChild(label);

  return(label);
}




#ifdef __STDC__
Widget create_text_field(Widget form, Widget below, char *label_string,
			 char *string, int width, 
			 XtPointer fn, XtPointer data)
#else
Widget create_text_field(form,below,label_string,string,width,fn,data)
     Widget form;
     Widget below;
     char *label_string;
     char *string;
     int width;
     XtPointer fn;
     XtPointer data;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  XtTranslations translations;
  Widget this_widget;
  Widget label;
  char buffer[FILEBUFFLEN];

  XtSetArg(args[n], XmNtopAttachment,   XmATTACH_WIDGET ); n ++;
  XtSetArg(args[n], XmNtopWidget,       below           ); n ++;
  XtSetArg(args[n], XmNleftAttachment,  XmATTACH_FORM   ); n ++;
  XtSetArg(args[n], XmNborderWidth,     0               ); n ++;
  
  strcpy(buffer,label_string);
  strcat(buffer,"_lbl");
  label = XmCreateLabel(form, buffer, args, n);      n = 0;
  XtManageChild(label);

  XtSetArg(args[n], XmNtopAttachment,   XmATTACH_WIDGET ); n ++;
  XtSetArg(args[n], XmNtopWidget,       below           ); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM   ); n ++;
  XtSetArg(args[n], XmNleftAttachment,  XmATTACH_WIDGET ); n ++;
  XtSetArg(args[n], XmNleftWidget,      label           ); n ++;
  if(width) {
    XtSetArg(args[n], XmNcolumns,       width           ); n ++;
  }
  strcpy(buffer,label_string);
  strcat(buffer,"_textfld");

  this_widget = XmCreateTextField(form, buffer, args, n);  n = 0;
  XtManageChild(this_widget);

  if(string)
    AppendText(this_widget,string);

  translations = XtParseTranslationTable(GLOBAL_text_field_translations);
  XtOverrideTranslations(this_widget,translations);

  if(fn) {
    translations = 
      XtParseTranslationTable(GLOBAL_terminal_text_field_translations);
    XtOverrideTranslations(this_widget,translations);  
    XtAddCallback(this_widget, XmNactivateCallback, (XtCallbackProc) fn, data);
  }
  else {
    translations = 
      XtParseTranslationTable(GLOBAL_nonterminal_text_field_translations);
    XtOverrideTranslations(this_widget,translations);
  }

  XtAddCallback(this_widget, XmNmodifyVerifyCallback, 
		(XtCallbackProc) text_field_edit, NULL);

  return(this_widget);
}

#ifndef XEW

#ifdef __STDC__
void help(Widget w, char *title, char *filename, Doc_Type doctype)
#else
void help(w,title,filename,doctype)
     Widget w;
     char *title;
     char *filename;
     Doc_Type doctype;
#endif
{
  char buffer[MAXPATHLEN];
  char altbuffer[MAXPATHLEN];
  char *local_language = NULL;

  sprintf(buffer,"%s/%s",preferences.help_directory,filename);
  strcpy(altbuffer,buffer);
  local_language = getenv("LANG");
  if(local_language == NULL)
    local_language = getenv("ML_LANG");
  if(local_language != NULL) {
    strcat(altbuffer,".");
    strcat(altbuffer,local_language);
  }
  if((access(altbuffer,R_OK)) == SYSCALL_SUCCESS)
    display_text(w,title,altbuffer,NULL,doctype);
  else
    display_text(w,title,buffer,NULL,doctype);
  return;
}



#ifdef __STDC__
void display_text(Widget w, char *title, char *filename, 
		  char *string, Doc_Type doctype)
#else
void display_text(w,title,filename,string, doctype)
     Widget w;
     char *title;
     char *filename;
     char *string;
     Doc_Type doctype;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  Widget form, menubar, text;
  XtTranslations translations;
  Gentext *gentext;
  FILE *fp;
  char buffer[FILEBUFFLEN];

  if(filename) {
    if((fp = fopen(filename,"r")) == NULL) {
      sprintf(buffer,CANNOT_READ_FILE,filename);
      mm_log(buffer,WARN);
      return;
    }
  }

  gentext = (Gentext *) fs_get(sizeof(Gentext));
  gentext->next = NULL;
  gentext->prev = NULL;


  if(title) {
    XtSetArg(args[n], XmNtitle, title); n ++;
  }

  gentext->shell = XtCreateWidget((filename) ? "helpwin" : "textwin",
				  topLevelShellWidgetClass,
				    w, args, n ); 
  n = 0;

  setup_editres(gentext->shell);

  if(help_icon != (Pixmap) None)
    XtVaSetValues(gentext->shell,
		  XmNiconPixmap,help_icon,
		  NULL);


  form = XmCreateForm(gentext->shell, "form", args, n); n = 0;

  XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n ++;
  menubar = XmCreateMenuBar(form,"menubar", args, n); n = 0;
  XtManageChild(menubar);

  create_buttons(NULL, menubar,
		      help_help_menu, XtNumber(help_help_menu), 
		      BTN_ON, (XtPointer) gentext, ROOTMENULEVEL);

  XtSetArg(args[n], XmNeditable, FALSE);                         n ++;
  XtSetArg(args[n], XmNeditMode, XmMULTI_LINE_EDIT);             n ++;
  XtSetArg(args[n], XmNwordWrap, TRUE);                          n ++;
  XtSetArg(args[n], XmNscrollVertical, TRUE);                    n ++;
  XtSetArg(args[n], XmNscrollHorizontal, FALSE );                n ++;
  gentext->text = XmCreateScrolledText(form, "text", args, n);
  n = 0;

  translations = XtParseTranslationTable(GLOBAL_text_translations);
  XtOverrideTranslations(gentext->text, translations);

  XtManageChild(gentext->text);

  if(filename) {
    while(fgets(buffer,sizeof(buffer),fp))
      AppendText(gentext->text,buffer);
    fclose(fp);
  }
  else 
    XmTextSetString(gentext->text,string);


  XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET);          n ++;
  XtSetArg(args[n], XmNtopWidget, menubar);                      n ++;
  XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM);           n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM);          n ++;
  XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM);         n ++;

  XtSetValues(XtParent(gentext->text),args,n); n = 0;

  XtManageChild(form);

  XmTextShowPosition(gentext->text,0);
  XmTextSetInsertionPosition(gentext->text,0);


  XtAddCallback(gentext->shell, XmNdestroyCallback,
		(XtCallbackProc) help_destroy, gentext);

  if(session.gentext == NULL)
    session.gentext = gentext;
  else {
    gentext->next = session.gentext;
    session.gentext->prev = gentext;
    session.gentext = gentext;
  }

  XtManageChild(gentext->shell);
  XtRealizeWidget(gentext->shell);
  XFlush(display);

  return;
}

#ifdef __STDC__
void help_dismiss(Widget w, Gentext *gentext, XtPointer xp)
#else
void help_dismiss(w,gentext,xp)
     Widget w;
     Gentext *gentext;
     XtPointer xp;
#endif
{
  XtRemoveCallback(gentext->shell, XmNdestroyCallback,
		(XtCallbackProc) help_destroy, gentext);
  XtDestroyWidget(gentext->shell);
  if(gentext->next)
    gentext->next->prev = gentext->prev;
  if(gentext->prev)
    gentext->prev->next = gentext->next;
  else
    session.gentext = gentext->next;
  fs_give((void **) &gentext);

  return;
}


#ifdef __STDC__
void help_print(Widget w, Gentext *gentext, XtPointer xp)
#else
void help_print(w,gentext,xp)
     Widget w;
     Gentext *gentext;
     XtPointer xp;
#endif
{
  char *txt;

  if(*preferences.print_command == NUL_TERM) {
    mm_log(NO_PRINT_COMMAND, WARN);
    return;
  }

  push_cursor(WATCH_CURSOR);

  txt = XmTextGetString(gentext->text);
  if(txt != NULL) {
    if(write_to_pipe(preferences.print_command, NULL,
		     txt, strlen(txt)))
      mm_log(PRINT_FAILED, WARN);
    else
      mm_log(PRINT_OK, NIL);
    fs_give((void **) &txt);
  }
  pop_cursor();
  return;
}

#ifdef __STDC__
void help_help(Widget w, Gentext *gentext, XtPointer xp)
#else
void help_help(w,gentext,xp)
     Widget w;
     Gentext *gentext;
     XtPointer xp;
#endif
{
  help(gentext->shell, HELPHELPTITLE, HELPHELPFILE, DOCTYPE_ENRICHED);
  return;
}


#ifdef __STDC__
void help_destroy(Widget w, Gentext *gentext, XtPointer xp)
#else
void help_destroy(w,gentext,xp)
     Widget w;
     Gentext *gentext;
     XtPointer xp;
#endif
{
  if(gentext->next)
    gentext->next->prev = gentext->prev;
  if(gentext->prev)
    gentext->prev->next = gentext->next;
  else
    session.gentext = gentext->next;

  fs_give((void **) &gentext);

  return;
}




#else /* XEW */

#include <X11/Xew/Basic.h>
#include <X11/Xew/Text.h>
#include <X11/Xew/TextEd.h>
#include <X11/Xew/Frame.h>
#include <Xm/ScrolledW.h>



static char default_charset[] = "\33(B\33)B\33.A\33/A\17\33\175";

typedef struct RTMapping
{
  char *tag;
  char *sgr;
} RTMapping;

static RTMapping mapping[] =
{
	{ "bold",	"\033[1m" },
	{ "italic",	"\033[3m" },
	{ "underline",	"\033[4m" },
	{ "excerpt",	"\033[34m" },
	{ "fixed",	"\033[19m" }, /* Assuming alternate font 9!! */
	{ "/bold",	"\033[22m" },
	{ "/italic",	"\033[23m" },
	{ "/underline",	"\033[24m" },
	{ "/excerpt",	"\033[39m" },
	{ "/fixed",	"\033[10m" }, /* Assuming primary font (0)!! */
	{ "bigger",	"\033[150 B" },
	{ "smaller",	"\033[70 B" },
	{ "/bigger",	"\033[100 B" },
	{ "/smaller",	"\033[100 B" },

	{ "flushleft",	"\033[1;5 F" },
	{ "flushright",	"\033[1;7 F" },
	{ "flushboth",	"\033[1;2 F" },
	{ "center",	"\033[1;6 F" },
	{ "nofill",	"\033[0 F" },
	{ "verbatim",	"\033[0 F" },

	{ "/flushleft",	"\033[1 F" },
	{ "/flushright","\033[1 F" },
	{ "/flushboth",	"\033[1 F" },
	{ "/center",	"\033[2 H\033[0;1 F" },
	{ "/nofill",    "\033[1 F" },
	{ "/verbatim",	"\033[1 F" },
	{ "x-black",    "\033[30m" },
	{ "x-red",      "\033[31m" },
	{ "x-green",    "\033[32m" },
	{ "x-yellow",   "\033[33m" },
	{ "x-blue",     "\033[34m" },
	{ "x-magenta",  "\033[35m" },
	{ "x-cyan",     "\033[36m" },
	{ "x-white",    "\033[37m" },
	{ "/x-black",   "\033[39m" },
	{ "/x-red",     "\033[39m" },
	{ "/x-green",   "\033[39m" },
	{ "/x-yellow",  "\033[39m" },
	{ "/x-blue",    "\033[39m" },
	{ "/x-magenta", "\033[39m" },
	{ "/x-cyan",    "\033[39m" },
	{ "/x-white",   "\033[39m" },
	{ "/x-color",   "\033[39m" },
	{ NULL,		NULL },
      };


static RTMapping params[] = {

	{ "black",    "\033[30m" },
	{ "red",      "\033[31m" },
	{ "green",    "\033[32m" },
	{ "yellow",   "\033[33m" },
	{ "blue",     "\033[34m" },
	{ "magenta",  "\033[35m" },
	{ "cyan",     "\033[36m" },
	{ "white",    "\033[37m" },
	{ NULL,		NULL },

      };

unsigned char plain_prefix[] = "\033[0 F\033[0;19m"; 
unsigned char enriched_prefix[] = "\033[0m\033[10m\033[0 F\033[1 F";

/* This call munges the calling string. */

#ifdef __STDC__
unsigned char * RichToPlain(char *s, Doc_Type doctype)
#else
unsigned char * RichToPlain(s, doctype)
     char *s;
     Doc_Type doctype;
#endif
{

  unsigned char *ret;
  unsigned char *src;
  unsigned char *dst;
  unsigned char *ptr;
  unsigned char *ptr2;
  Boolean x_color = FALSE;
  RTMapping *h;

  if(s == NULL)
    return((unsigned char *) cpystr(EMPTYSTR));

  ret = (unsigned char *) 
    fs_get(((long) strlen(s)) + ((long) (strlen(s)) / 2L) 
	   + (long) FILEBUFFLEN);
  *ret = NUL_TERM;
  dst = ret;

  if(doctype == DOCTYPE_ENRICHED) {
    strcat((char *) ret, (char *) enriched_prefix);
    dst += strlen((char *) enriched_prefix);
  }
  else {
    strcat((char *)ret,(char *)plain_prefix);
    strcat((char *)ret,s);
    return(ret);
  }
  
  for(src = (unsigned char *) s; *src != NUL_TERM; src ++) {
    if((strncasecmp((char *)src,"<param>",7)) == STRMATCH) {
      if(x_color == TRUE) {
	src += 7;
	for(h = params; h->tag; h ++) {
	  if((strncasecmp((char *) src, 
			  (char *) h->tag,
			  strlen((char *) h->tag))) == STRMATCH) {
	    strcat((char *) dst,(char *) h->sgr);
	    dst += strlen((char *) h->sgr);
	    break;
	  }
	}
      }
      while((*src != NUL_TERM) 
	    && (strncasecmp((char *)src,"</param>",8)) != STRMATCH)
	src ++;
      if(*src != NUL_TERM)
	src = (unsigned char *) strchr((char *)src,'>');
      continue;
    }
    if((strncasecmp((char *) src, "<x-color>",9)) == STRMATCH) {
      x_color = TRUE;
      src += 8;
      continue;
    }
    if((strncasecmp((char *)src,"<verbatim>",10)) == STRMATCH) {
      src += 10;
      strcat((char *) dst,(char *) plain_prefix);
      dst += strlen((char *) plain_prefix);
      while((*src != NUL_TERM) 
	    && (strncasecmp((char *)src,"</verbatim>",11)) != STRMATCH) {
	*dst = *src;
	dst ++;
	*dst = NUL_TERM;
	src ++;
      }
      if(*src != NUL_TERM) {
	src = (unsigned char *) strchr((char *)src,'>');
	strcat((char *) dst,(char *) enriched_prefix);
	dst += strlen((char *) enriched_prefix);
      }
      continue;
    }
    if((*src == '<') && (*(src + 1) == '<')) {
      *dst = *src;
      dst ++;
      *dst = NUL_TERM;
      src ++;
      continue;
    }
    if(*src == '<') {
      src ++;
      ptr = (unsigned char *) strchr((char *) src,'>');
      if(! ptr) {
	mm_log(RICHTEXTNOTAG, ERROR);
	*ret = NUL_TERM;
	return(ret);
      }
      *ptr = NUL_TERM;
      for(h = mapping; h->tag; h ++) {
	if((strcasecmp((char *) src, (char *) h->tag)) == STRMATCH) {
	  strcat((char *) dst,(char *) h->sgr);
	  dst += strlen((char *) h->sgr);
	  if((strcasecmp(h->tag,"/x-color")) == STRMATCH)
	    x_color = FALSE;
	  break;
	}
      }
      src = ptr;
      continue;
    }
    if(*src == LFCHAR) {
      if(*(src + 1) == LFCHAR) {
	while(*(src + 1) == LFCHAR) {
	  *dst = LFCHAR;
	  dst ++;
	  *dst = NUL_TERM;
	  src ++;
	}
	continue;
      }
      else {
	*dst = SPACECHAR;
	dst ++;
	*dst = NUL_TERM;
	continue;
      }
    }
    *dst = *src;
    dst ++;
    *dst = NUL_TERM;
  }
  *dst = NUL_TERM;
  return(ret);
}



#ifdef __STDC__
void help(Widget w, char *title, char *filename, Doc_Type doctype)
#else
void help(w,title,filename, doctype)
     Widget w;
     char *title;
     char *filename;
     Doc_Type doctype;
#endif
{
  char buffer[MAXPATHLEN];
  sprintf(buffer,"%s/%s",preferences.help_directory,filename);
  display_text(w,title,buffer,NULL,doctype);
  return;
}



#ifdef __STDC__
void display_text(Widget w, char *title, char *filename, 
		  char *string, Doc_Type doctype)
#else
void display_text(w,title,filename,string, doctype)
     Widget w;
     char *title;
     char *filename;
     char *string;
     Doc_Type doctype;
#endif
{
  Arg args[ARGLISTSIZE];
  int n = 0;
  Widget form, form2, menubar;
  XtTranslations translations;
  Gentext *gentext;
  FILE *fp;
  char buffer[FILEBUFFLEN];
  Binary_Buffer *binary_buffer;
  Dimension height, width;
  Dimension scroll_height;
  Dimension text_height;

  if(filename) {
    if((fp = fopen(filename,"r")) == NULL) {
      sprintf(buffer,CANNOT_READ_FILE,filename);
      mm_log(buffer,WARN);
      return;
    }
  }

  push_cursor(WATCH_CURSOR);

  gentext = (Gentext *) fs_get(sizeof(Gentext));
  gentext->next = NULL;
  gentext->prev = NULL;
  gentext->contents = NULL;

  if(title) {
    XtSetArg(args[n], XmNtitle, title); n ++;
  }

  XtSetArg(args[n], XtNx, -1000); n ++;
  XtSetArg(args[n], XtNy, -1000); n ++;

  gentext->shell = XtCreateWidget((filename) ? "helpwin" : "textwin",
				  topLevelShellWidgetClass,
				    w, args, n ); 
  n = 0;

  setup_editres(gentext->shell);

  if(help_icon != (Pixmap) None)
    XtVaSetValues(gentext->shell,
		  XmNiconPixmap,help_icon,
		  NULL);

  form = XmCreateForm(gentext->shell, "form", args, n); n = 0;

  XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n ++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n ++;
  menubar = XmCreateMenuBar(form,"menubar", args, n); n = 0;
  XtManageChild(menubar);

  create_buttons(NULL, menubar,
		      help_help_menu, XtNumber(help_help_menu), 
		      BTN_ON, (XtPointer) gentext, ROOTMENULEVEL);


  /* dang pixel values.... */

  XtSetArg(args[n], XtNwidth, 650); n ++;  
  XtSetArg(args[n], XtNheight,300); n ++;

  XtSetArg(args[n], XmNtraversalOn,      FALSE);                 n ++;
  XtSetArg(args[n], XmNscrollingPolicy,  XmAUTOMATIC);           n ++;
  XtSetArg(args[n], XmNtopAttachment,    XmATTACH_WIDGET);       n ++;
  XtSetArg(args[n], XmNtopWidget,        menubar);               n ++;
  XtSetArg(args[n], XmNleftAttachment,   XmATTACH_FORM);         n ++;
  XtSetArg(args[n], XmNrightAttachment,  XmATTACH_FORM);         n ++;
  XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM);         n ++;

  gentext->scroll = XmCreateScrolledWindow(form,"scroll", args, n); n = 0;

  if((doctype == DOCTYPE_PLAIN) || (doctype == DOCTYPE_ENRICHED)) {

    if(filename) {
      binary_buffer = load_binary_file(filename);
      if(binary_buffer && binary_buffer->data) {
	gentext->contents = RichToPlain((char *) binary_buffer->data, doctype);
      }
    }
    else {
      binary_buffer = (Binary_Buffer *)fs_get(sizeof(Binary_Buffer));
      binary_buffer->data = (unsigned char *) cpystr(string);
      gentext->contents = RichToPlain((char *) binary_buffer->data, doctype);
    }

    free_binary_buffer(binary_buffer);
  }

  XtManageChild(gentext->scroll); 

  XtVaGetValues(gentext->scroll, XmNclipWindow, &gentext->clip, NULL);

  XtSetArg(args[n], XtNalignment,     XeAlignment_JUSTIFIED);      n ++;
  XtSetArg(args[n], XtNwidth,         620);                        n ++;
  XtSetArg(args[n], XtNformat,        FALSE);                      n ++;
  XtSetArg(args[n], XtNproportional,  TRUE);                       n ++;
  XtSetArg(args[n], XtNresize,        TRUE);                       n ++;
  XtSetArg(args[n], XtNcontentString, gentext->contents);          n ++;
  XtSetArg(args[n], XtNcontentLength, strlen((char *)gentext->contents)); n ++;
  XtSetArg(args[n], XtNallowEdit,     FALSE);                      n ++;
  XtSetArg(args[n], XtNdisplayCaret,  FALSE);                      n ++;
  XtSetArg(args[n], XtNexportFormat,  XeTextExport_STRING);        n ++;
  XtSetArg(args[n], XtNborderWidth,   0);                          n ++;
  gentext->text = XtCreateWidget("xetext",xeTextEdWidgetClass,
				 gentext->scroll,args,n); 
  n = 0;

  XmScrolledWindowSetAreas(gentext->scroll, 
			   (Widget) NULL, (Widget) NULL, gentext->text);

  XtManageChild(gentext->text);

  XtAddEventHandler(gentext->scroll, SubstructureNotifyMask, FALSE,
		    (XtEventHandler) resize_xetext, gentext);

  XtManageChild(form);

  XtAddCallback(gentext->shell, XmNdestroyCallback,
		(XtCallbackProc) help_destroy, gentext);

  if(session.gentext == NULL)
    session.gentext = gentext;
  else {
    gentext->next = session.gentext;
    session.gentext->prev = gentext;
    session.gentext = gentext;
  }

  XtManageChild(gentext->shell);
  XtRealizeWidget(gentext->shell);

  resize_xetext(gentext->scroll, gentext, NULL);
  XmUpdateDisplay(gentext->shell);
  position_popup_widget(gentext->shell);
  pop_cursor();
  return;
}


#ifdef __STDC__
void resize_xetext(Widget w, Gentext *gentext, XEvent *event)
#else
void resize_xetext(w,gentext,event)
     Widget w;
     Gentext *gentext;
     XEvent *event;
#endif
{

  Dimension width;

  XtRemoveEventHandler(gentext->scroll,SubstructureNotifyMask,
		       FALSE, (XtEventHandler) resize_xetext, gentext); 


  if((XtIsRealized(gentext->shell))
     && (XtIsRealized(gentext->text))) {
    if((event == NULL) || (event->type == ConfigureNotify)) {
      XtVaGetValues(gentext->clip, XmNwidth, &width, NULL);
      if(width > 5)
	XtVaSetValues(gentext->text, XtNcolumnWidth,width - 5, NULL);
      
    }
  }

  XtAddEventHandler(gentext->scroll, SubstructureNotifyMask, FALSE,
		    (XtEventHandler) resize_xetext, gentext);

  return;
}






#ifdef __STDC__
void help_dismiss(Widget w, Gentext *gentext, XtPointer xp)
#else
void help_dismiss(w,gentext,xp)
     Widget w;
     Gentext *gentext;
     XtPointer xp;
#endif
{
  XtRemoveEventHandler(gentext->scroll,SubstructureNotifyMask,
		       FALSE, (XtEventHandler) resize_xetext, gentext); 
  
  XtRemoveCallback(gentext->shell, XmNdestroyCallback,
		   (XtCallbackProc) help_destroy, gentext);
  XtDestroyWidget(gentext->shell);

  if(gentext->next)
    gentext->next->prev = gentext->prev;
  if(gentext->prev)
    gentext->prev->next = gentext->next;
  else
    session.gentext = gentext->next;
  if(gentext->contents)
    fs_give((void **) &gentext->contents);
  fs_give((void **) &gentext);
  return;
}


#ifdef __STDC__
void help_print(Widget w, Gentext *gentext, XtPointer xp)
#else
void help_print(w,gentext,xp)
     Widget w;
     Gentext *gentext;
     XtPointer xp;
#endif
{
  char *txt;
  char *wrapped_text;
  int error = 0;
  long length;

  if(*preferences.print_command == NUL_TERM) {
    mm_log(NO_PRINT_COMMAND, WARN);
    return;
  }

  push_cursor(WATCH_CURSOR);

  txt = XeTextGetString(gentext->text, &length);
  
  if(txt != NULL) {
    wrapped_text = wrap_text(txt, COMPOSEWIDTH - 1);
    if(wrapped_text) {
      if(write_to_pipe(preferences.print_command, NULL,
		       wrapped_text, strlen(wrapped_text)))
	error ++;
      fs_give((void **) &wrapped_text);
    }
    else 
      error ++;
    fs_give((void **) &txt);
  }
  if(error)
    mm_log(PRINT_FAILED, WARN);
  else
    mm_log(PRINT_OK, NIL);

  pop_cursor();
  return;
}

#ifdef __STDC__
void help_help(Widget w, Gentext *gentext, XtPointer xp)
#else
void help_help(w,gentext,xp)
     Widget w;
     Gentext *gentext;
     XtPointer xp;
#endif
{
  help(gentext->shell, HELPHELPTITLE, HELPHELPFILE, DOCTYPE_ENRICHED);
  return;
}


#ifdef __STDC__
void help_destroy(Widget w, Gentext *gentext, XtPointer xp)
#else
void help_destroy(w,gentext,xp)
     Widget w;
     Gentext *gentext;
     XtPointer xp;
#endif
{
  XtRemoveEventHandler(gentext->scroll,SubstructureNotifyMask,
		       FALSE, (XtEventHandler) resize_xetext, gentext); 

  if(gentext->next)
    gentext->next->prev = gentext->prev;
  if(gentext->prev)
    gentext->prev->next = gentext->next;
  else
    session.gentext = gentext->next;
  if(gentext->contents)
    fs_give((void **) &gentext->contents);
  fs_give((void **) &gentext);

  return;
}


#endif /* XEW */
